/* -*- mode: c++; -*-
 *-----------------------------------------------------------------------------
 * $RCSfile: Security.h,v $
 *
 * See Copyright for the status of this software.
 *
 * The OpenSOAP Project
 * http://opensoap.jp/
 *-----------------------------------------------------------------------------
 */
#ifndef OpenSOAP_Security_H
#define OpenSOAP_Security_H

#include <stdio.h>

#include <OpenSOAP/Envelope.h>

#ifdef __cplusplus
extern "C" {
#endif /* __cplusplus */

/* Hash Algorithms */
#define OPENSOAP_HA_MD2    (1)  /* MD2 */
#define OPENSOAP_HA_MD4    (2)  /* MD4 */
#define OPENSOAP_HA_MD5    (3)  /* MD5 */
#define OPENSOAP_HA_MDC2   (4)  /* MDC2 */
#define OPENSOAP_HA_RIPEMD (5)  /* RIPEMD */
#define OPENSOAP_HA_SHA    (6)  /* SHA */

	/* Generate RSA Keys To Stream */
	int
	OPENSOAP_API
	OpenSOAPSecGenerateRSAKeys
	(const unsigned char* szSeedPhrase, /* (i)  Seed phrase */
	 FILE*          fpPrivKey,    /* (i)  RSA Private Key File stream */
	 FILE*          fpPubKey);    /* (i)  RSA Public Key File stream */

	/* Generate RSA Keys To File */
	int
	OPENSOAP_API
	OpenSOAPSecGenerateRSAKeysToFile
	(const unsigned char* szSeedPhrase,      /* (i)  Seed phrase */
	 const char*          szPrivKeyFileName, /* (i)  RSA Private Key File Name */
	 const char*          szPubKeyFileName); /* (i)  RSA Public Key File Name */

	/* Encrypt Envelope Using Public Key File Stream */
	int
	OPENSOAP_API
	OpenSOAPSecEncWithStream
	(OpenSOAPEnvelopePtr env,       /* (i)  Envelope */
	 FILE*               fpPubKey); /* (i)  Public Key File stream */
	
	/* Encrypt Envelope Using Public Key File */
	int
	OPENSOAP_API
	OpenSOAPSecEncWithFile
	(OpenSOAPEnvelopePtr env,         /* (i)  Envelope */
	 const char*         szPubKName); /* (i)  Public Key File Name */

	/* Decrypt Envelope Using Public Key File Stream */
	int
	OPENSOAP_API
	OpenSOAPSecDecWithStream
	(OpenSOAPEnvelopePtr env,        /* (i)  Envelope */
	 FILE*               fpPrivKey); /* (i)  Private Key File stream */
	
	/* Encrypt Envelope Using Public Key File */
	int
	OPENSOAP_API
	OpenSOAPSecDecWithFile
	(OpenSOAPEnvelopePtr env,          /* (i)  Envelope */
	 const char*         szPrivKName); /* (i)  Private Key File Name */

	/* Add Signature to SOAP Message using PrivateKey File Stream */
	int
	OPENSOAP_API
	OpenSOAPSecAddSignWithStream
	(OpenSOAPEnvelopePtr env,        /* (io) Envelope */
	 int                 iType,      /* (i)  Hash Type(OPENSOAP_HA_*) */
	 FILE*               fpPrivKey,  /* (i)  Private Key File stream */
	 OpenSOAPStringPtr   name);      /* (i)  Name of Sign */

	/* Add Signature to SOAP Message using PrivateKey File */
	int
	OPENSOAP_API
	OpenSOAPSecAddSignWithFile
	(OpenSOAPEnvelopePtr env,          /* (io) Envelope */
	 int                 iType,        /* (i)  Hash Type(OPENSOAP_HA_*) */
	 const char*         szPrivKName,  /* (i)  Private Key File Name */
	 OpenSOAPStringPtr   name);        /* (i)  Name of Sign */

	/* Count Signature */
	int
	OPENSOAP_API
	OpenSOAPSecCntSign
	(const OpenSOAPEnvelopePtr env,    /* (i)  Envelope */
	 int*                      pnSig); /* (o)  Signature Count */

	/* Get "SignedBy" List of Signature */
	int
	OPENSOAP_API
	OpenSOAPSecGetSignedByList
	(OpenSOAPEnvelopePtr env,          /* (i)  Envelope */
	 int                 nCntMax,      /* (i)  Maximum Count */
	 OpenSOAPStringPtr   list[],       /* (io) SignedBy List  */
	 int*                pnCntPacked); /* (o)  Packed Count */

	/* Verify Signature using PublicKey File Stream */
	int
	OPENSOAP_API
	OpenSOAPSecVerifySignWithStream
	(OpenSOAPEnvelopePtr env,       /* (io) Envelope */
	 FILE*               fpPubKey); /* (i)  Public Key File stream */

	/* Verify Signature using PublicKey File */
	int
	OPENSOAP_API
	OpenSOAPSecVerifySignWithFile
	(OpenSOAPEnvelopePtr env,         /* (io) Envelope */
	 const char*         szPubKName); /* (i)  Public Key File Name */

#ifdef __cplusplus
}
#endif /* __cplusplus */

#endif /* OpenSOAP_Security_H */
