/*-----------------------------------------------------------------------------
 * $RCSfile: XMLAttr.c,v $
 *
 * See Copyright for the status of this software.
 *
 * The OpenSOAP Project
 * http://opensoap.jp/
 *-----------------------------------------------------------------------------
 */

#if (!defined(_WIN32)) || defined(_DEBUG)
static  const   char    CVS_ID[]        =       "$Id: XMLAttr.c,v 1.22 2002/01/21 02:01:09 jin Exp $";
#endif  /* _DEBUG */

#include "XMLAttr.h"
#include <OpenSOAP/Serializer.h>

/*
=begin
= OpenSOAP XML Attribute class
=end
 */

static
int
OpenSOAPXMLAttrReleaseMembers(OpenSOAPXMLAttrPtr /* [in, out] */ attr) {
    int ret = OpenSOAPXMLNodeReleaseMembers((OpenSOAPXMLNodePtr)attr);

    return ret;
}

static
int
OpenSOAPXMLAttrFree(OpenSOAPObjectPtr /* [in, out] */ obj) {
    int ret = OpenSOAPXMLAttrReleaseMembers((OpenSOAPXMLAttrPtr)obj);

    if (OPENSOAP_SUCCEEDED(ret)) {
        free(obj);
    }

    return ret;
}

static
int
OpenSOAPXMLAttrInitialize(OpenSOAPXMLAttrPtr /* [in, out] */ attr) {
    int ret = OpenSOAPXMLNodeInitialize((OpenSOAPXMLNodePtr)attr,
                                        OpenSOAPXMLAttrFree);

    if (attr) {
        attr->definedElement = NULL;
    }

    return ret;
}

/*
=begin
--- function#OpenSOAPXMLAttrCreateString(attr_name, attr)
    OpenSOAP XML Attribute instance create(OpenSOAPString)

    :Parameters
      :OpenSOAPStringPtr [in] ((|attr_name|))
        XML Attribute name.
      :OpenSOAPXMLAttrPtr * [out] ((|attr|))
        XML Attribute pointer
    :Return Value
      :int
	    Return Code Is ErrorCode
=end
 */
extern
int
OPENSOAP_API
OpenSOAPXMLAttrCreateString(OpenSOAPStringPtr /* [in] */ attr_name,
                            OpenSOAPXMLAttrPtr * /* [out] */ attr) {
    int ret = OPENSOAP_PARAMETER_BADVALUE;

    if (attr) {
        ret = OPENSOAP_MEM_BADALLOC;
        *attr = malloc(sizeof(OpenSOAPXMLAttr));
        if (*attr) {
            ret = OpenSOAPXMLAttrInitialize(*attr);
            if (OPENSOAP_SUCCEEDED(ret)) {
                ret = OpenSOAPXMLNodeSetName((OpenSOAPXMLNodePtr)*attr,
                                             attr_name,
                                             1);
                if (OPENSOAP_FAILED(ret)) {
                    OpenSOAPXMLAttrReleaseMembers(*attr);
                }
            }
            if (OPENSOAP_FAILED(ret)) {
                free(*attr);
                *attr = NULL;
            }
        }
    }
    
    return ret;
}

/*
=begin
--- function#OpenSOAPXMLAttrCreateMB(attr_name, attr)
    OpenSOAP XML Attribute instance create(MB)

    :Parameters
      :const char * [in] ((|attr_name|))
        XML Attribute name.
      :OpenSOAPXMLAttrPtr * [out] ((|attr|))
        XML Attribute pointer
    :Return Value
      :int
	    Return Code Is ErrorCode
=end
 */
extern
int
OPENSOAP_API
OpenSOAPXMLAttrCreateMB(const char * /* [in] */ attr_name,
                        OpenSOAPXMLAttrPtr * /* [out] */ attr) {
#if 1
    OpenSOAPStringPtr attr_name_str = NULL;
    int ret = OpenSOAPStringCreateWithMB(attr_name, &attr_name_str);

    if (OPENSOAP_SUCCEEDED(ret)) {
        ret = OpenSOAPXMLAttrCreateString(attr_name_str, attr);
        OpenSOAPStringRelease(attr_name_str);
    }
#else    
    int ret = OPENSOAP_YET_IMPLEMENTATION;
    if(attr){
        *attr = malloc(sizeof(OpenSOAPXMLAttr));
        if(*attr){
            (*attr)->super.prev = NULL;
            (*attr)->super.next = NULL;
            (*attr)->super.parent = NULL;
            (*attr)->super.children = NULL;

            ret = OpenSOAPStringCreate( &((*attr)->super.name) );
            if( OPENSOAP_FAILED(ret) ){
                return ret;
            }

            ret = OpenSOAPStringSetStringMB((*attr)->super.name, attr_name);
            if(OPENSOAP_FAILED(ret)){
                return ret;
            }
            (*attr)->super.value = NULL;
            (*attr)->super.thisNamespace = NULL;
            (*attr)->definedElement = NULL;

            ret = OPENSOAP_NO_ERROR;
        }
        else{
            ret = OPENSOAP_MEM_BADALLOC;
        }
    }
    else{
        ret = OPENSOAP_PARAMETER_BADVALUE;
    }
#endif    
    return ret;
}

/*
=begin
--- function#OpenSOAPXMLAttrCreateWC(attr_name, attr)
    OpenSOAP XML Attribute instance create(WC)

    :Parameters
      :const wchar_t * [in] ((|attr_name|))
        XML Attribute name.
      :OpenSOAPXMLAttrPtr * [out] ((|attr|))
        XML Attribute pointer
    :Return Value
      :int
	    Return Code Is ErrorCode
=end
 */
extern
int
OPENSOAP_API
OpenSOAPXMLAttrCreateWC(const wchar_t * /* [in] */ attr_name,
                        OpenSOAPXMLAttrPtr * /* [out] */ attr) {
    OpenSOAPStringPtr attr_name_str = NULL;
    int ret = OpenSOAPStringCreateWithWC(attr_name, &attr_name_str);

    if (OPENSOAP_SUCCEEDED(ret)) {
        ret = OpenSOAPXMLAttrCreateString(attr_name_str, attr);
        OpenSOAPStringRelease(attr_name_str);
    }
    return ret;
}

/*
=begin
--- function#OpenSOAPXMLAttrSetNamespaceString(elm, ns_uri, ns_prefix)
    Set Namespace of XML Element(OpenSOAPString)

    :Parameters
      :OpenSOAPXMLAttrPtr [in, out] ((|elm|))
        XML Attribute.
      :OpenSOAPStringPtr [in] ((|ns_uri|))
        Namespace URI
      :OpenSOAPStringPtr [in] ((|ns_prefix|))
        Namespace Prefix

    :Return
      :int
	    Return Code Is ErrorCode
	:Note
	  This Function Is Static Function
=end
 */
/* extern */
static
int
/* OPENSOAP_API */
OpenSOAPXMLAttrSetNamespaceString(OpenSOAPXMLAttrPtr /* [in, out] */ elm,
                                  OpenSOAPStringPtr /* [in] */ ns_uri,
                                  OpenSOAPStringPtr /* [in] */ ns_prefix) {
    int ret = OPENSOAP_YET_IMPLEMENTATION;
    int compare_result;

    OpenSOAPXMLElmPtr next_parent;
    OpenSOAPXMLElmPtr  attr_elm;
    OpenSOAPXMLNamespacePtr next_tmp;

    if ( !elm ){
        ret = OPENSOAP_PARAMETER_BADVALUE;
        return ret;
    }

    attr_elm = elm->definedElement;


    /* check if there is the ns_uri in itself definedXMLNsList */
    next_tmp = attr_elm->definedXMLNsList;

    while( next_tmp != NULL ){
        ret = OpenSOAPStringCompare(next_tmp->uri, ns_uri, &compare_result);
        if( OPENSOAP_FAILED(ret) ){
            return ret;
        }

        if( compare_result != 0 ){
            next_tmp = next_tmp->next;
        }
        else{
            elm->super.thisNamespace = next_tmp;
            ret = OPENSOAP_NO_ERROR;
            return ret;
        }
    }

    /* If there is no ns_uri in itself definedXMLNsList,
       check if there is the ns_uri in its parent's definedXMLNsList */
    next_parent = (OpenSOAPXMLElmPtr)((OpenSOAPXMLNodePtr)(attr_elm))->parent;
	
    while( next_parent != NULL ){
        next_tmp = next_parent->definedXMLNsList;

        while( next_tmp != NULL ){
            ret = OpenSOAPStringCompare(next_tmp->uri,
                                        ns_uri,
                                        &compare_result);
            if( OPENSOAP_FAILED(ret) ){
                return ret;
            }

            if( compare_result != 0 ){
                next_tmp = next_tmp->next;
            }
            else{
                elm->super.thisNamespace = next_tmp;
                ret = OPENSOAP_NO_ERROR;
                return ret;
            }
        }
        next_parent = (OpenSOAPXMLElmPtr)(next_parent->super.parent);
    }

    /* If there is no ns_uri, make a new XMLNamespace, 
       input the new XMLNamespace to definedXMLNsList,
       input its own XMLElmPtr to definedElment */
    ret = OpenSOAPXMLNamespaceCreateString(	ns_uri,
                                                ns_prefix,
                                                &(elm->super.thisNamespace));
    if( OPENSOAP_FAILED(ret) ){
        return ret;
    }

    if( elm->definedElement->definedXMLNsList == NULL ){
        elm->definedElement->definedXMLNsList = elm->super.thisNamespace;
    }
    else{
        ret = OpenSOAPXMLNodeInsertNext( (OpenSOAPXMLNodePtr)(elm->definedElement->definedXMLNsList), 
                                         (OpenSOAPXMLNodePtr)(elm->super.thisNamespace));
        if( OPENSOAP_FAILED(ret) ){
            return ret;
        }
    }

    elm->super.thisNamespace->definedElement = elm->definedElement;

    return ret;
}

/*
=begin
--- function#OpenSOAPXMLAttrSetNamespaceMB(elm, ns_uri, ns_prefix)
    Set Namespace of XML Element(MB)

    :Parameters
      :OpenSOAPXMLAttrPtr [in, out] ((|elm|))
        XML Attribute.
      :const char * [in] ((|ns_uri|))
        Namespace URI
      :const char * [in] ((|ns_prefix|))
        Namespace Prefix

    :Return
      :int
	    Return Code Is ErrorCode
=end
 */
extern
int
OPENSOAP_API
OpenSOAPXMLAttrSetNamespaceMB(OpenSOAPXMLAttrPtr /* [in, out] */ elm,
                              const char * /* [in] */ ns_uri,
                              const char * /* [in] */ ns_prefix) {
    int ret = OPENSOAP_YET_IMPLEMENTATION;

	OpenSOAPXMLElmPtr  attr_elm = NULL;
	OpenSOAPXMLNamespacePtr ns = NULL;

	attr_elm = elm->definedElement;

	ret = OpenSOAPXMLElmSearchNamespaceMB( attr_elm, ns_uri, ns_prefix, &ns );
	if( OPENSOAP_FAILED(ret) ){
		return ret;
	}


	/* there is a namespace defined by ns_uri and ns_prefix */
	if( ns != NULL ){
		elm->super.thisNamespace = ns;
	}
	/* If there is no ns_uri and ns_prefix, make a new XMLNamespace,
	   input the new XMLNamespace to the definedXMLNsList,
	   input itself XMLElmPtr to definedElment */
	else{
		ret = OpenSOAPXMLNamespaceCreateMB(	ns_uri,
											ns_prefix,
											&(elm->super.thisNamespace));
		if( OPENSOAP_FAILED(ret) ){
			return ret;
		}

		if( !(elm->definedElement->definedXMLNsList)){
			elm->definedElement->definedXMLNsList = elm->super.thisNamespace;
		}
		else{
			ret = OpenSOAPXMLNodeInsertNext( (OpenSOAPXMLNodePtr)(elm->definedElement->definedXMLNsList), 
											 (OpenSOAPXMLNodePtr)(elm->super.thisNamespace));
			if( OPENSOAP_FAILED(ret) ){
				return ret;
			}
		}

		elm->super.thisNamespace->definedElement = elm->definedElement;
	}

	ret = OPENSOAP_NO_ERROR;

    return ret;
}

/*
=begin
--- function#OpenSOAPXMLAttrSetNamespaceWC(elm, ns_uri, ns_prefix)
    Set Namespace of XML Element(WC)

    :Parameters
      :OpenSOAPXMLAttrPtr [in, out] ((|elm|))
        XML Attribute.
      :const wchar_t * [in] ((|ns_uri|))
        Namespace URI
      :const wchar_t * [in] ((|ns_prefix|))
        Namespace Prefix

    :Return
      :int
	    Return Code Is ErrorCode
=end
 */
extern
int
OPENSOAP_API
OpenSOAPXMLAttrSetNamespaceWC(OpenSOAPXMLAttrPtr /* [in, out] */ elm,
                              const wchar_t * /* [in] */ ns_uri,
                              const wchar_t * /* [in] */ ns_prefix) {
    OpenSOAPStringPtr ns_uri_str = NULL;
    int ret = OpenSOAPStringCreateWithWC(ns_uri, &ns_uri_str);

    if (OPENSOAP_SUCCEEDED(ret)) {
        OpenSOAPStringPtr ns_prefix_str = NULL;
        int ret = OpenSOAPStringCreateWithWC(ns_prefix, &ns_prefix_str);
        if (OPENSOAP_SUCCEEDED(ret)) {
            ret = OpenSOAPXMLAttrSetNamespaceString(elm, ns_uri_str, ns_prefix_str);
            OpenSOAPStringRelease(ns_prefix_str);
        }
        OpenSOAPStringRelease(ns_uri_str);
    }
    return ret;
}

/*
=begin
--- function#OpenSOAPXMLAttrGetNamespace(elm, ns)
    Get Namespace of XML Element

    :Parameters
      :OpenSOAPXMLAttrPtr [in, out] ((|elm|))
        XML Attribute
      :OpenSOAPXMLNamespacePtr * [out] ((|ns|))
        Namespace 
    :Return
      :int
	    Return Code Is ErrorCode
=end
 */
extern
int
OPENSOAP_API
OpenSOAPXMLAttrGetNamespace(OpenSOAPXMLAttrPtr /* [in, out] */ elm,
                            OpenSOAPXMLNamespacePtr * /* [out] */ ns) {
    int ret = OPENSOAP_YET_IMPLEMENTATION;

    if( !elm || !ns ){
        ret = OPENSOAP_PARAMETER_BADVALUE;
    }
    else{
        *ns = elm->super.thisNamespace;
        ret = OPENSOAP_NO_ERROR;
    }

    return ret;
}

/*
=begin
--- function#OpenSOAPXMLAttrGetValueMB(elm, type_name, value)
    Get Value of XML Attribute(MB)

    :Parameters
      :OpenSOAPXMLAttrPtr [in] ((|elm|))
        OpenSOAP XML Attribute.
      :const char * [in] ((|type_name|))
        Type Name
      :void * [out] ((|value|))
        Strage Buffer Pointer
    :Return value
      :int
	    Return Code Is ErrorCode
=end
*/
extern
int
OPENSOAP_API
OpenSOAPXMLAttrGetValueMB(OpenSOAPXMLAttrPtr /* [in] */ elm,
                         const char * /* [in] */ type_name,
                         void * /* [out] */ value) {
    int ret = OPENSOAP_YET_IMPLEMENTATION;

	OpenSOAPDeserializerFunc deserializer = NULL;

	if( !elm ){
		ret = OPENSOAP_PARAMETER_BADVALUE;
		return ret;
	}

	ret = OpenSOAPGetDeserializerMB(type_name, &deserializer);
	if (OPENSOAP_SUCCEEDED(ret)) {
		ret = deserializer(((OpenSOAPXMLNodePtr)elm)->value, value);
		if( OPENSOAP_FAILED(ret) ){
			return ret;
		}
	}
    return ret;
}

/*
=begin
--- function#OpenSOAPXMLAttrGetValueWC(elm, type_name, value)
    Get Value of XML Attribute(WC)

    :Parameters
      :OpenSOAPXMLAttrPtr [in] ((|elm|))
        OpenSOAP XML Attribute.
      :const wchar_t * [in] ((|type_name|))
        Type Name 
	  :void * [out] ((|value|))
        Strage Buffer Pointer
    :Return value
      :int
	    Return Code Is ErrorCode
=end
*/
extern
int
OPENSOAP_API
OpenSOAPXMLAttrGetValueWC(OpenSOAPXMLAttrPtr /* [in] */ elm,
                         const wchar_t * /* [in] */ type_name,
                         void * /* [out] */ value) {
    int ret = OPENSOAP_YET_IMPLEMENTATION;

    return ret;
}

/*
=begin
--- function#OpenSOAPXMLAttrSetValueString(elm, type_name, value)
    Set Value of XML Attribute(OpenSOAPString)

    :Parameters
      :OpenSOAPXMLAttrPtr [in] ((|elm|))
        OpenSOAP XML Attribute.
      :OpenSOAPStringPtr [in] ((|type_name|))
        Type Name
      :void * [in] ((|value|))
        Strage Buffer Pointer
    :Return value
      :int
	    Return Code Is ErrorCode
=end
*/
extern
int
OPENSOAP_API
OpenSOAPXMLAttrSetValueString(OpenSOAPXMLAttrPtr /* [in] */ elm,
                         OpenSOAPStringPtr /* [in] */ type_name,
                              void * /* [in] */ value) {
    int ret = OPENSOAP_YET_IMPLEMENTATION;

    OpenSOAPSerializerFunc serializer = NULL;

    if( !elm ){
        ret = OPENSOAP_PARAMETER_BADVALUE;
        return ret;
    }

    ret = OpenSOAPStringCreate( &(elm->super.value) );
    if( OPENSOAP_FAILED(ret) ){
        return ret;
    }

    ret = OpenSOAPGetSerializer(type_name, &serializer);
    if (OPENSOAP_SUCCEEDED(ret)) {
        ret = serializer(value, ((OpenSOAPXMLNodePtr)elm)->value);
    } 
	
    return ret;
}

/*
=begin
--- function#OpenSOAPXMLAttrSetValueMB(elm, type_name, value)
    Set Value of XML Attribute(MB)

    :Parameters
      :OpenSOAPXMLAttrPtr [in] ((|elm|))
        OpenSOAP XML Attribute.
      :const char * [in] ((|type_name|))
        Type Name
      :void * [in] ((|value|))
        Strage Buffer Pointer
    :Return value
      :int
	    Return Code Is ErrorCode
=end
*/
extern
int
OPENSOAP_API
OpenSOAPXMLAttrSetValueMB(OpenSOAPXMLAttrPtr /* [in] */ elm,
                         const char * /* [in] */ type_name,
                          void * /* [in] */ value) {
    int ret = OPENSOAP_YET_IMPLEMENTATION;

    OpenSOAPSerializerFunc serializer = NULL;

    if( !elm ){
        ret = OPENSOAP_PARAMETER_BADVALUE;
        return ret;
    }

    ret = OpenSOAPStringCreate( &(elm->super.value) );
    if( OPENSOAP_FAILED(ret) ){
        return ret;
    }

    ret = OpenSOAPGetSerializerMB(type_name, &serializer);
    if (OPENSOAP_SUCCEEDED(ret)) {
        ret = serializer(value, ((OpenSOAPXMLNodePtr)elm)->value);
    } 
	
    return ret;
}

/*
=begin
--- function#OpenSOAPXMLAttrSetValueWC(elm, type_name, value)
    Set Value of XML Attribute(WC)

    :Parameters
      :OpenSOAPXMLAttrPtr [in] ((|elm|))
        OpenSOAP XML Attribute.
      :const wchar_t * [in] ((|type_name|))
        Type Name 
      :void * [in] ((|value|))
        Strage Buffer Pointer
    :Return value
      :int
	    Return Code Is ErrorCode
=end
*/
extern
int
OPENSOAP_API
OpenSOAPXMLAttrSetValueWC(OpenSOAPXMLAttrPtr /* [in] */ elm,
                         const wchar_t * /* [in] */ type_name,
                          void * /* [in] */ value) {
    OpenSOAPStringPtr type_name_str = NULL;
    int ret = OpenSOAPStringCreateWithWC(type_name, &type_name_str);

    if (OPENSOAP_SUCCEEDED(ret)) {
        ret = OpenSOAPXMLAttrSetValueString(elm, type_name_str, value);
        OpenSOAPStringRelease(type_name_str);
    }
    return ret;
}



/*
=begin
--- function#OpenSOAPXMLAttrGetName(attr, name)
    Getting of SOAP Attribute Name

    :Parameters
      :OpenSOAPXMLAttrPtr [in] ((|attr|))
        SOAP Attribute Pointer.
      :OpenSOAPStringPtr * [out] ((|name|))
        Result Name of SOAP Attribute.
    :Return Value
      :int
	    Return Code Is ErrorCode
=end
 */
extern
int
OPENSOAP_API
OpenSOAPXMLAttrGetName(OpenSOAPXMLAttrPtr /* [in] */ attr,
                       OpenSOAPStringPtr * /* [out] */ name) {
    int ret = OpenSOAPXMLNodeGetName((OpenSOAPXMLNodePtr)attr,
                                     name);

    return ret;
}

