/*-----------------------------------------------------------------------------
 * $RCSfile: XMLElm.c,v $
 *
 * See Copyright for the status of this software.
 *
 * The OpenSOAP Project
 * http://opensoap.jp/
 *-----------------------------------------------------------------------------
 */


#if (!defined(_WIN32)) || defined(_DEBUG)
static  const   char    CVS_ID[]        =       "$Id: XMLElm.c,v 1.51 2002/01/21 02:00:06 jin Exp $";
#endif  /* _DEBUG */

#include <string.h>

#include "XMLElm.h"
#include <OpenSOAP/Serializer.h>
#ifdef HAVE_CONFIG_H
# include "config.h"
#endif /* HAVE_CONFIG_H */

#if defined(HAVE_LIBXML2_XMLVERSION_H) 
#include <libxml2/tree.h>
#include <libxml2/parser.h>
#include <libxml2/xmlmemory.h>
#else
#include <libxml/tree.h>
#include <libxml/parser.h>
#include <libxml/xmlmemory.h>
#endif /* !HAVE_LIBXML2_XMLVERSION_H */


/*static
int
OpenSOAPXMLElmAddNodeWithLibxml( xmlDocPtr doc, xmlNodePtr node, OpenSOAPXMLElmPtr elm );
*/
/*static
int
OpenSOAPXMLElmConvertEncodingToUTF8Libxml(const char *name,
										  OpenSOAPByteArrayPtr utf8_name,
										  const unsigned char **utf8_name_beg );

*/

/*
=begin
= OpenSOAP XML Element class
=end
 */
extern
int
OpenSOAPXMLElmReleaseMembers(OpenSOAPXMLElmPtr /* [in, out] */ elm) {
    int ret = OpenSOAPXMLNodeReleaseMembers((OpenSOAPXMLNodePtr)elm);

    if (OPENSOAP_SUCCEEDED(ret)) {
		ret = OpenSOAPXMLNodeReleaseList((OpenSOAPXMLNodePtr)(elm->
															  xmlAttrList));
        if (OPENSOAP_SUCCEEDED(ret) && elm->definedXMLNsList) {
            ret = OpenSOAPXMLNamespaceReleaseList(elm->definedXMLNsList);
        }
    }

    return ret;
}

extern
int
OpenSOAPXMLElmRelease(OpenSOAPXMLElmPtr /* [in, out] */ elm) {
    int ret = OpenSOAPObjectRelease((OpenSOAPObjectPtr)elm);

    return ret;
}

static
int
OpenSOAPXMLElmFree(OpenSOAPObjectPtr /* [in, out] */ obj) {
    int ret = OpenSOAPXMLElmReleaseMembers((OpenSOAPXMLElmPtr)obj);

    if (OPENSOAP_SUCCEEDED(ret)) {
        free(obj);
    }

    return ret;
}

/*
=begin
--- function#OpenSOAPXMLElmInitialize(elm)
    Initialize XML Element Struct.

    :Parameters
      :OpenSOAPXMLElmPtr [in, out] ((|elm|))
        XML Element.
    :Return Value
      :int
	    Return Code Is ErrorCode
=end
 */
#if 1
extern
int
OpenSOAPXMLElmInitialize(OpenSOAPXMLElmPtr /* [in, out] */ elm,
                          OpenSOAPObjectFreeFunc /* [in] */ free_func) {

    int ret = OPENSOAP_PARAMETER_BADVALUE;

    if (elm) {
	elm->xmlAttrList = NULL;
	elm->definedXMLNsList = NULL;
        
        ret = OpenSOAPXMLNodeInitialize((OpenSOAPXMLNodePtr)elm,
                                        free_func ? free_func
                                        : OpenSOAPXMLElmFree);
    }
    
    return ret;
}
#else
extern
int
OpenSOAPXMLElmInitialize(OpenSOAPXMLElmPtr /* [in, out] */ elm){
	int ret = OPENSOAP_YET_IMPLEMENTATION;

	if( !elm ){
		ret = OPENSOAP_PARAMETER_BADVALUE;
		return ret;
	}

	ret = OpenSOAPXMLNodeInitialize( (OpenSOAPXMLNodePtr)elm );
	if( OPENSOAP_FAILED(ret) ){
		return ret;
	}

	elm->xmlAttrList = NULL;
	elm->definedXMLNsList = NULL;

	ret = OPENSOAP_NO_ERROR;

	return ret;
}
#endif

/*
=begin
--- function#OpenSOAPXMLElmCreate(attr)
    OpenSOAP XML Element instance create

    :Parameters
      :OpenSOAPXMLElmPtr * [out] ((|elm|))
        XML Element pointer
    :Return Values
      :int
	    Return Code Is ErrorCode
=end
 */
extern
int
OPENSOAP_API
OpenSOAPXMLElmCreate(OpenSOAPXMLElmPtr * /* [out] */ elm) {
    int ret = OPENSOAP_YET_IMPLEMENTATION;

	if(elm){
		*elm = malloc(sizeof(OpenSOAPXMLElm));

		if(*elm){
			ret = OpenSOAPXMLElmInitialize( *elm, NULL );
			if( OPENSOAP_FAILED(ret) ){
				return ret;
			}
		}
		else{
			ret = OPENSOAP_MEM_BADALLOC;
		}
	}
	else{
		ret = OPENSOAP_PARAMETER_BADVALUE;
	}

    return ret;
}

/*
=begin
--- function#OpenSOAPXMLElmSetNamespaceMB(elm, ns_uri, ns_prefix)
    Set Namespace of XML Element(MB)

    :Parameters
      :OpenSOAPXMLElmPtr [in, out] ((|elm|))
        XML Element.
      :const char * [in] ((|ns_uri|))
        Namespace URI
      :const char * [in] ((|ns_prefix|))
        Namespace Prefix
    :Return Value
      :int
	    Return Code Is ErrorCode
=end
 */
extern
int
OPENSOAP_API
OpenSOAPXMLElmSetNamespaceMB(OpenSOAPXMLElmPtr /* [in, out] */ elm,
                             const char * /* [in] */ ns_uri,
                             const char * /* [in] */ ns_prefix) {
    int ret = OPENSOAP_YET_IMPLEMENTATION;

	OpenSOAPXMLNamespacePtr ns = NULL;

	ret = OpenSOAPXMLElmSearchNamespaceMB( elm, ns_uri, ns_prefix, &ns );
	if( OPENSOAP_FAILED(ret) ){
		return ret;
	}

	/* there is a namespace defined by ns_uri and ns_prefix*/
	if( ns != NULL ){
		elm->super.thisNamespace = ns;
	}
	/* If there is no ns_uri and ns_prefix, make a new XMLNamespace,
	   input the new XMLNamespace to the definedXMLNsList,
	   input itself XMLElmPtr to definedElment */
	else{
		ret = OpenSOAPXMLNamespaceCreateMB(	ns_uri,
											ns_prefix,
											&(elm->super.thisNamespace));
		if( OPENSOAP_FAILED(ret) ){
			return ret;
		}

		if( !(elm->definedXMLNsList)){
			elm->definedXMLNsList = (OpenSOAPXMLNamespacePtr)(elm->super.thisNamespace);
		}
		else{
			ret = OpenSOAPXMLNodeInsertNext( (OpenSOAPXMLNodePtr)(elm->definedXMLNsList), 
											 (OpenSOAPXMLNodePtr)(elm->super.thisNamespace));
			if( OPENSOAP_FAILED(ret) ){
				return ret;
			}
		}

		elm->super.thisNamespace->definedElement = elm;
	}

	ret = OPENSOAP_NO_ERROR;
    return ret;
}

/*
=begin
--- function#OpenSOAPXMLElmSetNamespaceWC(elm, ns_uri, ns_prefix)
    Set Namespace of XML Element(WC)

    :Parameters
      :OpenSOAPXMLElmPtr [in, out] ((|elm|))
        XML Element.
      :const wchar_t * [in] ((|ns_uri|))
        Namespace URI
      :const wchar_t * [in] ((|ns_prefix|))
        Namespace Prefix

    :Return Value
      :int
	    Return Code Is ErrorCode
=end
 */
extern
int
OPENSOAP_API
OpenSOAPXMLElmSetNamespaceWC(OpenSOAPXMLElmPtr /* [in, out] */ elm,
                             const wchar_t * /* [in] */ ns_uri,
                             const wchar_t * /* [in] */ ns_prefix) {
    int ret = OPENSOAP_YET_IMPLEMENTATION;

    return ret;
}

/*
=begin
--- function#OpenSOAPXMLElmSearchNamespaceString(elm, ns_uri, ns_prefix, ns)
    Search Namespace of XML Element(OpenSOAPString).

    :Parameters
      :OpenSOAPXMLElmPtr [in] ((|elm|))
        XML Element.
      :OpenSOAPStringPtr [in] ((|ns_uri|))
        Namespace URI
      :OpenSOAPStringPtr [in] ((|ns_prefix|))
        Namespace Prefix
      :OpenSOAPXMLNamespacePtr * [out] ((|ns|))
        Namespace
    :Return Value
      :int
	    Return Code Is ErrorCode
	:Note
	  This Function Is Static Function
=end
 */
/* extern */
static
int
/* OPENSOAP_API */
OpenSOAPXMLElmSearchNamespaceString(OpenSOAPXMLElmPtr /* [in] */ elm,
                                    OpenSOAPStringPtr /* [in] */ ns_uri,
                                    OpenSOAPStringPtr /* [in] */ ns_prefix,
                                    OpenSOAPXMLNamespacePtr * /* [out] */ ns) {
    int ret = OPENSOAP_YET_IMPLEMENTATION;
#if 0    
    int compare_result;

    OpenSOAPXMLElmPtr next_parent;
    OpenSOAPXMLNamespacePtr next_tmp;

    if( !elm ){
        ret = OPENSOAP_PARAMETER_BADVALUE;
        return ret;
    }

    /* check if there is the ns_uri in itself definedXMLNsList */
    next_tmp = elm->definedXMLNsList;

    while( next_tmp != NULL 
           && OPENSOAP_SUCCEEDED(OpenSOAPStringCompareMB(next_tmp->uri, ns_uri, &compare_result))
           && compare_result != 0 ){
        next_tmp = next_tmp->next;
    }
    if( next_tmp != NULL ){
        elm->super.thisNamespace = next_tmp;
        ret = OPENSOAP_NO_ERROR;
        return ret;
    }

    /* If there is no ns_uri in itself definedXMLNsList,
       check if there is teh ns_uri in its parent's definedXMLNsList */
    next_parent = (OpenSOAPXMLElmPtr)(((OpenSOAPXMLNodePtr)(elm))->parent);

    while( next_parent != NULL ){
        next_tmp = next_parent->definedXMLNsList;
        while( next_tmp != NULL 
               && OPENSOAP_SUCCEEDED(OpenSOAPStringCompareMB(next_tmp->uri,
                                                             ns_uri,
                                                             &compare_result)) 
               && compare_result != 0 ){
            next_tmp = next_tmp->next;
        }
        if( next_tmp != NULL ){
            elm->super.thisNamespace = next_tmp;
            ret = OPENSOAP_NO_ERROR;
            return ret;
        }
        next_parent = (OpenSOAPXMLElmPtr)(next_parent->super.parent);
    }

    /* If there is no ns_uri, make a new XMLNamespace,
       input the new XMLNamespace to the definedXMLNsList,
       input itself XMLElmPtr to definedElment */
    ret = OpenSOAPXMLNamespaceCreateMB(	ns_uri,
                                        ns_prefix,
                                        &(elm->super.thisNamespace));
    if( OPENSOAP_FAILED(ret) ){
        return ret;
    }

    if( !(elm->definedXMLNsList)){
        elm->definedXMLNsList = (OpenSOAPXMLNamespacePtr)(elm->super.thisNamespace);
    }
    else{
        ret = OpenSOAPXMLNodeInsertNext( (OpenSOAPXMLNodePtr)(elm->definedXMLNsList), 
                                         (OpenSOAPXMLNodePtr)(elm->super.thisNamespace));
        if( OPENSOAP_FAILED(ret) ){
            return ret;
        }
    }

    elm->super.thisNamespace->definedElement = elm;

    ret = OPENSOAP_NO_ERROR;
#endif    
    return ret;
}

/*
=begin
--- function#OpenSOAPXMLElmSearchNamespaceMB(elm, ns_uri, ns_prefix, ns)
    Search Namespace of XML Element(MB).

    :Parameters
      :OpenSOAPXMLElmPtr [in] ((|elm|))
        XML Element.
      :const char * [in] ((|ns_uri|))
        Namespace URI
      :const char * [in] ((|ns_prefix|))
        Namespace Prefix
      :OpenSOAPXMLNamespacePtr * [out] ((|ns|))
        Namespace
    :Return Value
      :int
	    Return Code Is ErrorCode
=end
 */
extern
int
OPENSOAP_API
OpenSOAPXMLElmSearchNamespaceMB(OpenSOAPXMLElmPtr /* [in] */ elm,
                                const char * /* [in] */ ns_uri,
                                const char * /* [in] */ ns_prefix,
                                OpenSOAPXMLNamespacePtr * /* [out] */ ns) {
#if 1
    int ret = OPENSOAP_YET_IMPLEMENTATION;

	int compare_result_uri;
	int compare_result_prefix;

	OpenSOAPXMLElmPtr next_parent = NULL;
	OpenSOAPXMLNamespacePtr next_tmp = NULL;

	if( !elm || !ns ){
		ret = OPENSOAP_PARAMETER_BADVALUE;
		return ret;
	}

	/* check if there is the ns_uri and ns_prefix in itself definedXMLNsList */

	if( elm->definedXMLNsList != NULL ){
		next_tmp = elm->definedXMLNsList;
		while( next_tmp != NULL ){
			if( next_tmp->uri != NULL && ns_uri != NULL ) {
				ret = OpenSOAPStringCompareMB(next_tmp->uri, ns_uri, &compare_result_uri);
				if( OPENSOAP_FAILED(ret) ){
					return ret;
				}
			}
			else{
				if( next_tmp->uri == NULL && ns_uri == NULL ){
					compare_result_uri = 0;
				}
				else{
					compare_result_uri = 1;
				}
			}

			if( next_tmp->prefix != NULL && ns_prefix != NULL ) {
				ret = OpenSOAPStringCompareMB(next_tmp->prefix, ns_prefix, &compare_result_prefix);
				if( OPENSOAP_FAILED(ret) ){
					return ret;
				}
			}
			else{
				if( next_tmp->prefix == NULL && ns_prefix == NULL ){
					compare_result_prefix = 0;
				}
				else{
					compare_result_prefix = 1;
				}
			}

			if( compare_result_prefix == 0 ){
				if( compare_result_uri != 0 ){
					ret = OPENSOAP_PARAMETER_BADVALUE;
					return ret;
				}
				else{
					*ns = next_tmp;
					ret = OPENSOAP_NO_ERROR;
					return ret;
				}
			}
			else{
				next_tmp = next_tmp->next;
			}
		}
	}


	/* If there is no ns_uri and ns_prefix in itself definedXMLNsList,
	   check if there is the ns_uri and ns_prefix in its parent's definedXMLNsList */
	next_parent = (OpenSOAPXMLElmPtr)(((OpenSOAPXMLNodePtr)(elm))->parent);

	while( next_parent != NULL ){
		next_tmp = next_parent->definedXMLNsList;

		while( next_tmp != NULL ){
			if( next_tmp->uri != NULL && ns_uri != NULL ) {
				ret = OpenSOAPStringCompareMB( next_tmp->uri,
											   ns_uri,
											   &compare_result_uri );
				if( OPENSOAP_FAILED(ret) ){
					return ret;
				}
			}
			else{
				if( next_tmp->uri == NULL && ns_uri == NULL ){
					compare_result_uri = 0;
				}
				else{
					compare_result_uri = 0;
				}
			}

			if( compare_result_uri == 0 ){
				if( next_tmp->prefix != NULL && ns_prefix != NULL ) {
					ret = OpenSOAPStringCompareMB( next_tmp->prefix,
												   ns_prefix,
												   &compare_result_prefix );
					if( OPENSOAP_FAILED(ret) ){
						return ret;
					}
				}
				else{
					if( next_tmp->prefix == NULL && ns_prefix == NULL ){
						compare_result_prefix = 0;
					}
					else{
						compare_result_prefix = 1;
					}
				}

				if( compare_result_prefix == 0 ){
					*ns = next_tmp;
					ret = OPENSOAP_NO_ERROR;
					return ret;
				}
			}
			next_tmp = next_tmp->next;
		}

		next_parent = (OpenSOAPXMLElmPtr)(next_parent->super.parent);
	}

	/* there is no ns_uri and ns_prefix namespace */ 
	*ns = NULL;

    return ret;

#else
    OpenSOAPStringPtr ns_uri_str = NULL;
    int ret = OpenSOAPStringCreateWithMB(ns_uri, &ns_uri_str);
    if (OPENSOAP_SUCCEEDED(ret)) {
        OpenSOAPStringPtr ns_prefix_str = NULL;
        ret = OpenSOAPStringCreateWithMB(ns_prefix, &ns_prefix_str);
        if (OPENSOAP_SUCCEEDED(ret)) {
            ret = OpenSOAPXMLElmSearchNamespaceString(elm,
                                                      ns_uri_str,
                                                      ns_prefix_str,
                                                      ns);
            
            OpenSOAPStringRelease(ns_prefix_str);
        }
        OpenSOAPStringRelease(ns_uri_str);
    }
    
    return ret;
#endif
}


/*
=begin
--- function#OpenSOAPXMLElmSearchNamespaceWC(elm, ns_uri, ns_prefix, ns)
    Search Namespace of XML Element(WC).

    :Parameters
      :OpenSOAPXMLElmPtr [in, out] ((|elm|))
        XML Element.
      :const wchar_t * [in] ((|ns_uri|))
        Namespace URI
      :const wchar_t * [in] ((|ns_prefix|))
        Namespace Prefix
      :OpenSOAPXMLNamespacePtr * [out] ((|ns|))
        Namespace
    :Return Value
      :int
	    Return Code Is ErrorCode
=end
 */
extern
int
OPENSOAP_API
OpenSOAPXMLElmSearchNamespaceWC(OpenSOAPXMLElmPtr /* [in] */ elm,
                                const wchar_t * /* [in] */ ns_uri,
                                const wchar_t * /* [in] */ ns_prefix,
                                OpenSOAPXMLNamespacePtr * /* [out] */ ns) {
    OpenSOAPStringPtr ns_uri_str = NULL;
    int ret = OpenSOAPStringCreateWithWC(ns_uri, &ns_uri_str);
    if (OPENSOAP_SUCCEEDED(ret)) {
        OpenSOAPStringPtr ns_prefix_str = NULL;
        ret = OpenSOAPStringCreateWithWC(ns_prefix, &ns_prefix_str);
        if (OPENSOAP_SUCCEEDED(ret)) {
            ret = OpenSOAPXMLElmSearchNamespaceString(elm,
                                                      ns_uri_str,
                                                      ns_prefix_str,
                                                      ns);
            
            OpenSOAPStringRelease(ns_prefix_str);
        }
        OpenSOAPStringRelease(ns_uri_str);
    }
    
    return ret;
}


/*
=begin
--- function#OpenSOAPXMLElmGetNamespace(elm, ns)
    Get Namespace Pointer

    :Parameters
      :OpenSOAPXMLElmPtr [in, out] ((|elm|))
        XML Element
      :OpenSOAPXMLNamespacePtr * [out] ((|ns|))
        Namespace 
    :Return Value
      :int
	    Return Code Is ErrorCode
=end
 */
extern
int
OPENSOAP_API
OpenSOAPXMLElmGetNamespace(OpenSOAPXMLElmPtr /* [in, out] */ elm,
                           OpenSOAPXMLNamespacePtr * /* [out] */ ns) {
    int ret = OPENSOAP_YET_IMPLEMENTATION;

	if( !elm || !ns ){
		ret = OPENSOAP_PARAMETER_BADVALUE;
	}
	else{
		*ns = elm->super.thisNamespace;
		ret = OPENSOAP_NO_ERROR;
	}

    return ret;
}

/*
=begin
--- function#OpenSOAPXMLElmDefineNamespaceMB(elm, ns_uri, ns_prefix, ns)
    Define XML namespace.(MB)

    :Parameters
      :OpenSOAPXMLElmPtr [in, out] ((|elm|))
        XML Element.
      :const char * [in] ((|ns_uri|))
        Namespace URI
      :const char * [in] ((|ns_prefix|))
        Namespace Prefix
      :OpenSOAPXMLNamespacePtr * [out] ((|ns|))
        OpenSOAP XML Namespace.
    :Return Value
      :int
	    Return Code Is ErrorCode
=end
 */
extern
int
OPENSOAP_API
OpenSOAPXMLElmDefineNamespaceMB(OpenSOAPXMLElmPtr /* [in, out] */ elm,
                                const char * /* [in] */ ns_uri,
                                const char * /* [in] */ ns_prefix,
                                OpenSOAPXMLNamespacePtr * /* [out] */ ns) {
    int ret = OPENSOAP_YET_IMPLEMENTATION;

	OpenSOAPXMLNamespacePtr ns_next = NULL;

	int compare_result_uri;
	int compare_result_prefix;

	if( !elm || !ns ){
		ret = OPENSOAP_PARAMETER_BADVALUE;
		return ret;
	}

	if( elm->definedXMLNsList != NULL ){
		ns_next = elm->definedXMLNsList;
		while( ns_next != NULL ){
			if( ns_next->uri != NULL && ns_uri != NULL ){
				ret = OpenSOAPStringCompareMB(ns_next->uri, ns_uri, &compare_result_uri);
				if( OPENSOAP_FAILED(ret) ){
					return ret;
				}
			}
			else{
				if( ns_next->uri == NULL && ns_uri == NULL ){
					compare_result_uri = 0;
				}
				else{
					compare_result_uri = 1;
				}
			}

			if( ns_next->prefix != NULL && ns_prefix != NULL ){
				ret = OpenSOAPStringCompareMB(ns_next->prefix, ns_prefix, &compare_result_prefix);
				if( OPENSOAP_FAILED(ret) ){
					return ret;
				}
			}
			else{
				if( ns_next->prefix == NULL && ns_prefix == NULL ){
					compare_result_prefix = 0;
				}
				else{
					compare_result_prefix = 1;
				}
			}

			if( compare_result_prefix == 0 ){
				if( compare_result_uri != 0 ){
					ret = OPENSOAP_PARAMETER_BADVALUE;
					return ret;
				}
				else{
					*ns = ns_next;
					ret = OPENSOAP_NO_ERROR;
					return ret;
				}
			}
			else{
				ns_next = ns_next->next;
			}
		}
	}

    ret = OpenSOAPXMLNamespaceCreateMB(	ns_uri,
										ns_prefix,
										ns);
	if( OPENSOAP_FAILED(ret) ){
		return ret;
	}

	if( !(elm->definedXMLNsList)){
		elm->definedXMLNsList = *ns;
	}
	else{
		ret = OpenSOAPXMLNodeInsertNext( (OpenSOAPXMLNodePtr)(elm->definedXMLNsList), 
										 (OpenSOAPXMLNodePtr)(*ns));
		if( OPENSOAP_FAILED(ret) ){
			return ret;
		}
	}

	(*ns)->definedElement = elm;

	ret = OPENSOAP_NO_ERROR;

    return ret;
}

/*
=begin
--- function#OpenSOAPXMLElmDefineNamespaceWC(elm, ns_uri, ns_prefix, ns)
    Define XML namespace.(WC)

    :Parameters
      :OpenSOAPXMLElmPtr [in, out] ((|elm|))
        XML Element.
      :const wchar_t * [in] ((|ns_uri|))
        Namespace URI
      :const wchar_t * [in] ((|ns_prefix|))
        Namespace Prefix
      :OpenSOAPXMLNamespacePtr * [out] ((|ns|))
        OpenSOAP XML Namespace.
    :Return Value
      :int
	    Return Code Is ErrorCode
=end
 */
extern
int
OPENSOAP_API
OpenSOAPXMLElmDefineNamespaceWC(OpenSOAPXMLElmPtr /* [in, out] */ elm,
                                const wchar_t * /* [in] */ ns_uri,
                                const wchar_t * /* [in] */ ns_prefix,
                                OpenSOAPXMLNamespacePtr * /* [out] */ ns) {
    int ret = OPENSOAP_YET_IMPLEMENTATION;

    return ret;
}

/*
=begin
--- function#OpenSOAPXMLElmAddAttributeString(elm, attr_name, attr_type, attr_value, attr)
    Add Attribute(OpenSOAPString)

    :Parameters
      :OpenSOAPXMLElmPtr [in, out] ((|elm|))
        SOAP Element.
      :OpenSOAPStringPtr [in] ((|attr_name|))
        Attribute Name
      :OpenSOAPStringPtr [in] ((|attr_type|))
        Attribute Type
	  :void * [in] ((|attr_value|))
        Attribute Value
      :OpenSOAPXMLAttrPtr * [out] ((|attr|))
        OpenSOAP XML Attribute
    :Return Value
      :int
	    Return Code Is ErrorCode
=end
 */
extern
int
OPENSOAP_API
OpenSOAPXMLElmAddAttributeString(OpenSOAPXMLElmPtr /* [in, out] */ elm,
                             OpenSOAPStringPtr /* [in] */ attr_name,
                             OpenSOAPStringPtr /* [in] */ attr_type,
                             void * /* [in] */ attr_value,
                             OpenSOAPXMLAttrPtr * /* [out] */ attr) {

    int ret = OPENSOAP_YET_IMPLEMENTATION;

	OpenSOAPXMLAttrPtr next_attr;

	if( elm ){
		ret = OpenSOAPXMLAttrCreateString( attr_name, attr );
		if( OPENSOAP_FAILED(ret) ){
			return ret;
		}
		ret = OpenSOAPXMLAttrSetValueString( *attr, attr_type, attr_value );
		if( OPENSOAP_FAILED(ret) ){
			return ret;
		}

		(*attr)->definedElement = elm;

		next_attr = elm->xmlAttrList;
		if( !next_attr ){
			elm->xmlAttrList = *attr;
		}
		else{
			ret = OpenSOAPXMLNodeInsertNext( (OpenSOAPXMLNodePtr)next_attr,
											 (OpenSOAPXMLNodePtr)(*attr) );
		}
	}
	else{
		ret = OPENSOAP_PARAMETER_BADVALUE;
	}
    return ret;
}

/*
=begin
--- function#OpenSOAPXMLElmAddAttributeMB(elm, attr_name, attr_type, attr_value, attr)
    Add Attribute(MB)

    :Parameters
      :OpenSOAPXMLElmPtr [in, out] ((|elm|))
        SOAP Element.
      :const char * [in] ((|attr_name|))
        Attribute Name
      :const char * [in] ((|attr_type|))
        Attribute Type
      :void * [in] ((|attr_value|))
        Attribute Value
      :OpenSOAPXMLAttrPtr * [out] ((|attr|))
        OpenSOAP XML Attribute
    :Return Value
      :int
	    Return Code Is ErrorCode
=end
 */
extern
int
OPENSOAP_API
OpenSOAPXMLElmAddAttributeMB(OpenSOAPXMLElmPtr /* [in, out] */ elm,
                             const char * /* [in] */ attr_name,
                             const char * /* [in] */ attr_type,
                             void * /* [in] */ attr_value,
                             OpenSOAPXMLAttrPtr * /* [out] */ attr) {
    int ret = OPENSOAP_YET_IMPLEMENTATION;

	OpenSOAPXMLAttrPtr next_attr;

	if( elm ){
		ret = OpenSOAPXMLAttrCreateMB( attr_name, attr );
		if( OPENSOAP_FAILED(ret) ){
			return ret;
		}

		ret = OpenSOAPXMLAttrSetValueMB( *attr, attr_type, attr_value );
		if( OPENSOAP_FAILED(ret) ){
			return ret;
		}

		(*attr)->definedElement = elm;

		next_attr = elm->xmlAttrList;
		if( !next_attr ){
			elm->xmlAttrList = *attr;
		}
		else{
			ret = OpenSOAPXMLNodeInsertNext( (OpenSOAPXMLNodePtr)next_attr,
											 (OpenSOAPXMLNodePtr)(*attr) );
		}

	}
	else{
		ret = OPENSOAP_PARAMETER_BADVALUE;
	}

    return ret;
}


/*
=begin
--- function#OpenSOAPXMLElmAddAttributeWC(elm, attr_name, attr_type, attr_value, attr)
    Add Attribute(WC)

    :Parameters
      :OpenSOAPXMLElmPtr [in, out] ((|elm|))
        SOAP Element.
      :const wchar_t * [in] ((|attr_name|))
        Attribute Name
      :const wchar_t * [in] ((|attr_type|))
        Attribute Type
	  :void * [in] ((|attr_value|))
        Attribute Value
      :OpenSOAPXMLAttrPtr * [out] ((|attr|))
        OpenSOAP XML Attribute
    :Return Value
      :int
	    Return Code Is ErrorCode
=end
 */
extern
int
OPENSOAP_API
OpenSOAPXMLElmAddAttributeWC(OpenSOAPXMLElmPtr /* [in, out] */ elm,
                             const wchar_t * /* [in] */ attr_name,
                             const wchar_t * /* [in] */ attr_type,
                             void * /* [in] */ attr_value,
                             OpenSOAPXMLAttrPtr * /* [out] */ attr) {
	OpenSOAPStringPtr attr_name_str = NULL;
	int ret = OpenSOAPStringCreateWithWC(attr_name, &attr_name_str);

	if (OPENSOAP_SUCCEEDED(ret)) {
		OpenSOAPStringPtr attr_type_str = NULL;
		ret = OpenSOAPStringCreateWithWC(attr_type, &attr_type_str);
		if (OPENSOAP_SUCCEEDED(ret)) {
			ret = OpenSOAPXMLElmAddAttributeString(elm, attr_name_str, attr_type_str, attr_value, attr);
			OpenSOAPStringRelease(attr_type_str);
		}
		OpenSOAPStringRelease(attr_name_str);
	}

	return ret;
}

/*
=begin
--- function#OpenSOAPXMLElmGetAttributeString(elm, attr_name, attr)
    Get Attribute Pointer of Assignment Attribute Name(OpenSOAPString)

    :Parameters
      :OpenSOAPXMLElmPtr [in, out] ((|elm|))
        XML Element.
      :OpenSOAPStringPtr [in] ((|attr_name|))
        Attribute Name
      :OpenSOAPXMLAttrPtr * [out] ((|attr|))
        OpenSOAP XML Attribute
    :Return Value
      :int
	    Return Code Is ErrorCode
=end
 */
extern
int
OPENSOAP_API
OpenSOAPXMLElmGetAttributeString(OpenSOAPXMLElmPtr /* [in, out] */ elm,
                             OpenSOAPStringPtr /* [in] */ attr_name,
                             OpenSOAPXMLAttrPtr * /* [out] */ attr) {
    int ret = OPENSOAP_YET_IMPLEMENTATION;
	int compare_result;

	OpenSOAPXMLAttrPtr  next_attr;

	if( !elm || !attr ){
		ret = OPENSOAP_PARAMETER_BADVALUE;
		return ret;
	}

	next_attr = elm->xmlAttrList;

	while( next_attr != NULL ){
		ret = OpenSOAPStringCompare( next_attr->super.name, attr_name, &compare_result );
		if( OPENSOAP_FAILED(ret) ){
			return ret;
		}

		if( compare_result == 0 ){
			*attr = next_attr;
			ret = OPENSOAP_NO_ERROR;
			return ret;
		}
		next_attr = (OpenSOAPXMLAttrPtr)(next_attr->super.next);
	}

	/* Cannot Find */
	if( next_attr == NULL ){
		*attr = NULL;
		ret = OPENSOAP_NO_ERROR;
	}

    return ret;
}

/*
=begin
--- function#OpenSOAPXMLElmGetAttributeMB(elm, attr_name, attr)
    Get Attribute Pointer of Assignment Attribute Name(MB)

    :Parameters
      :OpenSOAPXMLElmPtr [in, out] ((|elm|))
        XML Element.
      :const char * [in] ((|attr_name|))
        Attribute Name
      :OpenSOAPXMLAttrPtr * [out] ((|attr|))
        OpenSOAP XML Attribute
    :Return Value
      :int
	    Return Code Is ErrorCode
=end
 */
extern
int
OPENSOAP_API
OpenSOAPXMLElmGetAttributeMB(OpenSOAPXMLElmPtr /* [in, out] */ elm,
                             const char * /* [in] */ attr_name,
                             OpenSOAPXMLAttrPtr * /* [out] */ attr) {
    int ret = OPENSOAP_YET_IMPLEMENTATION;
	int compare_result;

	OpenSOAPXMLAttrPtr  next_attr;

	if( !elm || !attr ){
		ret = OPENSOAP_PARAMETER_BADVALUE;
		return ret;
	}

	next_attr = elm->xmlAttrList;

	while( next_attr != NULL ){
		ret = OpenSOAPStringCompareMB( next_attr->super.name, attr_name, &compare_result );
		if( OPENSOAP_FAILED(ret) ){
			return ret;
		}

		if( compare_result == 0 ){
			*attr = next_attr;
			ret = OPENSOAP_NO_ERROR;
			return ret;
		}
		next_attr = (OpenSOAPXMLAttrPtr)(next_attr->super.next);
	}

	/* Cannot Find */
	if( next_attr == NULL ){
		*attr = NULL;
		ret = OPENSOAP_NO_ERROR;
	}

    return ret;
}

/*
=begin
--- function#OpenSOAPXMLElmGetAttributeWC(elm, attr_name, attr)
    Get Attribute Pointer of Assignment Attribute Name(WC)

    :Parameters
      :OpenSOAPXMLElmPtr [in, out] ((|elm|))
        XML Element.
      :const wchar_t * [in] ((|attr_name|))
        Attribute Name
      :OpenSOAPXMLAttrPtr * [out] ((|attr|))
        OpenSOAP XML Attribute
    :Return Value
      :int
	    Return Code Is ErrorCode
=end
 */
extern
int
OPENSOAP_API
OpenSOAPXMLElmGetAttributeWC(OpenSOAPXMLElmPtr /* [in, out] */ elm,
                             const wchar_t * /* [in] */ attr_name,
                             OpenSOAPXMLAttrPtr * /* [out] */ attr) {
	OpenSOAPStringPtr attr_name_str = NULL;
	int ret = OpenSOAPStringCreateWithWC(attr_name, &attr_name_str);

	if (OPENSOAP_SUCCEEDED(ret)) {
		ret = OpenSOAPXMLElmGetAttributeString(elm, attr_name_str, attr);
		OpenSOAPStringRelease(attr_name_str);
	}
	return ret;
}

/*
=begin
--- function#OpenSOAPXMLElmGetNextChild(elm, cld_elm)
    Get next child XML Element.

    :Parameters
      :OpenSOAPXMLElmPtr [in] ((|elm|))
        OpenSOAP XML Element.
      :OpenSOAPXMLElmPtr * [in, out] ((|cld_elm|))
        Next XML Element Pointer of Assignment XML Element
        But If *cld_elm as in NULL then Return To First XML Element Pointer 
    :Return Value value
      :int
	    Return Code Is ErrorCode
=end
*/
extern
int
OPENSOAP_API
OpenSOAPXMLElmGetNextChild(OpenSOAPXMLElmPtr /* [in] */ elm,
                           OpenSOAPXMLElmPtr * /* [in, out] */ cld_elm) {
    int ret = OPENSOAP_YET_IMPLEMENTATION;

	OpenSOAPXMLNodePtr  child_next;

	if( !elm ){
		ret = OPENSOAP_PARAMETER_BADVALUE;
		return ret;
	}

	if( !(*cld_elm) ){
		*cld_elm = (OpenSOAPXMLElmPtr)(elm->super.children);
		ret = OPENSOAP_NO_ERROR;
	}
	else{
		child_next = elm->super.children;
		while( child_next != NULL ){
			if( child_next == (OpenSOAPXMLNodePtr)(*cld_elm) ){
				*cld_elm = (OpenSOAPXMLElmPtr)((*cld_elm)->super.next);
				ret = OPENSOAP_NO_ERROR;
				return ret;
			}
			child_next = child_next->next;
		}
	}
    return ret;
}

/*
=begin
--- function#OpenSOAPXMLElmAddChildString(elm, cld_name, cld_elm)
    Add Child XML Element(OpenSOAPString)

    :Parameters
      :OpenSOAPXMLElmPtr [in] ((|elm|))
        OpenSOAP XML Element.
      :OpenSOAPStringPtr [in] ((|cld_name|))
        Add Child Element Name
      :OpenSOAPXMLElmPtr * [out] ((|cld_elm|))
        Add Child Element Pointer
    :Return Value value
      :int
	    Return Code Is ErrorCode
=end
*/
extern
int
OPENSOAP_API
OpenSOAPXMLElmAddChildString(OpenSOAPXMLElmPtr /* [in] */ elm,
                           OpenSOAPStringPtr /* [in] */ cld_name,
                           OpenSOAPXMLElmPtr * /* [out] */ cld_elm) {
    int ret = OPENSOAP_YET_IMPLEMENTATION;

	if( !elm || !cld_elm ){
		ret = OPENSOAP_PARAMETER_BADVALUE;
		return ret;
	}

	if( !*cld_elm ){
		ret = OpenSOAPXMLElmCreate( cld_elm );
		if( OPENSOAP_FAILED(ret) ){
			return ret;
		}
	}

	ret = OpenSOAPStringCreate( &((*cld_elm)->super.name) );
	if( OPENSOAP_FAILED(ret) ){
		return ret;
	}

    ret = OpenSOAPStringDuplicate((*cld_elm)->super.name, &cld_name);
	if(ret == OPENSOAP_PARAMETER_BADVALUE){
		return ret;
	}

	(*cld_elm)->super.parent = (OpenSOAPXMLNodePtr)elm;

	if( !elm->super.children ){
		elm->super.children = (OpenSOAPXMLNodePtr)(*cld_elm);
	}
	else{
		ret = OpenSOAPXMLNodeInsertNext( (OpenSOAPXMLNodePtr)(elm->super.children),
										 (OpenSOAPXMLNodePtr)(*cld_elm) );
		if( OPENSOAP_FAILED(ret) ){
			return ret;
		}
	}

	ret = OPENSOAP_NO_ERROR;
    return ret;
}

/*
=begin
--- function#OpenSOAPXMLElmAddChildMB(elm, cld_name, cld_elm)
    Add Child XML Element(MB)

    :Parameters
      :OpenSOAPXMLElmPtr [in] ((|elm|))
        OpenSOAP XML Element.
      :cosnt char * [in] ((|cld_name|))
        Add Child Element Name
      :OpenSOAPXMLElmPtr * [out] ((|cld_elm|))
        Add Child Element Pointer
    :Return Value value
      :int
	    Return Code Is ErrorCode
=end
*/
extern
int
OPENSOAP_API
OpenSOAPXMLElmAddChildMB(OpenSOAPXMLElmPtr /* [in] */ elm,
                           const char * /* [in] */ cld_name,
                           OpenSOAPXMLElmPtr * /* [out] */ cld_elm) {
    int ret = OPENSOAP_YET_IMPLEMENTATION;

	if( !elm || !cld_elm ){
		ret = OPENSOAP_PARAMETER_BADVALUE;
		return ret;
	}

	if( !*cld_elm ){
		ret = OpenSOAPXMLElmCreate( cld_elm );
		if( OPENSOAP_FAILED(ret) ){
			return ret;
		}
	}

	ret = OpenSOAPStringCreate( &((*cld_elm)->super.name) );
	if( OPENSOAP_FAILED(ret) ){
		return ret;
	}

    ret = OpenSOAPStringSetStringMB((*cld_elm)->super.name, cld_name);
	if(ret == OPENSOAP_PARAMETER_BADVALUE){
		return ret;
	}

	(*cld_elm)->super.parent = (OpenSOAPXMLNodePtr)elm;

	if( !elm->super.children ){
		elm->super.children = (OpenSOAPXMLNodePtr)(*cld_elm);
	}
	else{
		ret = OpenSOAPXMLNodeInsertNext( (OpenSOAPXMLNodePtr)(elm->super.children),
										 (OpenSOAPXMLNodePtr)(*cld_elm) );
		if( OPENSOAP_FAILED(ret) ){
			return ret;
		}
	}

	ret = OPENSOAP_NO_ERROR;
    return ret;
}

/*
=begin
--- function#OpenSOAPXMLElmAddChildWC(elm, cld_name, cld_elm)
    Add Child XML Element(WC)

    :Parameters
      :OpenSOAPXMLElmPtr [in] ((|elm|))
        OpenSOAP XML Element.
      :cosnt wchar_t * [in] ((|cld_name|))
        Add Child Element Name
      :OpenSOAPXMLElmPtr * [out] ((|cld_elm|))
        Add Child Element Pointer
    :Return Value
      :int
	    Return Code Is ErrorCode
=end
*/
extern
int
OPENSOAP_API
OpenSOAPXMLElmAddChildWC(OpenSOAPXMLElmPtr /* [in] */ elm,
                           const wchar_t * /* [in] */ cld_name,
                           OpenSOAPXMLElmPtr * /* [out] */ cld_elm) {
	OpenSOAPStringPtr cld_name_str = NULL;
	int ret = OpenSOAPStringCreateWithWC(cld_name, &cld_name_str);

	if (OPENSOAP_SUCCEEDED(ret)) {
		ret = OpenSOAPXMLElmAddChildString(elm, cld_name_str, cld_elm);
		OpenSOAPStringRelease(cld_name_str);
	}
	return ret;
}

/*
=begin
--- function#OpenSOAPXMLElmGetChildString(elm, cld_name, cld_elm)
    Get Child XML Element(OpenSOAPString)

    :Parameters
      :OpenSOAPXMLElmPtr [in] ((|elm|))
        OpenSOAP XML Element.
      :OpenSOAPStringPtr [in] ((|cld_name|))
        Assignment Element Name 
      :OpenSOAPXMLElmPtr * [out] ((|cld_elm|))
        XML Element Pointer of Assignment Element Name 
    :Return Value
      :int
	    Return Code Is ErrorCode
=end
*/
extern
int
OPENSOAP_API
OpenSOAPXMLElmGetChildString(OpenSOAPXMLElmPtr /* [in] */ elm,
                           OpenSOAPStringPtr /* [in] */ cld_name,
                           OpenSOAPXMLElmPtr * /* [out] */ cld_elm) {
    int ret = OPENSOAP_YET_IMPLEMENTATION;
	int compare_result;

	OpenSOAPXMLNodePtr  child;

	if( !elm || !cld_elm ){
		ret = OPENSOAP_PARAMETER_BADVALUE;
		return ret;
	}

	child = elm->super.children;

	while( child != NULL ){
		ret = OpenSOAPStringCompare( child->name, cld_name, &compare_result );
		if( OPENSOAP_FAILED(ret) ){
			return ret;
		}

		if( compare_result == 0 ){
			*cld_elm = ( OpenSOAPXMLElmPtr ) child;
			ret = OPENSOAP_NO_ERROR;
			return ret;
		}
		child = child->next;
	}

	// Cannot Find
	if( child == NULL ){
		*cld_elm = NULL;
		ret = OPENSOAP_NO_ERROR;
	}

    return ret;
}

/*
=begin
--- function#OpenSOAPXMLElmGetChildMB(elm, cld_name, cld_elm)
    Get Child XML Element(MB)

    :Parameters
      :OpenSOAPXMLElmPtr [in] ((|elm|))
        OpenSOAP XML Element.
      :cosnt char * [in] ((|cld_name|))
        Assignment Element Name 
      :OpenSOAPXMLElmPtr * [out] ((|cld_elm|))
        XML Element Pointer of Assignment Element Name 
    :Return Value
      :int
	    Return Code Is ErrorCode
=end
*/
extern
int
OPENSOAP_API
OpenSOAPXMLElmGetChildMB(OpenSOAPXMLElmPtr /* [in] */ elm,
                           const char * /* [in] */ cld_name,
                           OpenSOAPXMLElmPtr * /* [out] */ cld_elm) {
    int ret = OPENSOAP_YET_IMPLEMENTATION;
	int compare_result;

	OpenSOAPXMLNodePtr  child;

	if( !elm || !cld_elm ){
		ret = OPENSOAP_PARAMETER_BADVALUE;
		return ret;
	}

	child = elm->super.children;

	while( child != NULL ){
		ret = OpenSOAPStringCompareMB( child->name, cld_name, &compare_result );
		if( OPENSOAP_FAILED(ret) ){
			return ret;
		}

		if( compare_result == 0 ){
			*cld_elm = ( OpenSOAPXMLElmPtr ) child;
			ret = OPENSOAP_NO_ERROR;
			return ret;
		}
		child = child->next;
	}

	// Cannot Find
	if( child == NULL ){
		*cld_elm = NULL;
		ret = OPENSOAP_NO_ERROR;
	}

    return ret;
}

/*
=begin
--- function#OpenSOAPXMLElmGetChildWC(elm, cld_name, cld_elm)
    Get Child XML Element(WC)

    :Parameters
      :OpenSOAPXMLElmPtr [in] ((|elm|))
        OpenSOAP XML Element.
      :cosnt wchar_t * [in] ((|cld_name|))
        Assignment Element Name 
      :OpenSOAPXMLElmPtr * [out] ((|cld_elm|))
        XML Element Pointer of Assignment Element Name 
    :Return Value
      :int
	    Return Code Is ErrorCode
=end
*/
extern
int
OPENSOAP_API
OpenSOAPXMLElmGetChildWC(OpenSOAPXMLElmPtr /* [in] */ elm,
                           const wchar_t * /* [in] */ cld_name,
                           OpenSOAPXMLElmPtr * /* [out] */ cld_elm) {
	OpenSOAPStringPtr cld_name_str = NULL;
	int ret = OpenSOAPStringCreateWithWC(cld_name, &cld_name_str);

	if (OPENSOAP_SUCCEEDED(ret)) {
		ret = OpenSOAPXMLElmAddChildString(elm, cld_name_str, cld_elm);
		OpenSOAPStringRelease(cld_name_str);
	}
	return ret;
}

/*
=begin
--- function#OpenSOAPXMLElmGetValueString(elm, type_name, value)
    Get Value of XML Element(OpenSOAPString)

    :Parameters
      :OpenSOAPXMLElmPtr [in] ((|elm|))
        OpenSOAP XML Element.
      :OpenSOAPString [in] ((|type_name|))
        Type Name
      :void * [out] ((|value|))
        Setting Buffer Pointer
    :Return Value
      :int
	    Return Code Is ErrorCode
=end
*/
extern
int
OPENSOAP_API
OpenSOAPXMLElmGetValueString(OpenSOAPXMLElmPtr /* [in] */ elm,
                         OpenSOAPStringPtr /* [in] */ type_name,
                         void * /* [out] */ value) {
    int ret = OPENSOAP_YET_IMPLEMENTATION;

	OpenSOAPDeserializerFunc deserializer = NULL;

	if( !elm ){
		ret = OPENSOAP_PARAMETER_BADVALUE;
		return ret;
	}

	ret = OpenSOAPGetDeserializer(type_name, &deserializer);
	if (OPENSOAP_SUCCEEDED(ret)) {
		ret = deserializer(((OpenSOAPXMLNodePtr)elm)->value, value);
		if( OPENSOAP_FAILED(ret) ){
			return ret;
		}
	}
	return ret;
}

/*
=begin
--- function#OpenSOAPXMLElmGetValueMB(elm, type_name, value)
    Get Value of XML Element(MB)

    :Parameters
      :OpenSOAPXMLElmPtr [in] ((|elm|))
        OpenSOAP XML Element.
      :const char * [in] ((|type_name|))
        Type Name
      :void * [out] ((|value|))
        Setting Buffer Pointer
    :Return Value
      :int
	    Return Code Is ErrorCode
=end
*/
extern
int
OPENSOAP_API
OpenSOAPXMLElmGetValueMB(OpenSOAPXMLElmPtr /* [in] */ elm,
                         const char * /* [in] */ type_name,
                         void * /* [out] */ value) {
    int ret = OPENSOAP_YET_IMPLEMENTATION;

	OpenSOAPDeserializerFunc deserializer = NULL;

	if( !elm ){
		ret = OPENSOAP_PARAMETER_BADVALUE;
		return ret;
	}

	ret = OpenSOAPGetDeserializerMB(type_name, &deserializer);
	if (OPENSOAP_SUCCEEDED(ret)) {
		ret = deserializer(((OpenSOAPXMLNodePtr)elm)->value, value);
		if( OPENSOAP_FAILED(ret) ){
			return ret;
		}
	}
	return ret;
}

/*
=begin
--- function#OpenSOAPXMLElmGetValueWC(elm, type_name, value)
    Get Value of XML Element(WC)

    :Parameters
      :OpenSOAPXMLElmPtr [in] ((|elm|))
        OpenSOAP XML Element.
      :const wchar_t * [in] ((|type_name|))
        Type Name
      :void * [out] ((|value|))
        Setting Buffer Pointer
    :Return Value
      :int
	    Return Code Is ErrorCode
=end
*/
extern
int
OPENSOAP_API
OpenSOAPXMLElmGetValueWC(OpenSOAPXMLElmPtr /* [in] */ elm,
                         const wchar_t * /* [in] */ type_name,
                         void * /* [out] */ value) {
	OpenSOAPStringPtr type_name_str = NULL;
	int ret = OpenSOAPStringCreateWithWC(type_name, &type_name_str);

	if (OPENSOAP_SUCCEEDED(ret)) {
		ret = OpenSOAPXMLElmGetValueString(elm, type_name_str, value);
		OpenSOAPStringRelease(type_name_str);
	}
	return ret;
}

/*
=begin
--- function#OpenSOAPXMLElmSetValueString(elm, type_name, value)
    Setting Value of XML Element(OpenSOAPString)

    :Parameters
      :OpenSOAPXMLElmPtr [in] ((|elm|))
        OpenSOAP XML Element.
      :OpenSOAPStringPtr [in] ((|type_name|))
        Type Name
      :void * [in] ((|value|))
        Setting Buffer Pointer
    :Return Value
      :int
	    Return Code Is ErrorCode
=end
*/
extern
int
OPENSOAP_API
OpenSOAPXMLElmSetValueString(OpenSOAPXMLElmPtr /* [in] */ elm,
                         OpenSOAPStringPtr /* [in] */ type_name,
                         void * /* [in] */ value) {
    int ret = OPENSOAP_YET_IMPLEMENTATION;

	OpenSOAPSerializerFunc serializer = NULL;

	if( !elm ){
		ret = OPENSOAP_PARAMETER_BADVALUE;
		return ret;
	}

	ret = OpenSOAPStringCreate( &(elm->super.value) );
	if( OPENSOAP_FAILED(ret) ){
		return ret;
	}

	ret = OpenSOAPGetSerializer(type_name, &serializer);
	if (OPENSOAP_SUCCEEDED(ret)) {
		ret = serializer(value, ((OpenSOAPXMLNodePtr)elm)->value);
	}

	return ret;
}

/*
=begin
--- function#OpenSOAPXMLElmSetValueMB(elm, type_name, value)
    Setting Value of XML Element(MB)

    :Parameters
      :OpenSOAPXMLElmPtr [in] ((|elm|))
        OpenSOAP XML Element.
      :const char * [in] ((|type_name|))
        Type Name
      :void * [in] ((|value|))
        Setting Buffer Pointer
    :Return Value
      :int
	    Return Code Is ErrorCode
=end
*/
extern
int
OPENSOAP_API
OpenSOAPXMLElmSetValueMB(OpenSOAPXMLElmPtr /* [in] */ elm,
                         const char * /* [in] */ type_name,
                         void * /* [in] */ value) {
    int ret = OPENSOAP_YET_IMPLEMENTATION;

	OpenSOAPSerializerFunc serializer = NULL;

	if( !elm ){
		ret = OPENSOAP_PARAMETER_BADVALUE;
		return ret;
	}

	ret = OpenSOAPStringCreate( &(elm->super.value) );
	if( OPENSOAP_FAILED(ret) ){
		return ret;
	}

	ret = OpenSOAPGetSerializerMB(type_name, &serializer);
	if (OPENSOAP_SUCCEEDED(ret)) {
		ret = serializer(value, ((OpenSOAPXMLNodePtr)elm)->value);
	}

	return ret;
}

/*
=begin
--- function#OpenSOAPXMLElmSetValueWC(elm, type_name, value)
    Setting Value of XML Element(WC)

    :Parameters
      :OpenSOAPXMLElmPtr [in] ((|elm|))
        OpenSOAP XML Element.
      :const wchar_t * [in] ((|type_name|))
        Type Name
      :void * [in] ((|value|))
        Setting Buffer Pointer
    :Return Value
      :int
	    Return Code Is ErrorCode
=end
*/
extern
int
OPENSOAP_API
OpenSOAPXMLElmSetValueWC(OpenSOAPXMLElmPtr /* [in] */ elm,
                         const wchar_t * /* [in] */ type_name,
                         void * /* [in] */ value) {
	OpenSOAPStringPtr type_name_str = NULL;
	int ret = OpenSOAPStringCreateWithWC(type_name, &type_name_str);

	if (OPENSOAP_SUCCEEDED(ret)) {
		ret = OpenSOAPXMLElmSetValueString(elm, type_name_str, value);
		OpenSOAPStringRelease(type_name_str);
	}
	return ret;
}

/*
=begin
--- function#OpenSOAPXMLElmGetCharEncodingString(elm, ch_enc, b_ary)
    XML Element character encoding output

    :Parameters
      :OpenSOAPXMLElmPtr [in] ((|elm|))
        OpenSOAP XML Element.
      :const char  * [in] ((|ch_enc|))
        character encoding. i.e. "EUC-JP", "SHIFT-JIS", "UTF-8"
      :OpenSOAPByteArrayPtr [out] ((|b_ary|))
        Result Buffer        
    :Return Value
      :int
	    Return Code Is ErrorCode
=end
*/
extern
int
OPENSOAP_API
OpenSOAPXMLElmGetCharEncodingString(OpenSOAPXMLElmPtr /* [in] */ elm,
                                    const char * /* [in] */ ch_enc,
                                    OpenSOAPByteArrayPtr /* [out] */ b_ary) {
    int ret = OPENSOAP_PARAMETER_BADVALUE;

    if (elm) {

	xmlDocPtr	doc = NULL;
	xmlNsPtr	elmNs = NULL;
	xmlAttrPtr	elmAttr = NULL;

	xmlNodePtr	elmNode = NULL;

	OpenSOAPXMLAttrPtr next_attr = NULL;
	OpenSOAPXMLElmPtr  next_elm = NULL;

	xmlChar *xml_text = NULL;
	size_t xml_text_len = 0;
	OpenSOAPByteArrayPtr tmp_b_ary = NULL;

	unsigned char *xml_text_noProcLine;
	size_t xml_text_noProcLine_sz = 0;
	unsigned int i;
	unsigned int no_procLine = 0;

	char *name_elm = NULL;
	size_t len_name_elm = 0;
	OpenSOAPByteArrayPtr utf8_name_elm = NULL;
	const unsigned char *utf8_name_elm_beg = NULL;

	char *value_elm = NULL;
	size_t len_value_elm = 0;
	OpenSOAPByteArrayPtr utf8_value_elm = NULL;
	const unsigned char *utf8_value_elm_beg = NULL;

	char *elm_ns_uri = NULL;
	size_t elm_ns_uri_len = 0;
	OpenSOAPByteArrayPtr utf8_elm_ns_uri = NULL;
	const unsigned char *utf8_elm_ns_uri_beg = NULL;

	char *elm_ns_prefix = NULL;
	size_t elm_ns_prefix_len = 0;
	OpenSOAPByteArrayPtr utf8_elm_ns_prefix = NULL;
	const unsigned char *utf8_elm_ns_prefix_beg = NULL;

	char *attr_name = NULL;
	size_t attr_name_len = 0;
	OpenSOAPByteArrayPtr utf8_attr_name = NULL;
	const unsigned char *utf8_attr_name_beg = NULL;

	char *attr_value = NULL;
	size_t attr_value_len = 0;
	OpenSOAPByteArrayPtr utf8_attr_value = NULL;
	const unsigned char *utf8_attr_value_beg = NULL;


	/* make a new xmlDoc */ 
	doc = xmlNewDoc( "1.0" );
	next_elm = elm;  /* detail element */

	/* make detail element node */
	if( next_elm->super.value != NULL ){
		ret = OpenSOAPStringGetLengthMB(next_elm->super.name, &len_name_elm);
		if( OPENSOAP_FAILED(ret) ){
			return ret;
		}
		name_elm = malloc( (len_name_elm+1)*sizeof(char) );
		if( !name_elm ){
			ret = OPENSOAP_MEM_BADALLOC;
			return ret;
		}
		ret = OpenSOAPStringGetStringMB(next_elm->super.name, &len_name_elm, name_elm);
		if( OPENSOAP_FAILED(ret) ){
			return ret;
		}

		ret = OpenSOAPStringGetLengthMB(next_elm->super.value, &len_value_elm);
		if( OPENSOAP_FAILED(ret) ){
			return ret;
		}
		value_elm = malloc( (len_value_elm+1)*sizeof(char) );
		if( !value_elm ){
			ret = OPENSOAP_MEM_BADALLOC;
			return ret;
		}
		ret = OpenSOAPStringGetStringMB(next_elm->super.value, &len_value_elm, value_elm);
		if( OPENSOAP_FAILED(ret) ){
			return ret;
		}

		ret = OpenSOAPByteArrayCreate( &utf8_name_elm );
		if( OPENSOAP_FAILED(ret) ){
			return ret;
		}
		ret = OpenSOAPXMLElmConvertEncodingToUTF8Libxml( name_elm,
														 utf8_name_elm,
														 &utf8_name_elm_beg );
		if( OPENSOAP_FAILED(ret) ){
			return ret;
		}
		ret = OpenSOAPByteArrayCreate( &utf8_value_elm );
		if( OPENSOAP_FAILED(ret) ){
			return ret;
		}
		ret = OpenSOAPXMLElmConvertEncodingToUTF8Libxml( value_elm,
														 utf8_value_elm,
														 &utf8_value_elm_beg );
		if( OPENSOAP_FAILED(ret) ){
			return ret;
		}

		doc->children = xmlNewDocNode( doc, 
									   NULL, 
									   utf8_name_elm_beg,
									   utf8_value_elm_beg );
		elmNode = doc->children;

		free( name_elm );
		free( value_elm );
		ret = OpenSOAPByteArrayRelease( utf8_name_elm );
		if( OPENSOAP_FAILED(ret) ){
			return ret;
		}
		ret = OpenSOAPByteArrayRelease( utf8_value_elm );
		if( OPENSOAP_FAILED(ret) ){
			return ret;
		}
	}
	else{
		ret = OpenSOAPStringGetLengthMB(next_elm->super.name, &len_name_elm);
		if( OPENSOAP_FAILED(ret) ){
			return ret;
		}
		name_elm = malloc( (len_name_elm+1)*sizeof(char) );
		if( !name_elm ){
			ret = OPENSOAP_MEM_BADALLOC;
			return ret;
		}
		ret = OpenSOAPStringGetStringMB(next_elm->super.name, &len_name_elm, name_elm);
		if( OPENSOAP_FAILED(ret) ){
			return ret;
		}
		ret = OpenSOAPByteArrayCreate( &utf8_name_elm );
		if( OPENSOAP_FAILED(ret) ){
			return ret;
		}
		ret = OpenSOAPXMLElmConvertEncodingToUTF8Libxml( name_elm,
														 utf8_name_elm,
														 &utf8_name_elm_beg );
		if( OPENSOAP_FAILED(ret) ){
			return ret;
		}

		doc->children = xmlNewDocNode( doc, 
									   NULL, 
									   utf8_name_elm_beg,
									   NULL );
		elmNode = doc->children;

		free( name_elm );
		ret = OpenSOAPByteArrayRelease( utf8_name_elm );
		if( OPENSOAP_FAILED(ret) ){
			return ret;
		}
	}
	
	/* add root node's namespace */
	ret = OpenSOAPXMLElmAddNsInEncodingString( doc, next_elm, elmNode );
	if( OPENSOAP_FAILED(ret) ) {
		return ret;
	}

	/* add root node's attributes */
	ret = OpenSOAPXMLElmAddAttrInEncodingString( doc, next_elm, elmNode );
	if( OPENSOAP_FAILED(ret) ) {
		return ret;
	}

	/* add children node */
	ret =  OpenSOAPXMLElmAddNodeWithLibxml( doc, elmNode, next_elm );
	if( OPENSOAP_FAILED(ret) ){
		return ret;
	}

	/* change xmlDoc to text */
	xmlDocDumpMemory(doc, &xml_text, &xml_text_len);

	/* delete the processing line <?xml ... ?> */
	xml_text_noProcLine = malloc( xml_text_len * sizeof(unsigned char) );
	if( xml_text_noProcLine == NULL ){
		ret = OPENSOAP_MEM_BADALLOC;
		return ret;
	}

	for( i=2; i<xml_text_len; i++ ){
		if( xml_text[i-2] == '?' && xml_text[i-1] == '>'){
			no_procLine = i;
			break;
		}
	}

	if( no_procLine > 2 ){
		for( i=no_procLine; i<xml_text_len; i++ ) {
			if( xml_text[i] != '\n' && xml_text[i] != '\r' ){
				xml_text_noProcLine[xml_text_noProcLine_sz] = xml_text[i];
				xml_text_noProcLine_sz++;
			}
		}
	}

	ret = OpenSOAPByteArrayCreateWithData(xml_text_noProcLine, xml_text_noProcLine_sz, &tmp_b_ary);
	if (OPENSOAP_SUCCEEDED(ret)) {
		ret = OpenSOAPStringConvertCharEncoding("UTF-8", tmp_b_ary, ch_enc, b_ary);

	   OpenSOAPByteArrayRelease(tmp_b_ary);
	}
	xmlFree(xml_text);

	/* free memory */
	xmlFreeDoc( doc );
	}

    return ret;
}



/*
=begin
--- function#OpenSOAPXMLElmAddNodeWithLibxml( doc, node, elm )
    create libxml structure from opensoap structure 
    
    :Parameters
      :xmlDocPtr [in ] ((|doc|))
        Libxml Document Pointer
      :xmlNodePtr [in, out] ((|node|))
        Libxml Node Pointer
      :OpenSOAPXMLElmPtr [in] ((|elm|))
        OpenSOAP Element Pointer
    :Return value
      :type int
        error code.
      
=end
*/

int
OpenSOAPXMLElmAddNodeWithLibxml( xmlDocPtr doc, xmlNodePtr node, OpenSOAPXMLElmPtr elm ){
	int ret = OPENSOAP_YET_IMPLEMENTATION;

	OpenSOAPXMLElmPtr	elm_next = NULL;
	OpenSOAPXMLElmPtr	elm_child = NULL;
	OpenSOAPXMLAttrPtr	attr_next = NULL;
	OpenSOAPXMLNamespacePtr ns_next = NULL;

	xmlNodePtr	parentNode = NULL;
	xmlNodePtr	nextNode = NULL;
	xmlNodePtr	childNode = NULL;
	xmlNsPtr	childNs = NULL;
	xmlNsPtr	searchNs = NULL;
	xmlAttrPtr  childAttr = NULL;

	char *node_name = NULL;
	size_t node_name_len = 0;
	OpenSOAPByteArrayPtr utf8_node_name = NULL;
	const unsigned char *utf8_node_name_beg = NULL;

	char *node_value = NULL;
	size_t node_value_len = 0;
	OpenSOAPByteArrayPtr utf8_node_value = NULL;
	const unsigned char *utf8_node_value_beg = NULL;

	char *node_ns_uri = NULL;
	size_t node_ns_uri_len = 0;
	OpenSOAPByteArrayPtr utf8_node_ns_uri = NULL;
	const unsigned char *utf8_node_ns_uri_beg = NULL;

	char *node_ns_prefix = NULL;
	size_t node_ns_prefix_len = 0;
	OpenSOAPByteArrayPtr utf8_node_ns_prefix = NULL;
	const unsigned char *utf8_node_ns_prefix_beg = NULL;

	char *node_attr_name = NULL;
	size_t node_attr_name_len = 0;
	OpenSOAPByteArrayPtr utf8_node_attr_name = NULL;
	const unsigned char *utf8_node_attr_name_beg = NULL;

	char *node_attr_value = NULL;
	size_t node_attr_value_len = 0;
	OpenSOAPByteArrayPtr utf8_node_attr_value = NULL;
	const unsigned char *utf8_node_attr_value_beg = NULL;


	if( elm == NULL || node == NULL ){
		ret = OPENSOAP_PARAMETER_BADVALUE;
		return ret;
	}

	elm_child = (OpenSOAPXMLElmPtr)((OpenSOAPXMLNodePtr)elm)->children;
	parentNode = node;

	while( elm_child != NULL ){
		if( elm_child->super.value != NULL ){
			ret = OpenSOAPStringGetLengthMB(elm_child->super.name, &node_name_len);
			if( OPENSOAP_FAILED(ret) ){
				return ret;
			}
			node_name = malloc( (node_name_len+1)*sizeof(char) );
			if( !node_name ){
				ret = OPENSOAP_MEM_BADALLOC;
				return ret;
			}
			ret = OpenSOAPStringGetStringMB(elm_child->super.name, &node_name_len, node_name);
			if( OPENSOAP_FAILED(ret) ){
				return ret;
			}
			ret = OpenSOAPStringGetLengthMB(elm_child->super.value, &node_value_len);
			if( OPENSOAP_FAILED(ret) ){
				return ret;
			}
			node_value = malloc( (node_value_len+1)*sizeof(char) );
			if( !node_value ){
				ret = OPENSOAP_MEM_BADALLOC;
				return ret;
			}
			ret = OpenSOAPStringGetStringMB(elm_child->super.value, &node_value_len, node_value);
			if( OPENSOAP_FAILED(ret) ){
				return ret;
			}
			ret = OpenSOAPByteArrayCreate( &utf8_node_name );
			if( OPENSOAP_FAILED(ret) ){
				return ret;
			}
			ret = OpenSOAPXMLElmConvertEncodingToUTF8Libxml( node_name,
															   utf8_node_name,
															   &utf8_node_name_beg );
			if( OPENSOAP_FAILED(ret) ){
				return ret;
			}
			ret = OpenSOAPByteArrayCreate( &utf8_node_value );
			if( OPENSOAP_FAILED(ret) ){
				return ret;
			}
			ret = OpenSOAPXMLElmConvertEncodingToUTF8Libxml( node_value,
															   utf8_node_value,
															   &utf8_node_value_beg );
			if( OPENSOAP_FAILED(ret) ){
				return ret;
			}

			childNode = xmlNewChild( parentNode, 
									 NULL, 
									 utf8_node_name_beg, 
									 utf8_node_value_beg);

			free( node_name );
			free( node_value );
			ret = OpenSOAPByteArrayRelease( utf8_node_name );
			if( OPENSOAP_FAILED(ret) ){
				return ret;
			}
			ret = OpenSOAPByteArrayRelease( utf8_node_value );
			if( OPENSOAP_FAILED(ret) ){
				return ret;
			}
		}
		else{
			ret = OpenSOAPStringGetLengthMB(elm_child->super.name, &node_name_len);
			if( OPENSOAP_FAILED(ret) ){
				return ret;
			}
			node_name = malloc( (node_name_len+1)*sizeof(char) );
			if( !node_name ){
				ret = OPENSOAP_MEM_BADALLOC;
				return ret;
			}
			ret = OpenSOAPStringGetStringMB(elm_child->super.name, &node_name_len, node_name);
			if( OPENSOAP_FAILED(ret) ){
				return ret;
			}
			ret = OpenSOAPByteArrayCreate( &utf8_node_name );
			if( OPENSOAP_FAILED(ret) ){
				return ret;
			}
			ret = OpenSOAPXMLElmConvertEncodingToUTF8Libxml( node_name,
															   utf8_node_name,
															   &utf8_node_name_beg );
			if( OPENSOAP_FAILED(ret) ){
				return ret;
			}
			childNode = xmlNewChild( parentNode, 
									 NULL, 
									 utf8_node_name_beg, 
									 NULL );
			free( node_name );
			ret = OpenSOAPByteArrayRelease( utf8_node_name );
			if( OPENSOAP_FAILED(ret) ){
				return ret;
			}
		}

		/* add element's namespace */
		ret = OpenSOAPXMLElmAddNsInEncodingString( doc, elm_child, childNode );
		if( OPENSOAP_FAILED(ret) ){
			return ret;
		}

		/* add attributes */
		ret = OpenSOAPXMLElmAddAttrInEncodingString( doc, elm_child, childNode );
		if( OPENSOAP_FAILED(ret) ) {
			return ret;
		}

		/* add node */
		ret = OpenSOAPXMLElmAddNodeWithLibxml( doc, childNode, elm_child );
		if( OPENSOAP_FAILED(ret) ){
			return ret;
		}

		elm_next = (OpenSOAPXMLElmPtr)((OpenSOAPXMLNodePtr)elm_child)->next;
		elm_child = elm_next;
	}
		
	ret = OPENSOAP_NO_ERROR;

	return ret;
}


/*
=begin
--- function#OpenSOAPXMLElmConvertEncodingToUTF8Libxml( name, utf8_name, utf8_name_beg )
    Convert encoding from local encoding to utf8-encoding 
    
    :Parameters
      :const char * [in ] ((|name|))
        Const Charactor Pointer
      :OpenSOAPByteArrayPtr [in] ((|utf8_name|))
        OpenSOAP Byte Array Pointer
      :const unsigned char ** [in] ((|utf8_name_beg|))
        Const Unsigned Charactor Pointer's Pointer
    :Return value
      :type int
        error code.
      
=end
*/
int
OpenSOAPXMLElmConvertEncodingToUTF8Libxml(const char *name,
										  OpenSOAPByteArrayPtr utf8_name,
										  const unsigned char **utf8_name_beg ){
	int ret = OPENSOAP_YET_IMPLEMENTATION;

	OpenSOAPByteArrayPtr  mb_name = NULL;

	unsigned char null_char = '\0';
	size_t utf8_name_size = 0;

	ret = OpenSOAPByteArrayCreateWithData(name,
										  name ? strlen(name) : 0,
										  &mb_name);
	if( OPENSOAP_FAILED(ret) ){
		return ret;
	}

	ret = OpenSOAPStringConvertCharEncoding( NULL,
											 mb_name,
											 "UTF-8",
											 utf8_name );
	if( OPENSOAP_FAILED(ret) ){
		return ret;
	}


	ret = OpenSOAPByteArrayAppend( utf8_name, &null_char, sizeof(null_char) );
	if( OPENSOAP_FAILED(ret) ){
		return ret;
	}


	ret = OpenSOAPByteArrayGetBeginSizeConst( utf8_name,
											  utf8_name_beg,
											  &utf8_name_size );
	if( OPENSOAP_FAILED(ret) ){
		return ret;
	}

	ret = OPENSOAP_NO_ERROR;
	
	return ret;
}



/*
=begin
--- function#OpenSOAPXMLElmAddNsInEncodingString(doc, elm_child, childNode)
    Add namespace to the element node in libxml structure from opensoap structure 
    
    :Parameters
      :xmlDocPtr [in ] ((|doc|))
        Libxml Document Pointer
      :OpenSOAPXMLElmPtr [in] ((|elm_child|))
        OpenSOAP Element Pointer
      :xmlNodePtr [in, out] ((|childNode|))
        Libxml Node Pointer
    :Return value
      :type int
        error code.
      
=end
*/
int
OpenSOAPXMLElmAddNsInEncodingString( xmlDocPtr doc, OpenSOAPXMLElmPtr elm_child, xmlNodePtr childNode ){

	int ret = OPENSOAP_YET_IMPLEMENTATION;


	OpenSOAPXMLNamespacePtr ns_next = NULL;
	OpenSOAPXMLElmPtr	next_parent = NULL;

	xmlNsPtr	childNs = NULL;
	xmlNsPtr	searchNs = NULL;
	char *node_ns_uri = NULL;
	size_t node_ns_uri_len = 0;
	OpenSOAPByteArrayPtr utf8_node_ns_uri = NULL;
	const unsigned char *utf8_node_ns_uri_beg = NULL;

	char *node_ns_prefix = NULL;
	size_t node_ns_prefix_len = 0;
	OpenSOAPByteArrayPtr utf8_node_ns_prefix = NULL;
	const unsigned char *utf8_node_ns_prefix_beg = NULL;

	int compare_result;
	int hasSame_Ns;   /* 0:no same ns;  1:has same ns */

	if( doc == NULL || elm_child == NULL || childNode == NULL ) {
		ret = OPENSOAP_PARAMETER_BADVALUE; 
	}

	/* add this element's namespace */
	if( elm_child->super.thisNamespace != NULL ){
		hasSame_Ns = 0;  /* no same ns */
		/* check if there is a same ns in its parent's definedXMLNsList*/
		next_parent = (OpenSOAPXMLElmPtr)(((OpenSOAPXMLNodePtr)(elm_child))->parent);

		while( next_parent != NULL ){
			ns_next = next_parent->definedXMLNsList;
			while( ns_next != NULL && elm_child->super.thisNamespace != ns_next ){
				ns_next = ns_next->next;
			}
			if( ns_next != NULL ){
				hasSame_Ns = 1;  /* has same ns */
				break;
			}
			next_parent = (OpenSOAPXMLElmPtr)(next_parent->super.parent);
		}

		/*get string uri and prefix*/
		if( elm_child->super.thisNamespace->uri != NULL ){
			ret = OpenSOAPStringGetLengthMB(elm_child->super.thisNamespace->uri, &node_ns_uri_len);
			if( OPENSOAP_FAILED(ret) ){
				return ret;
			}
			node_ns_uri = malloc( (node_ns_uri_len+1)*sizeof(char) );
			if( !node_ns_uri ){
				ret = OPENSOAP_MEM_BADALLOC;
				return ret;
			}
			ret = OpenSOAPStringGetStringMB(elm_child->super.thisNamespace->uri, &node_ns_uri_len, node_ns_uri);
			if( OPENSOAP_FAILED(ret) ){
				return ret;
			}
			ret = OpenSOAPByteArrayCreate( &utf8_node_ns_uri );
			if( OPENSOAP_FAILED(ret) ){
				return ret;
			}
			ret = OpenSOAPXMLElmConvertEncodingToUTF8Libxml( node_ns_uri,
															 utf8_node_ns_uri,
															 &utf8_node_ns_uri_beg );
			if( OPENSOAP_FAILED(ret) ){
				return ret;
			}
		}

		if( elm_child->super.thisNamespace->prefix != NULL ){
			ret = OpenSOAPStringGetLengthMB(elm_child->super.thisNamespace->prefix, &node_ns_prefix_len);
			if( OPENSOAP_FAILED(ret) ){
				return ret;
			}
			node_ns_prefix = malloc( (node_ns_prefix_len+1)*sizeof(char) );
			if( !node_ns_prefix ){
				ret = OPENSOAP_MEM_BADALLOC;
				return ret;
			}
			ret = OpenSOAPStringGetStringMB(elm_child->super.thisNamespace->prefix, &node_ns_prefix_len, node_ns_prefix);
			if( OPENSOAP_FAILED(ret) ){
				return ret;
			}

			ret = OpenSOAPByteArrayCreate( &utf8_node_ns_prefix );
			if( OPENSOAP_FAILED(ret) ){
				return ret;
			}
			ret = OpenSOAPXMLElmConvertEncodingToUTF8Libxml( node_ns_prefix,
															 utf8_node_ns_prefix,
															 &utf8_node_ns_prefix_beg );
			if( OPENSOAP_FAILED(ret) ){
				return ret;
			}
		}

		/* check weather there is a namespace defined by node_ns_uri */
		searchNs = xmlSearchNsByHref(doc,
									 childNode,
									 utf8_node_ns_uri_beg);
		if (searchNs == NULL) {
			if (hasSame_Ns == 1) {  /* there is the same ns in parent */
				childNs = xmlNewNs(childNode, 
								   NULL,
								   utf8_node_ns_prefix_beg);
			}
			else {  /* there is no same ns in parent */
				childNs = xmlNewNs(childNode, 
								   utf8_node_ns_uri_beg,
								   utf8_node_ns_prefix_beg);
			}
		}
		else {
			if (searchNs->prefix != NULL && utf8_node_ns_prefix_beg != NULL) {
				compare_result
					= strcmp(searchNs->prefix, utf8_node_ns_prefix_beg);
				if (compare_result == 0){
					if (hasSame_Ns == 1) {
						childNs = xmlNewNs(childNode, 
										   NULL,
										   utf8_node_ns_prefix_beg);
					}
					else {
						childNs = searchNs;
					}
				}
				else{
					if( hasSame_Ns == 1 ) {
						childNs = xmlNewNs( childNode, 
											NULL,
											utf8_node_ns_prefix_beg );
					}
					else {
						childNs = xmlNewNs( childNode, 
											utf8_node_ns_uri_beg,
											utf8_node_ns_prefix_beg );
					}
				}
			}
			else{
				childNs = searchNs;
			}
		}

		xmlSetNs( childNode, childNs );

		if( node_ns_uri ){
			free( node_ns_uri );
			node_ns_uri = NULL;
		}
		if( node_ns_prefix ){
			free( node_ns_prefix );
			node_ns_prefix = NULL;
		}
		if( utf8_node_ns_uri ){
			ret = OpenSOAPByteArrayRelease( utf8_node_ns_uri );
			if( OPENSOAP_FAILED(ret) ){
				return ret;
			}
			utf8_node_ns_uri = NULL;
		}
		if( utf8_node_ns_prefix ){
			ret = OpenSOAPByteArrayRelease( utf8_node_ns_prefix );
			if( OPENSOAP_FAILED(ret ) ){
				return ret;
			}
			utf8_node_ns_prefix = NULL;
		}
	}
	else{
		/* add NULL namespace to this element */
		childNs = xmlNewNs( childNode, NULL, NULL );
		xmlSetNs( childNode, childNs );
		ret = OPENSOAP_NO_ERROR;
	}

	/* add the other namespace defined in this element */
	if( elm_child->definedXMLNsList != NULL ){
		ns_next = elm_child->definedXMLNsList;
		while( ns_next != NULL ){
			if( ns_next->uri != NULL ){
				ret = OpenSOAPStringGetLengthMB(ns_next->uri, &node_ns_uri_len);
				if( OPENSOAP_FAILED(ret) ){
					return ret;
				}
				node_ns_uri = malloc( (node_ns_uri_len+1)*sizeof(char) );
				if( !node_ns_uri ){
					ret = OPENSOAP_MEM_BADALLOC;
					return ret;
				}
				ret = OpenSOAPStringGetStringMB(ns_next->uri, &node_ns_uri_len, node_ns_uri);
				if( OPENSOAP_FAILED(ret) ){
					return ret;
				}
				ret = OpenSOAPByteArrayCreate( &utf8_node_ns_uri );
				if( OPENSOAP_FAILED(ret) ){
					return ret;
				}
				ret = OpenSOAPXMLElmConvertEncodingToUTF8Libxml( node_ns_uri,
																   utf8_node_ns_uri,
																   &utf8_node_ns_uri_beg );
				if( OPENSOAP_FAILED(ret) ){
					return ret;
				}
			}

			if( ns_next->prefix != NULL ){
				ret = OpenSOAPStringGetLengthMB(ns_next->prefix, &node_ns_prefix_len);
				if( OPENSOAP_FAILED(ret) ){
					return ret;
				}
				node_ns_prefix = malloc( (node_ns_prefix_len+1)*sizeof(char) );
				if( !node_ns_prefix ){
					ret = OPENSOAP_MEM_BADALLOC;
					return ret;
				}
				ret = OpenSOAPStringGetStringMB(ns_next->prefix, &node_ns_prefix_len, node_ns_prefix);
				if( OPENSOAP_FAILED(ret) ){
					return ret;
				}

				ret = OpenSOAPByteArrayCreate( &utf8_node_ns_prefix );
				if( OPENSOAP_FAILED(ret) ){
					return ret;
				}
				ret = OpenSOAPXMLElmConvertEncodingToUTF8Libxml( node_ns_prefix,
																   utf8_node_ns_prefix,
																   &utf8_node_ns_prefix_beg );
				if( OPENSOAP_FAILED(ret) ){
					return ret;
				}
			}

			/* check weather there is a namespace defined by node_ns_uri */
			searchNs = xmlSearchNsByHref( doc,
										  childNode,
										  utf8_node_ns_uri_beg );

			if( searchNs == NULL ){
				childNs = xmlNewNs(childNode, 
								   utf8_node_ns_uri_beg,
								   utf8_node_ns_prefix_beg);
			}
			else{
				if (searchNs->prefix != NULL
					&& utf8_node_ns_prefix_beg != NULL) {
					compare_result 
						= strcmp(searchNs->prefix, utf8_node_ns_prefix_beg);
					if (compare_result == 0) {
						childNs = searchNs;
					}
					else {
						childNs = xmlNewNs(childNode, 
										   utf8_node_ns_uri_beg,
										   utf8_node_ns_prefix_beg);
					}
				}
				else{
					childNs = searchNs;
				}
			}

			if( node_ns_uri ){
				free( node_ns_uri );
				node_ns_uri = NULL;
			}
			if( node_ns_prefix ){
				free( node_ns_prefix );
				node_ns_prefix = NULL;
			}
			if( utf8_node_ns_uri ){
				ret = OpenSOAPByteArrayRelease( utf8_node_ns_uri );
				if( OPENSOAP_FAILED(ret) ){
					return ret;
				}
				utf8_node_ns_uri = NULL;
			}
			if( utf8_node_ns_prefix ){
				ret = OpenSOAPByteArrayRelease( utf8_node_ns_prefix );
				if( OPENSOAP_FAILED(ret ) ){
					return ret;
				}
				utf8_node_ns_prefix = NULL;
			}
			ns_next = ns_next->next;
		}
	}
	return ret;
}	



/*
=begin
--- function#OpenSOAPXMLElmAddAttrInEncodingString(doc, elm_child, childNode)
    Add attribute to the element node in libxml structure from opensoap structure 
    
    :Parameters
      :xmlDocPtr [in ] ((|doc|))
        Libxml Document Pointer
      :OpenSOAPXMLElmPtr [in] ((|elm_child|))
        OpenSOAP Element Pointer
      :xmlNodePtr [in, out] ((|childNode|))
        Libxml Node Pointer
    :Return value
      :type int
        error code.
      
=end
*/
int
OpenSOAPXMLElmAddAttrInEncodingString( xmlDocPtr doc, OpenSOAPXMLElmPtr elm_child, xmlNodePtr childNode ) {

	int ret = OPENSOAP_YET_IMPLEMENTATION;

	OpenSOAPXMLAttrPtr	attr_next = NULL;

	char *node_attr_name = NULL;
	size_t node_attr_name_len = 0;
	OpenSOAPByteArrayPtr utf8_node_attr_name = NULL;
	const unsigned char *utf8_node_attr_name_beg = NULL;

	char *node_attr_value = NULL;
	size_t node_attr_value_len = 0;
	OpenSOAPByteArrayPtr utf8_node_attr_value = NULL;
	const unsigned char *utf8_node_attr_value_beg = NULL;

	xmlAttrPtr  childAttr = NULL;
	xmlNsPtr	childNs = NULL;
	xmlNsPtr	searchNs = NULL;

	char *node_ns_uri = NULL;
	size_t node_ns_uri_len = 0;
	OpenSOAPByteArrayPtr utf8_node_ns_uri = NULL;
	const unsigned char *utf8_node_ns_uri_beg = NULL;

	char *node_ns_prefix = NULL;
	size_t node_ns_prefix_len = 0;
	OpenSOAPByteArrayPtr utf8_node_ns_prefix = NULL;
	const unsigned char *utf8_node_ns_prefix_beg = NULL;

	int compare_result;

	if( doc == NULL || elm_child == NULL || childNode == NULL ) {
		ret = OPENSOAP_PARAMETER_BADVALUE; 
	}

	/* add attributes */
	if( elm_child->xmlAttrList != NULL ){
		attr_next = elm_child->xmlAttrList;

		while( attr_next != NULL ){
			if( attr_next->super.name != NULL ){
				ret = OpenSOAPStringGetLengthMB(attr_next->super.name, &node_attr_name_len);
				if( OPENSOAP_FAILED(ret) ){
					return ret;
				}
				node_attr_name = malloc( (node_attr_name_len+1)*sizeof(char) );
				if( !node_attr_name ){
					ret = OPENSOAP_MEM_BADALLOC;
					return ret;
				}
				ret = OpenSOAPStringGetStringMB(attr_next->super.name, &node_attr_name_len, node_attr_name);
				if( OPENSOAP_FAILED(ret) ){
					return ret;
				}
				ret = OpenSOAPByteArrayCreate( &utf8_node_attr_name );
				if( OPENSOAP_FAILED(ret) ){
					return ret;
				}
				ret = OpenSOAPXMLElmConvertEncodingToUTF8Libxml( node_attr_name,
																   utf8_node_attr_name,
																   &utf8_node_attr_name_beg );
				if( OPENSOAP_FAILED(ret) ){
					return ret;
				}
			}
			else{
				node_attr_name = NULL;
				utf8_node_attr_name = NULL;
			}

			if( attr_next->super.value != NULL ){
				ret = OpenSOAPStringGetLengthMB(attr_next->super.value, &node_attr_value_len);
				if( OPENSOAP_FAILED(ret) ){
					return ret;
				}
				node_attr_value = malloc( (node_attr_value_len+1)*sizeof(char) );
				if( !node_attr_value ){
					ret = OPENSOAP_MEM_BADALLOC;
					return ret;
				}
				ret = OpenSOAPStringGetStringMB(attr_next->super.value, &node_attr_value_len, node_attr_value);
				if( OPENSOAP_FAILED(ret) ){
					return ret;
				}

				ret = OpenSOAPByteArrayCreate( &utf8_node_attr_value );
				if( OPENSOAP_FAILED(ret) ){
					return ret;
				}
				ret = OpenSOAPXMLElmConvertEncodingToUTF8Libxml( node_attr_value,
																   utf8_node_attr_value,
																   &utf8_node_attr_value_beg );
				if( OPENSOAP_FAILED(ret) ){
					return ret;
				}
			}
			else{
				node_attr_value = NULL;
				utf8_node_attr_value = NULL;
			}

			if( attr_next->super.thisNamespace == NULL ){
				/* add attribute without namespace */
				childAttr = xmlSetProp( childNode,
										utf8_node_attr_name_beg,
										utf8_node_attr_value_beg );

				if( node_attr_name ){
					free( node_attr_name );
					node_attr_name = NULL;
				}
				if( node_attr_value ){
					free( node_attr_value );
					node_attr_value = NULL;
				}
			}
			else{
				/* add attributes and attribute's namespace */
				if( attr_next->super.thisNamespace->uri != NULL ){
					ret = OpenSOAPStringGetLengthMB(attr_next->super.thisNamespace->uri, &node_ns_uri_len);
					if( OPENSOAP_FAILED(ret) ){
						return ret;
					}
					node_ns_uri = malloc( (node_ns_uri_len+1)*sizeof(char) );
					if( !node_ns_uri ){
						ret = OPENSOAP_MEM_BADALLOC;
						return ret;
					}
					ret = OpenSOAPStringGetStringMB(attr_next->super.thisNamespace->uri, &node_ns_uri_len, node_ns_uri);
					if( OPENSOAP_FAILED(ret) ){
						return ret;
					}
					ret = OpenSOAPByteArrayCreate( &utf8_node_ns_uri );
					if( OPENSOAP_FAILED(ret) ){
						return ret;
					}
					ret = OpenSOAPXMLElmConvertEncodingToUTF8Libxml( node_ns_uri,
																	   utf8_node_ns_uri,
																	   &utf8_node_ns_uri_beg );
					if( OPENSOAP_FAILED(ret) ){
						return ret;
					}
				}
				else{
					node_ns_uri = NULL;
					utf8_node_ns_uri = NULL;
				}

				if( attr_next->super.thisNamespace->prefix != NULL ){
					ret = OpenSOAPStringGetLengthMB(attr_next->super.thisNamespace->prefix, &node_ns_prefix_len);
					if( OPENSOAP_FAILED(ret) ){
						return ret;
					}
					node_ns_prefix = malloc( (node_ns_prefix_len+1)*sizeof(char) );
					if( !node_ns_prefix ){
						ret = OPENSOAP_MEM_BADALLOC;
						return ret;
					}
					ret = OpenSOAPStringGetStringMB(attr_next->super.thisNamespace->prefix, &node_ns_prefix_len, node_ns_prefix);
					if( OPENSOAP_FAILED(ret) ){
						return ret;
					}
					ret = OpenSOAPByteArrayCreate( &utf8_node_ns_prefix );
					if( OPENSOAP_FAILED(ret) ){
						return ret;
					}
					ret = OpenSOAPXMLElmConvertEncodingToUTF8Libxml( node_ns_prefix,
																	   utf8_node_ns_prefix,
																	   &utf8_node_ns_prefix_beg );
					if( OPENSOAP_FAILED(ret) ){
						return ret;
					}
				}
				else{
					node_ns_prefix = NULL;
					utf8_node_ns_prefix = NULL;
				}

				/* check weather there is a namespace definded by node_ns_uri */
				searchNs = xmlSearchNs( doc,
										childNode,
										utf8_node_ns_prefix_beg );
				if( searchNs == NULL ){
					childNs = xmlNewNs(childNode, 
									   utf8_node_ns_uri_beg,
									   utf8_node_ns_prefix_beg);
				}
				else {
					if (searchNs->href != NULL
						&& utf8_node_ns_uri_beg != NULL) {
						compare_result 
							= strcmp(searchNs->href, utf8_node_ns_uri_beg);
						if (compare_result == 0) {
							childNs = searchNs;
						}
						else {
							childNs = xmlNewNs(childNode, 
											   utf8_node_ns_uri_beg,
											   utf8_node_ns_prefix_beg);
						}
					}
					else{
						if( searchNs->href == NULL && utf8_node_ns_uri_beg == NULL ){
							childNs = searchNs;
						}
						else{
							childNs = xmlNewNs( childNode, 
												utf8_node_ns_uri_beg,
												utf8_node_ns_prefix_beg );
						}
					}
				}

				childAttr = xmlNewNsProp( childNode, 
										  childNs,
										  utf8_node_attr_name_beg,
										  utf8_node_attr_value_beg);
				if( node_attr_name ){
					free( node_attr_name );
					node_attr_name = NULL;
				}
				if( node_attr_value ){
					free( node_attr_value );
					node_attr_value = NULL;
				}
				if( node_ns_uri ){
					free( node_ns_uri );
					node_ns_uri = NULL;
				}
				if( node_ns_prefix ){
					free( node_ns_prefix );
					node_ns_prefix = NULL;
				}
				if( utf8_node_ns_uri ){
					ret = OpenSOAPByteArrayRelease( utf8_node_ns_uri );
					if( OPENSOAP_FAILED(ret) ){
						return ret;
					}
					utf8_node_ns_uri = NULL;
				}
				if( utf8_node_ns_prefix ){
					ret = OpenSOAPByteArrayRelease( utf8_node_ns_prefix );
					if( OPENSOAP_FAILED(ret) ){
						return ret;
					}
					utf8_node_ns_prefix = NULL;
				}
			}

			if( utf8_node_attr_name ){
				ret = OpenSOAPByteArrayRelease( utf8_node_attr_name );
				if( OPENSOAP_FAILED(ret) ){
					return ret;
				}
				utf8_node_attr_name = NULL;
			}
			if( utf8_node_attr_value ){
				ret = OpenSOAPByteArrayRelease( utf8_node_attr_value );
				if( OPENSOAP_FAILED(ret) ){
					return ret;
				}
				utf8_node_attr_value = NULL;
			}

			attr_next = (OpenSOAPXMLAttrPtr)(attr_next->super.next);
		}
	}
	else{
		ret = OPENSOAP_NO_ERROR;
	}
	return ret;
}



/*
=begin
--- function#OpenSOAPXMLElmCreateChildNodeFromLibxml( elm, node )
    create opensoap structure from libxml structure 
    
    :Parameters
      :OpenSOAPXMLElmPtr [in, out] ((|elm|))
        OpenSOAP Element Pointer
      :xmlNodePtr [in] ((|node|))
        Libxml Node Pointer
    :Return value
      :type int
        error code.
      
=end
*/
int
OpenSOAPXMLElmCreateChildNodeFromLibxml( OpenSOAPXMLElmPtr elm, xmlNodePtr node ){
	int ret = OPENSOAP_YET_IMPLEMENTATION;

	OpenSOAPXMLElmPtr	elm_next = NULL;
	OpenSOAPXMLElmPtr	elm_parent = NULL;
	OpenSOAPXMLElmPtr	elm_cld = NULL;

	OpenSOAPXMLAttrPtr	new_attr = NULL;
	OpenSOAPXMLNamespacePtr ns_def = NULL;

	OpenSOAPStringPtr	opensoap_nodevalue = NULL;
	OpenSOAPStringPtr	attr_name = NULL;
	OpenSOAPStringPtr	attr_value = NULL;

	xmlNodePtr	nextNode = NULL;
	xmlNodePtr	childNode = NULL;
	xmlAttrPtr	nextAttr = NULL;
	xmlNsPtr	nextNs = NULL;
	int isBlankNode;

	if( elm == NULL || node == NULL ){
		ret = OPENSOAP_PARAMETER_BADVALUE;
		return ret;
	}

	childNode = node->children;
	isBlankNode = xmlIsBlankNode( childNode );
	if( isBlankNode == 1 ) {
		childNode = childNode->next;
	}
	elm_parent = elm;

	while (childNode != NULL) {
		if (!childNode->name || strcmp(childNode->name, "text") != 0) {
			/* add node in OpenSOAPXMLElm */
			OpenSOAPByteArrayPtr mb_name = NULL;
			const unsigned char *mb_name_beg = NULL;

			ret = OpenSOAPByteArrayCreate(&mb_name);
			if (OPENSOAP_FAILED(ret)) {
				return ret;
			}

			ret = OpenSOAPXMLElmConvertEncodingFromUTF8Libxml( childNode->name,
																 mb_name,
																 &mb_name_beg );
			if( OPENSOAP_FAILED(ret) ){
				return ret;
			}

			elm_cld = NULL;
			ret = OpenSOAPXMLElmAddChildMB( elm_parent, mb_name_beg, &elm_cld );
			if( OPENSOAP_FAILED(ret) ){
				return ret;
			}

			ret = OpenSOAPByteArrayRelease( mb_name );
			if( OPENSOAP_FAILED(ret) ){
				return ret;
			}

			/* add element namespace */
			ret = OpenSOAPXMLElmAddNsInCharEncoding( childNode, elm_cld );
			if( OPENSOAP_FAILED(ret) ){
				return ret;
			}

			/* add element's attributes */
			ret = OpenSOAPXMLElmAddAttrInCharEncoding( childNode, elm_cld );
			if( OPENSOAP_FAILED(ret) ){
				return ret;
			}
			
			/* add element's child */
			ret = OpenSOAPXMLElmCreateChildNodeFromLibxml( elm_cld, childNode );

			nextNode = childNode->next;
			isBlankNode = xmlIsBlankNode(nextNode);
			if (isBlankNode == 1) {
				nextNode = nextNode->next;
			}
			childNode = nextNode;
		}
		else {
			OpenSOAPByteArrayPtr mb_content = NULL;
			const unsigned char *mb_content_beg = NULL;
            size_t mb_content_sz = 0;
            unsigned char *mb_content_beg_noLF = NULL;
            size_t mb_content_sz_noLF = 0;
            int i;

			ret = OpenSOAPByteArrayCreate( &mb_content );
			if( OPENSOAP_FAILED(ret) ){
				return ret;
			}
			ret = OpenSOAPXMLElmConvertEncodingFromUTF8Libxml( childNode->content,
																 mb_content,
																 &mb_content_beg );
			if( OPENSOAP_FAILED(ret) ){
				return ret;
			}

			/* delete '\t' '\n' '\r' in the content */
            ret = OpenSOAPByteArrayGetBeginSizeConst(mb_content,
													 &mb_content_beg,
													 &mb_content_sz);
            mb_content_beg_noLF
				= malloc(mb_content_sz * sizeof(unsigned char));
            if (mb_content_beg_noLF == NULL) {
                ret = OPENSOAP_MEM_BADALLOC;
                return ret;
            }

            for( i=0; i<(int)mb_content_sz; i++ ){
                if( mb_content_beg[i] != '\n' && mb_content_beg[i] != '\r' && mb_content_beg[i] != '\t'){
                    mb_content_beg_noLF[mb_content_sz_noLF] = mb_content_beg[i];
                    mb_content_sz_noLF++;
                }
            }

			ret = OpenSOAPStringCreateWithMB( mb_content_beg_noLF, &opensoap_nodevalue );
			if( OPENSOAP_FAILED(ret) ) {
				free( mb_content_beg_noLF );
				return ret;
			}
			free( mb_content_beg_noLF );

			ret = OpenSOAPXMLElmSetValueMB( elm_parent, "string", &opensoap_nodevalue );
			if( OPENSOAP_FAILED(ret) ) {
				return ret;
			}
			ret = OpenSOAPByteArrayRelease( mb_content );
			if( OPENSOAP_FAILED(ret) ){
				return ret;
			}

			nextNode = childNode->next;
			isBlankNode = xmlIsBlankNode( nextNode );
			if( isBlankNode == 1 ) {
				nextNode = nextNode->next;
			}
			childNode = nextNode;
		}
	}

	ret = OPENSOAP_NO_ERROR;

	return ret;
}



/*
=begin
--- function#OpenSOAPXMLElmConvertEncodingFromUTF8Libxml( name, mb_name, mb_name_beg )
    Convert encoding from utf8-enc libxml to local encoding
    
    :Parameters
      :const xmlChar * [in ] ((|name|))
        Libxml Charactor Pointer
      :OpenSOAPByteArrayPtr [in] ((|mb_name|))
        OpenSOAP Byte Array Pointer
      :const unsigned char ** [in] ((|mb_name_beg|))
        Const Unsigned Charactor Pointer's Pointer
    :Return value
      :type int
        error code.
      
=end
*/
int
OpenSOAPXMLElmConvertEncodingFromUTF8Libxml(const xmlChar *name,
											OpenSOAPByteArrayPtr mb_name,
											const unsigned char **
											mb_name_beg) {
	int ret = OPENSOAP_YET_IMPLEMENTATION;

	OpenSOAPByteArrayPtr  utf8_name = NULL;

	unsigned char null_char = '\0';
	size_t mb_name_size = 0;

	if (name == NULL && mb_name == NULL) {
		ret = OPENSOAP_PARAMETER_BADVALUE;
		return ret;
	}

	ret = OpenSOAPByteArrayCreateWithData(name,
										  name ? strlen(name) : 0,
										  &utf8_name);
	if (OPENSOAP_FAILED(ret)) {
		return ret;
	}

	ret = OpenSOAPStringConvertCharEncoding("UTF-8",
											utf8_name,
											NULL,
											mb_name);
	if (OPENSOAP_FAILED(ret)) {
		return ret;
	}


	ret = OpenSOAPByteArrayAppend(mb_name, &null_char, sizeof(null_char));
	if (OPENSOAP_FAILED(ret)) {
		return ret;
	}


	ret = OpenSOAPByteArrayGetBeginSizeConst( mb_name,
											  mb_name_beg,
											  &mb_name_size );
	if( OPENSOAP_FAILED(ret) ){
		return ret;
	}

	ret = OpenSOAPByteArrayRelease( utf8_name );
	if( OPENSOAP_FAILED(ret) ){
		return ret;
	}

	ret = OPENSOAP_NO_ERROR;
	
	return ret;
}


/*
=begin
--- function#OpenSOAPXMLElmAddAttrInCharEncoding( childNode, elm_cld )
    Add attribute to the element in opensoap structure from libxml structure  
    
    :Parameters
      :xmlNodePtr [in, out] ((|childNode|))
        Libxml Node Pointer
      :OpenSOAPXMLElmPtr [in] ((|elm_cld|))
        OpenSOAP Element Pointer
    :Return value
      :type int
        error code.
      
=end
*/
int
OpenSOAPXMLElmAddAttrInCharEncoding( xmlNodePtr childNode, OpenSOAPXMLElmPtr elm_cld ){
	int ret = OPENSOAP_YET_IMPLEMENTATION;

	xmlAttrPtr nextAttr = NULL;
	xmlNsPtr   nextNs = NULL;

	OpenSOAPStringPtr attr_value = NULL;
	OpenSOAPXMLAttrPtr new_attr = NULL;


	/* add element's attributes */
	if( childNode->properties != NULL ){
		nextAttr = childNode->properties;
		while( nextAttr != NULL ){
			OpenSOAPByteArrayPtr mb_name = NULL;
			const unsigned char *mb_name_beg = NULL;
			OpenSOAPByteArrayPtr mb_content = NULL;
			const unsigned char *mb_content_beg = NULL;

			if( nextAttr->name != NULL ){
				ret = OpenSOAPByteArrayCreate( &mb_name );
				if( OPENSOAP_FAILED(ret) ){
					return ret;
				}
				ret = OpenSOAPXMLElmConvertEncodingFromUTF8Libxml( nextAttr->name,
																	 mb_name,
																	 &mb_name_beg );
				if( OPENSOAP_FAILED(ret) ){
					return ret;
				}
			}
			else{
				mb_name = NULL;
				mb_name_beg = NULL;
			}

			if( nextAttr->children->content != NULL ){
				ret = OpenSOAPByteArrayCreate( &mb_content );
				if( OPENSOAP_FAILED(ret) ){
					return ret;
				}
				ret = OpenSOAPXMLElmConvertEncodingFromUTF8Libxml( nextAttr->children->content,
																	 mb_content,
																	 &mb_content_beg );
				if( OPENSOAP_FAILED(ret) ){
					return ret;
				}
			}
			else{
				mb_content = NULL;
				mb_content_beg = NULL;
			}

			if( mb_content_beg != NULL ){
				ret = OpenSOAPStringCreateWithMB( mb_content_beg, &attr_value );
				if( OPENSOAP_FAILED(ret) ) {
					return ret;
				}
			}
			else{
				attr_value = NULL;
			}

			new_attr = NULL;
			ret = OpenSOAPXMLElmAddAttributeMB( elm_cld, 
												mb_name_beg,
												"string", 
												&attr_value,
												&new_attr );
			if( OPENSOAP_FAILED(ret) ) {
				return ret;
			}

			if( mb_name != NULL ){ 
				ret = OpenSOAPByteArrayRelease( mb_name );
				if( OPENSOAP_FAILED(ret) ){
					return ret;
				}
				mb_name = NULL;
			}

			if( mb_content != NULL ){
				ret = OpenSOAPByteArrayRelease( mb_content );
				if( OPENSOAP_FAILED(ret) ){
					return ret;
				}
				mb_content = NULL;
			}

			/* add attributes' namespace */
			if( nextAttr->ns != NULL ){
				OpenSOAPByteArrayPtr mb_href = NULL;
				const unsigned char *mb_href_beg = NULL;
				OpenSOAPByteArrayPtr mb_prefix = NULL;
				const unsigned char *mb_prefix_beg = NULL;

				nextNs = nextAttr->ns;

				if( nextNs->href != NULL ){
					ret = OpenSOAPByteArrayCreate( &mb_href );
					if( OPENSOAP_FAILED(ret) ){
						return ret;
					}
					ret = OpenSOAPXMLElmConvertEncodingFromUTF8Libxml( nextNs->href,
																		 mb_href,
																		 &mb_href_beg );
					if( OPENSOAP_FAILED(ret) ){
						return ret;
					}
				}
				else{
					mb_href = NULL;
					mb_href_beg = NULL;
				}

				if( nextNs->prefix != NULL ){
					ret = OpenSOAPByteArrayCreate( &mb_prefix );
					if( OPENSOAP_FAILED(ret) ){
						return ret;
					}
					ret = OpenSOAPXMLElmConvertEncodingFromUTF8Libxml( nextNs->prefix,
																		 mb_prefix,
																		 &mb_prefix_beg );
					if( OPENSOAP_FAILED(ret) ){
						return ret;
					}
				}
				else{
					mb_prefix = NULL;
					mb_prefix_beg = NULL;
				}

				ret = OpenSOAPXMLAttrSetNamespaceMB( new_attr,
													 mb_href_beg,
													 mb_prefix_beg );
				if( OPENSOAP_FAILED(ret) ){
					return ret;
				}

				if( mb_href != NULL ){
					ret = OpenSOAPByteArrayRelease( mb_href );
					if( OPENSOAP_FAILED(ret) ){
						return ret;
					}
					mb_href = NULL;
				}
				if( mb_prefix != NULL ){
					ret = OpenSOAPByteArrayRelease( mb_prefix );
					if( OPENSOAP_FAILED(ret) ){
						return ret;
					}
					mb_prefix = NULL;
				}
			}
			nextAttr = nextAttr->next;
		}
	}
	else{
		ret = OPENSOAP_NO_ERROR;
	}
	return ret;
}


/*
=begin
--- function#OpenSOAPXMLElmAddNsInCharEncoding( childNode, elm_cld )
    Add namespace to the element in opensoap structure from libxml structure  
    
    :Parameters
      :xmlNodePtr [in, out] ((|childNode|))
        Libxml Node Pointer
      :OpenSOAPXMLElmPtr [in] ((|elm_cld|))
        OpenSOAP Element Pointer
    :Return value
      :type int
        error code.
      
=end
*/
int
OpenSOAPXMLElmAddNsInCharEncoding( xmlNodePtr childNode, OpenSOAPXMLElmPtr elm_cld ){
	int ret = OPENSOAP_YET_IMPLEMENTATION;

	xmlNsPtr   nextNs = NULL;

	OpenSOAPXMLNamespacePtr ns_def = NULL;

	/* add the associated namespace of this element */
	if( childNode->ns != NULL ){
		OpenSOAPByteArrayPtr mb_href = NULL;
		const unsigned char *mb_href_beg = NULL;
		OpenSOAPByteArrayPtr mb_prefix = NULL;
		const unsigned char *mb_prefix_beg = NULL;

		nextNs = childNode->ns;

		if( nextNs->href != NULL ){
			ret = OpenSOAPByteArrayCreate( &mb_href );
			if( OPENSOAP_FAILED(ret) ){
				return ret;
			}
			ret = OpenSOAPXMLElmConvertEncodingFromUTF8Libxml( nextNs->href,
																 mb_href,
																 &mb_href_beg );
			if( OPENSOAP_FAILED(ret) ){
				return ret;
			}
		}
		else{
			mb_href = NULL;
			mb_href_beg = NULL;
		}

		if( nextNs->prefix != NULL ){
			ret = OpenSOAPByteArrayCreate( &mb_prefix );
			if( OPENSOAP_FAILED(ret) ){
				return ret;
			}
			ret = OpenSOAPXMLElmConvertEncodingFromUTF8Libxml( nextNs->prefix,
																 mb_prefix,
																 &mb_prefix_beg );
			if( OPENSOAP_FAILED(ret) ){
				return ret;
			}
		}
		else{
			mb_prefix = NULL;
			mb_prefix_beg = NULL;
		}

		ret = OpenSOAPXMLElmSetNamespaceMB( elm_cld,
											mb_href_beg,
											mb_prefix_beg );
		if( OPENSOAP_FAILED(ret) ){
			return ret;
		}

		if( mb_href ){
			ret = OpenSOAPByteArrayRelease( mb_href );
			if( OPENSOAP_FAILED(ret) ){
				return ret;
			}
			mb_href = NULL;
		}
		if( mb_prefix ){
			ret = OpenSOAPByteArrayRelease( mb_prefix );
			if( OPENSOAP_FAILED(ret) ){
				return ret;
			}
			mb_prefix = NULL;
		}
	}

	/* add namespace definitions on this element */
	if( childNode->nsDef != NULL ){
		nextNs = childNode->nsDef;
		while( nextNs != NULL ){
			OpenSOAPByteArrayPtr mb_href = NULL;
			const unsigned char *mb_href_beg = NULL;
			OpenSOAPByteArrayPtr mb_prefix = NULL;
			const unsigned char *mb_prefix_beg = NULL;

			if( nextNs->href != NULL ){
				ret = OpenSOAPByteArrayCreate( &mb_href );
				if( OPENSOAP_FAILED(ret) ){
					return ret;
				}
				ret = OpenSOAPXMLElmConvertEncodingFromUTF8Libxml( nextNs->href,
																	 mb_href,
																	 &mb_href_beg );
				if( OPENSOAP_FAILED(ret) ){
					return ret;
				}
			}
			else{
				mb_href = NULL;
				mb_href_beg = NULL;
			}

			if( nextNs->prefix != NULL ){
				ret = OpenSOAPByteArrayCreate( &mb_prefix );
				if( OPENSOAP_FAILED(ret) ){
					return ret;
				}
				ret = OpenSOAPXMLElmConvertEncodingFromUTF8Libxml( nextNs->prefix,
																	 mb_prefix,
																	 &mb_prefix_beg );
				if( OPENSOAP_FAILED(ret) ){
					return ret;
				}
			}
			else{
				mb_prefix = NULL;
				mb_prefix_beg = NULL;
			}

			ret = OpenSOAPXMLElmDefineNamespaceMB( elm_cld,
												   mb_href_beg,
												   mb_prefix_beg,
												   &ns_def);
			if( OPENSOAP_FAILED(ret) ){
				return ret;
			}

			if( mb_href ){
				ret = OpenSOAPByteArrayRelease( mb_href );
				if( OPENSOAP_FAILED(ret) ){
					return ret;
				}
				mb_href = NULL;
			}
			if( mb_prefix ){
				ret = OpenSOAPByteArrayRelease( mb_prefix );
				if( OPENSOAP_FAILED(ret) ){
					return ret;
				}
				mb_prefix = NULL;
			}
			nextNs = nextNs->next;
		}
	}
	else{
		ret = OPENSOAP_NO_ERROR;
	}
	return ret;
}

/*
=begin
--- function#OpenSOAPXMLElmGetNameString(elm, name)
    Getting of XML Element's Name as OpenSOAPString

    :Parameters
      :[in]  OpenSOAPXMLElmPtr ((|elm|))
        XML Element Pointer.
      :[out] OpenSOAPStringPtr * ((|name|))
        Result Name of Node.
    :Return Value
      :int
	    Return Code Is ErrorCode
=end
 */
int
OPENSOAP_API
OpenSOAPXMLElmGetNameString(/* [in]  */ OpenSOAPXMLElmPtr elm,
							/* [out] */ OpenSOAPStringPtr *name) {
	int ret = OpenSOAPXMLNodeGetName((OpenSOAPXMLNodePtr)elm, name);

	return ret;
}



/*
=begin
--- function#OpenSOAPXMLElmGetNextAttr(elm, attr_elm)
    Get next attr XML Element.

    :Parameters
      :OpenSOAPXMLElmPtr [in] ((|elm|))
        OpenSOAP XML Element.
      :OpenSOAPXMLAttrPtr * [in, out] ((|attr_elm|))
        Next Attr Element Pointer of Assignment XML Element
        But If *attr_elm as in NULL then Return To First XML Element Pointer 
    :Return Value value
      :int
	    Return Code Is ErrorCode
=end
*/
extern
int
OPENSOAP_API
OpenSOAPXMLElmGetNextAttr(OpenSOAPXMLElmPtr /* [in] */ elm,
                           OpenSOAPXMLAttrPtr * /* [in, out] */ attr_elm) {
    int ret = OPENSOAP_YET_IMPLEMENTATION;

	OpenSOAPXMLNodePtr  attr_next;

	if( !elm ){
		ret = OPENSOAP_PARAMETER_BADVALUE;
		return ret;
	}

	if( !(*attr_elm) ){
		*attr_elm = (OpenSOAPXMLAttrPtr)(elm->xmlAttrList);
		ret = OPENSOAP_NO_ERROR;
	}
	else{
		attr_next = (OpenSOAPXMLNodePtr)(elm->xmlAttrList);
		while( attr_next != NULL ){
			if( attr_next == (OpenSOAPXMLNodePtr)(*attr_elm) ){
				*attr_elm = (OpenSOAPXMLAttrPtr)((*attr_elm)->super.next);
				ret = OPENSOAP_NO_ERROR;
				return ret;
			}
			attr_next = attr_next->next;
		}
	}
    return ret;
}

