/* -*- mode: c++; -*-
 *-----------------------------------------------------------------------------
 * $RCSfile: ChannelManager.cpp,v $
 *
 * See Copyright for the status of this software.
 *
 * The OpenSOAP Project
 * http://opensoap.jp/
 *-----------------------------------------------------------------------------
 */

#if HAVE_CONFIG_H
# include "config.h"
#endif

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/un.h>
#include <cstdio>
#include <unistd.h>
#include <string>

#include <stdexcept>

#include <pthread.h>

#include "ServerCommon.h"
#include "ChannelManager.h"

using namespace OpenSOAP;

/*
  ChannelManager::
*/

//for LOG
static const std::string CLASS_SIG = "ChannelManager";

ChannelManager::ChannelManager()
{
#ifdef DEBUG
  cerr << CLASS_SIG << "::" << CLASS_SIG << endl;
#endif /* DEBUG */  
  setMaskFlag = false;
}

ChannelManager::~ChannelManager()
{
#ifdef DEBUG
  cerr << CLASS_SIG << "::~" << CLASS_SIG << endl;
#endif /* DEBUG */  

}

void
ChannelManager::setMask(mode_t newMask)
{
  newMask_ = newMask;
  setMaskFlag = true;
}

bool ChannelManager::setSocketAddr(const std::string& addr)
{
#ifdef DEBUG
  cerr << CLASS_SIG << "::setSocketAddr() arg=[" 
       << addr << "]" << endl;
#endif /* DEBUG */
  /*
  if (0 != unlink(addr.c_str())) {
    return false;
  }
  */
  socketAddr_ = addr;

  return true;
}

void ChannelManager::run()
{
  //===================================================================
  // ʐM ChannelManager <-> ChannelSelector
  //===================================================================
  typedef struct sockaddr SockAddr;
  typedef struct sockaddr_un SockAddrUn;

  int sockfd = int();
  int newsockfd = int();
  SockAddrUn serverAddr;
  SockAddrUn clientAddr;
#if HAVE_TYPE_OF_ADDRLEN_AS_SOCKLEN_T
  socklen_t addrLen = socklen_t(); //socket_t();
  socklen_t clientLen = socklen_t();
#endif
#if HAVE_TYPE_OF_ADDRLEN_AS_SIZE_T
  // Solaris 2.6 does not have socklen_t
  size_t addrLen = size_t(); 
  size_t clientLen = size_t();
#endif
#if HAVE_TYPE_OF_ADDRLEN_AS_INT
  // Solaris 2.6 does not have socklen_t
  int addrLen = int(); 
  int clientLen = int();
#endif
  
  //ChannelSelectorƂ̒ʐM|[g
  if (0 > (sockfd = socket(AF_UNIX, SOCK_STREAM, 0))) {
    cerr << TAG_ERR << CLASS_SIG << "::" << CLASS_SIG
	 << ": open socket failed" << endl;
    //throw std::exception();
    // ONX̐݌vKv 
  }
#ifdef DEBUG
  cerr << CLASS_SIG << "::" << CLASS_SIG 
       << ": socket open OK!" << endl;
#endif

  memset((char*)&serverAddr, 0, sizeof(serverAddr));
  serverAddr.sun_family = AF_UNIX;
  strcpy(serverAddr.sun_path, socketAddr_.c_str());
  addrLen = sizeof(serverAddr.sun_family) + strlen(serverAddr.sun_path);

  unlink(socketAddr_.c_str());

  if (setMaskFlag == true) {
    curMask = umask(newMask_);
  }

  if (0 > bind(sockfd, (SockAddr*)&serverAddr, addrLen)) {
    cerr << TAG_ERR << CLASS_SIG << "::" << CLASS_SIG
	 << ": bind failed" << endl;
    //throw std::exception();
    // ONX̐݌vKv 
  }
  
  if (setMaskFlag == true) {
    umask(curMask);
  }

#ifdef DEBUG
  cerr << CLASS_SIG << "::" << CLASS_SIG 
       << ": bind OK!" << endl;
#endif

  if (0 > listen(sockfd, 5)) {
    cerr << TAG_ERR << CLASS_SIG << "::" << CLASS_SIG
	 << ": listen failed" << endl;
    //throw std::exception();
    // ONX̐݌vKv 
  }

#ifdef DEBUG
  cerr << CLASS_SIG << "::" << CLASS_SIG 
       << ": listen OK!" << endl;
#endif
    
  for (;;) {
    clientLen = sizeof(clientAddr);

#ifdef DEBUG
    cerr << CLASS_SIG << "::" << CLASS_SIG 
	 << ": accept ready!" << endl;
#endif
    
    newsockfd = accept(sockfd, (SockAddr*)&clientAddr, &clientLen);
    if (0 > newsockfd) {
      cerr << TAG_ERR << CLASS_SIG << "::" << CLASS_SIG
	   << ": accept failed" << endl;
      //throw std::exception();
      // ONX̐݌vKv 
    }

#ifdef DEBUG
    cerr << CLASS_SIG << "::" << CLASS_SIG 
	 << ": accept OK! fd=(" << newsockfd << ")" << endl;
#endif

    //XbhɓnKv̂f[^
    // Eint newsockfd;
    // EChannelManager* this;
    ThrData* thrData = new ThrData();
    thrData->that = this;
    thrData->sockfd = newsockfd;

    //ʐMXbh쐬
#if 0
    connectionThread((void*)thrData);
#else
    pthread_t thrId;
    int status = pthread_create(&thrId,
				NULL,
				connectionThread,
				(void*)thrData);
    if (0 != status) {
      cerr << TAG_ERR << CLASS_SIG << "::" << CLASS_SIG
	   << ": create thread failed" << endl;
      //throw std::exception();
      // ONX̐݌vKv 
    }
#endif

  } // end for (;;)
}

void* ChannelManager::connectionThread(void* aArg)
{
#ifdef DEBUG
  cerr << CLASS_SIG << "::connectionThread()" << endl;
#endif /* DEBUG */

  //ñ`FbN
  if (!aArg) {
    cerr << TAG_ERR << CLASS_SIG 
	 << "::connectionThread: invalid argument" << endl;
    pthread_exit(NULL);
  }

  //n̓WJ
  ThrData* thrData = (ThrData*)aArg;
  ChannelManager* that = thrData->that;
  int sockfd = thrData->sockfd;
  delete thrData;

#ifdef DEBUG
  cerr << CLASS_SIG << "::connectionThread: sockfd=(" << sockfd << ")" << endl;
#endif /* DEBUG */

  that->doProc(sockfd);

  if (0 < sockfd) {
    close(sockfd);
  }

  pthread_exit(NULL);

  return NULL;
}


#if 0
//std::string ChannelManager::parseReadData(const std::string& aReadData)
/*
int ChannelManager::parseReadData(const std::string& aReadData,
				  std::string& response)
*/
int ChannelManager::parseAndOpen(const std::string& aReadData,
				 std::string& response,
				 ChannelDescriptor& chnlDesc)
{
#ifdef DEBUG
  cerr << CLASS_SIG << "::parseReadData() arg=["
       << aReadData << "]" << endl;
#endif /* DEBUG */

  return 0;
}
#endif

void ChannelManager::doProc(int sockfd)
{
#ifdef DEBUG
  cerr << CLASS_SIG << "::doProc() sockfd=[" << sockfd << "]" << endl;
#endif /* DEBUG */

}

