/* -*- mode: c++; -*-
 *-----------------------------------------------------------------------------
 * $RCSfile: Forwarder.cpp,v $
 *
 * See Copyright for the status of this software.
 *
 * The OpenSOAP Project
 * http://opensoap.jp/
 *-----------------------------------------------------------------------------
 */

#include <algorithm>
#include <string>
#include <vector>

#include <sys/time.h>
#include <unistd.h>

/* common library */
#include "connection.h"
#include "SOAPMessage.h"
#include "StringUtil.h"
#include "HTTPMessage.h"
#include "FileIDFunc.h"
#include "Timeout.h"
#include "FwdInfoFunc.h"

/* server common library */
#include "ServerCommon.h"
#include "ChannelDescriptor.h"

#include "TTLTableRefChnlSelector.h"

/* forwarder library */
#include "Forwarder.h"

using namespace OpenSOAP;

Forwarder::Forwarder(const ChannelDescriptor& chnlDesc)
  : chnlDesc_(chnlDesc)
{
  cerr << "Forwarder::Forwarder" << endl;
  
  pthread_mutex_init(&_mt, NULL);
  
  defaultForwarding = getForwarderInfo(forwardHostname, forwardPort, 
				       forwardCGILocation);
  
  localhostName = getLocalhostName();
  std::vector<std::string> backwardURLs;
  bool hasBackwardURL = getBackwardURLs(backwardURLs);
  backwardURL = backwardURLs[0];
  if (!hasBackwardURL) {
    backwardURL = "http://localhost:80/soapInterface.cgi";
  }
}// end of Forwarder()

//----- Destructor -----//
Forwarder::~Forwarder()
{
  cerr << "Forwarder::~Forwarder" << endl;
  pthread_mutex_destroy(&_mt);
}// end of ~Forwarder()

//----- bZ[W]Xbh -----//
void*
Forwarder::syncMessageForward(void* arg)
{
  Forwarder* that = (Forwarder*)arg;
  
  int lock_status = pthread_mutex_lock(&(that->_mt));
  if (0 != lock_status) {
    cerr << "[[Error!!]] void* Forwarder::localSyncThrFunc(void*)" << endl;
    cerr << "            pthread_mutex_lock() failed." << endl;
    cerr.flush();
    return NULL;
  }
  
  
  // HTTP bZ[W쐬B
  std::string http_msg 
    = createHTTPMessage(that->forwardCGILocation,
			that->soapMsgFromMsgDrv, that->operationName);
  
  // for debug  
  cout << "Forwarder::syncMessageForward() thr_func wait..." << endl;
  
  int fd = openSocket(that->forwardHostname, that->forwardPort);
  if (fd < 0) {
    cerr << "Error: socket [" << that->forwardHostname << ", "
	 << that->forwardPort << "] cannot open!" << endl;
  }
  sendSocket(fd, http_msg);
  std::string recv_msg;
  recvSocket(fd, that->soapMsgFromOtherIf);
  close(fd);
  
  removeHTTPHeader(that->soapMsgFromOtherIf);
  int unlock_status = pthread_mutex_unlock(&(that->_mt));
  
  // mutex ̉
  if (0 != unlock_status) {
    cerr << "[[Error!!]] void* Forwarder::localSyncThrFunc(void*)" << endl;
    cerr << "            pthread_mutex_unlock() failed." << endl;
    cerr.flush();
    return NULL;
  }
  
  return NULL;
}

//----- bZ[W] -----//
bool
Forwarder::execSyncMessageForward()
{
  cout << "[execSyncMessageForward]" << endl;
  cout << "forwardHostname = " << forwardHostname << endl;
  cout << "forwardPort = " << forwardPort << endl;
  cout << "forwardCGILocation = " << forwardCGILocation << endl;
  cout << "operationName = " << operationName << endl;
  
  struct timeval tv_start, tv_end;
  struct timezone tz;
  
  gettimeofday(&tv_start, &tz);
  cerr << "tv_start.tv_sec  = " << tv_start.tv_sec << endl;
  cerr.flush();
  
  bool ttlHeaderExist;
  std::string ttlType;
  getTTLInfo(soapMsgFromMsgDrv, ttlHeaderExist, ttl, ttlType);
  // for debug
  cerr << "Forwarder::execSyncMessageForward() ttl=(" << ttl << ")" << endl;
  
  // for debug
  cout << "Forwarder::execSyncMessageForward() create thread..." << endl;
  
  //ړȈsȂXbh𐶐
  pthread_t thread_t;
  int status = pthread_create(&thread_t, NULL, syncMessageForward, this);
  
  // for debug
  cout << "Forwarder::execSyncMessageForward() " 
       << "thread create done... status=(" << status << ")" << endl;
  
  //^C}[IuWFNg
  OpenSOAP::Timeout timeout(ttl, thread_t);
  //^CAEgĎJn
  timeout.start();
  
  //ړȈɃXbh
  status = pthread_join(thread_t, NULL);
  
  if (status != 0) {
    cerr << "Forwarder::execSyncMessageForward() pthread join error" << endl;
  }
  
  //ړIIɂ^C}[~
  timeout.finish();
  
  // for debug
  cout << "Forwarder::execSyncMessageForward() bye..." << endl;
  cout << "timeout.targetThreadIsCanceled = " 
       << timeout.targetThreadIsCanceled() << endl;
  
  gettimeofday(&tv_end, &tz);
  cerr << "tv_end.tv_sec  = " << tv_end.tv_sec << endl;
  cerr.flush();
  
  return timeout.targetThreadIsCanceled();
}

bool
Forwarder::setFwdHostInfo()
{
  cout << "----- setFwdHostInfo() -----" << endl;
  cout << "localhostName = " << localhostName << endl;
  
  unsigned int i = 0;
  for (i = 0; i < hostnames.size() - 1; i++) {
    if (localhostName == hostnames[i]) {
      forwardHostname = hostnames[i + 1];
      forwardPort = ports[i + 1];
      forwardCGILocation = cgiLocations[i + 1];
      break;
    }
  }
  
  if (localhostName == *(hostnames.end() - 1)) {
    return false;
  }
  
  if (i == hostnames.size() - 1) {
    forwardHostname = hostnames[0];
    forwardPort = ports[0];
    forwardCGILocation = cgiLocations[0];
  }
  
  
  cout << "forwardHostname = " << forwardHostname << endl;
  cout << "forwardPort = " << forwardPort << endl;
  cout << "forwardCGILocation = " << forwardCGILocation << endl;
  cout << "----- setFwdHostInfo() -----" << endl;
  return true;
}// end of setFwdHostInfo()

//----- 2 ڈȍ~̃zXgł, received_path ǉ ------//
bool
Forwarder::addReceivedPath () {
  for (unsigned i = 0; i < hostnames.size(); i++) {
    if (hostnames[i] == localhostName) {
      attachReceivedPath(soapMsgFromMsgDrv, 
			 fileIDOfMsgFromMsgDrv, backwardURL);
      return true;
    }
  }// loop of for (i)
  
  return false;
}// end of attatchReceivePath()

//----- TTL Table 烁bZ[W ID ɑΉ -----/
//----- backward_path T -----/
bool
Forwarder::refToTTLTable(const std::string& msgIDContent, 
			 std::string& backwardPath)
{
  ChannelSelector *ttlTableRefChnlSelector 
    = new TTLTableRefChnlSelector();
  ChannelDescriptor ttlTableRefChnlDesc;
  
  if (0 != ttlTableRefChnlSelector->open(ttlTableRefChnlDesc)) {
    cerr << "Forwarder::refToTTLTable: ttlTableRefChnl open failed" << endl;
    return false;
  }
  
  cout << "send to ttl table" << endl;
  if (0 > ttlTableRefChnlDesc.write(msgIDContent)) {
    cerr << "Forwarder::refToTTLTable: write failed to TTLTableRef" << endl;
    return false;
  }
  
  
  cout << "recv from ttl table" << endl;  
  if (0 > ttlTableRefChnlDesc.read(backwardPath)) {
    cerr << "Forwarder::refToTTLTable: read failed from TTLTableRef" << endl;
    return false;
  }
  cerr << "** MsgDrv ** READ From TTLTable=[" 
       << backwardPath << "]" << endl;
  
  ttlTableRefChnlSelector->close(ttlTableRefChnlDesc);
  delete ttlTableRefChnlSelector;
  
  
  return true;
}// end of refToTTLTable()

//----- X|X߂s -----//
bool
Forwarder::execBackwarding()
{
  cerr << "----- Forwarder::execBackwarding begin -----" << endl;
  std::string backwardPath;
  std::string msgIDContent = getMsgIDContent(soapMsgFromMsgDrv);
  bool hasBackwardPath = refToTTLTable(msgIDContent, backwardPath);
  
  if (hasBackwardPath) {
    // TTL e[u backward pXꍇ֖͂߂D
    std::string prtcl;
    analyseFwdHostInfo(backwardPath, prtcl, forwardHostname,
		       forwardPort, forwardCGILocation);
    
    bool targetThreadIsCanceled = execSyncMessageForward();    
    if (targetThreadIsCanceled) {
      // ^CAEg thr_func LZďIꍇB
      soapMsgFromOtherIf = makeTimeoutMessage();
    }
    
  }
  cerr << "----- Forwarder::execBackwarding end   -----" << endl;
  return true;
}// end of execBackwarding()

void
Forwarder::run()
{
  
  cerr << "FORWARDER::run read ready." << endl;
  
  if (0 > chnlDesc_.read(fileIDOfMsgFromMsgDrv)) {
    cerr << TAG_ERR << "Forwarder::run: read failed" << endl;
    return;
  }

  cerr << "$$ Forwarder $$ READ From MsgDrv=[" 
       << fileIDOfMsgFromMsgDrv << "]" << endl;

  soapMsgFromMsgDrv = revertFromFileID(fileIDOfMsgFromMsgDrv);
  
  // forwarder  operationNames, ttl nĂ炦΁A
  // 
  cerr << "extract Operation and Parameter" << endl;
  std::vector<std::string> operationNames;
  std::vector<std::string> paramNames;
  std::vector<std::string> paramValues;

  bool operationExist 
    = extractOperationAndParam(soapMsgFromMsgDrv, operationNames, 
			       paramNames, paramValues);
  // ܂ł̏͂ȂB

  // SOAP Header ]ɊւW߂
  bool fwdPathExist;
  std::vector<std::string> fwdPath;
  bool hopCountExist;
  unsigned int hopCount;
  getFwdInfoFromHeader(soapMsgFromMsgDrv, fwdPathExist, fwdPath,
		       hopCountExist, hopCount);
  
  // X|XbZ[Wǂ𒲂ׂ
  bool itIsResMsg = isItResponseMessage(soapMsgFromMsgDrv);
  
  bool targetThreadIsCanceled = false;

  if (itIsResMsg) {
    execBackwarding();
  }
  if (defaultForwarding || fwdPathExist 
      || hopCountExist || hopCountExist) {
    // tH[hꍇB
    ttl = 5;
    if (fwdPathExist || hopCountExist) {
      analyseFwdHostInfo(fwdPath, prtcls, hostnames, ports, cgiLocations);
      addReceivedPath();
      updateBackwardPath(soapMsgFromMsgDrv, backwardURL);
      if (hopCountExist) {
	decreaseHopCount(soapMsgFromMsgDrv);
      }
      else {
	// if <hopcount> is specified, use default forward
	setFwdHostInfo();
      }
    }
    targetThreadIsCanceled = execSyncMessageForward();
  }// end of if (defaultForwarding)
  if (!defaultForwarding && operationExist) {
    // Iy[VȂĂtH[hȂ΂ꍇ
    soapMsgFromOtherIf = makeOperationNotFoundMessage();
  }
  
  if (targetThreadIsCanceled) {
    // ^CAEg thr_func LZďIꍇB
    soapMsgFromOtherIf = makeTimeoutMessage();
  }
  
  fileIDOfMsgFromOtherIf = convertToFileID(soapMsgFromOtherIf);
  
  cerr << "$$ Forwarder $$ WRITE To MsgDrv=[" 
       << fileIDOfMsgFromOtherIf << "]" << endl;
  
  if (0 > chnlDesc_.write(fileIDOfMsgFromOtherIf)) {
    cerr << TAG_ERR << "Forwarder::run: write failed" << endl;
    return;
  }
  
}






