/* -*- mode: c++; -*-
 *-----------------------------------------------------------------------------
 * $RCSfile: FwdInfoFunc.cpp,v $
 *
 * See Copyright for the status of this software.
 *
 * The OpenSOAP Project
 * http://opensoap.jp/
 *-----------------------------------------------------------------------------
 */

#include <algorithm>
#include <iostream>
#include <string>

#include <unistd.h>
#include <netdb.h>

/* libxml2 */
#include <tree.h>
#include <parser.h>

#include "FwdInfoFunc.h"

#include "SrvConfAttrHandler.h"

using namespace OpenSOAP;

//----- Forward ftHg̃zXg̏擾 -----//
bool
getForwarderInfo(std::string& forwardHostname, unsigned int& forwardPort, 
		 std::string& forwardCGILocation)
{
  SrvConfAttrHandler ssmlAttrHndl;
  std::vector<std::string> values;
  std::string query;
  
  query = "/forwarder/hostname=?";
  if (0 > ssmlAttrHndl.queryXml(query, values)) {
    cerr << "*** ERROR ***" << endl;
    return false;
  }
  
  cerr << "### RESULT ###" << endl;
  cerr << "FwdInfoFunc::getForwarderInfo:" 
       << "values[0] = " << values[0] << endl;
  forwardHostname = values[0];
  
  values.clear();
  query = "/forwarder/port=?";
  if (0 > ssmlAttrHndl.queryXml(query, values)) {
    cerr << "*** ERROR ***" << endl;
    return false;
  }
  
  cerr << "### RESULT ###" << endl;
  cerr << "FwdInfoFunc::getForwarderInfo:" 
       << "values[0] = " << values[0] << endl;
  forwardPort = atoi(values[0].c_str());
  
  values.clear();
  query = "/forwarder/cgi_location=?";
  if (0 > ssmlAttrHndl.queryXml(query, values)) {
    cerr << "*** ERROR ***" << endl;
    return false;
  }
  
  cerr << "### RESULT ###" << endl;
  cerr << "FwdInfoFunc::getForwarderInfo:" 
       << "values[0] = " << values[0] << endl;
  forwardCGILocation = values[0];
  
  std::string officialHostname = getLocalhostName();
  
  if (officialHostname == forwardHostname) {
    cerr << "Official hostname equals to forward hostname." << endl;
    cerr << "This will cause loop problem !!" << endl;
    return false;
  }
  
  return true;
}// end of getForwarderInfo(std::string&, unsigned int&, std::string&)


//----- Forward ꂽbZ[WԂĂ炤 URL 擾 -----//
bool
getBackwardURLs(std::vector<std::string>& backwardURLs)
{
  SrvConfAttrHandler ssmlAttrHndl;
  std::vector<std::string> values;
  std::string query;

  std::string protocol;
  std::string port;
  std::string cgiLocation;
  
  // vgR̎擾
  query = "/backward/protocol=?";
  values.clear();
  if (0 > ssmlAttrHndl.queryXml(query, values)) {
    cerr << "*** ERROR ***" << endl;
    return false;
  }
  
  cerr << "### RESULT ###" << endl;
  cerr << "FwdInfoFunc::getBackwardURLs:" 
       << "values[0] = " << values[0] << endl;
  protocol = values[0];
  
  // |[gԍ̎擾
  query = "/backward/port=?";
  values.clear();
  if (0 > ssmlAttrHndl.queryXml(query, values)) {
    cerr << "*** ERROR ***" << endl;
    cerr << "Use default port: 80" << endl;
    port = "80";
  }
  else {
    cerr << "### RESULT ###" << endl;
    cerr << "FwdInfoFunc::getBackwardURLs:" 
	 << "values[0] = " << values[0] << endl;
    port = values[0];
  }
  
  
  // CGI ւ̃pX̎擾
  query = "/backward/cgi_location=?";
  values.clear();
  if (0 > ssmlAttrHndl.queryXml(query, values)) {
    cerr << "*** ERROR ***" << endl;
    return false;
  }
  
  cerr << "### RESULT ###" << endl;
  cerr << "FwdInfoFunc::getBackwardURLs:" 
       << "values[0] = " << values[0] << endl;
  cgiLocation = values[0];
  
  
  // get hostname
  query = "/backward/hostname=??";
  values.clear();
  if (0 > ssmlAttrHndl.queryXml(query, values)) {
    cerr << "*** ERROR ***" << endl;
    return false;
  }
  
  cerr << "### RESULT ###" << endl;
  cerr << "FwdInfoFunc::getBackwardURLs:" << endl;
  std::vector<std::string>::iterator iter;
  for(iter = values.begin(); iter != values.end(); iter++) {
    cerr << "values = " << *iter << endl;
  }
  
  // make backwardURLs
  if (port == "80") {
    for(iter = values.begin(); iter != values.end(); iter++) {    
      backwardURLs.push_back(protocol + "://" 
			     + *iter + cgiLocation);
    }
  }
  else {
    for(iter = values.begin(); iter != values.end(); iter++) {
      backwardURLs.push_back(protocol + "://" 
			    + *iter + ":" + port + cgiLocation);
    }
  }
  for (std::vector<std::string>::iterator iter = backwardURLs.begin();
       iter != backwardURLs.end(); iter++) {
    cerr << "backwarkURLs = [" << *iter << "]" <<endl;
  }
  return true;
}// end of getBackwardURLs(std::vector<std::string>&)



//-- SOAP Header ɂ forward_path ̓e؂蕪 --//
bool
analyseFwdHostInfo(std::vector<std::string> fwdPath,
		   std::vector<std::string>& prtcls,
		   std::vector<std::string>& hostnames,
		   std::vector<unsigned int>& ports,
		   std::vector<std::string>& cgiLocations)
{
  prtcls.clear();
  hostnames.clear();
  ports.clear();
  cgiLocations.clear();
  
  std::string::iterator findPoint;
  std::string* tmp;
  std::string* lastStr;
  char delim;
  
  for (std::vector<std::string>::iterator iter = fwdPath.begin();
      iter != fwdPath.end(); iter++) {
    /* vgR擾 */
    delim = ':';
    findPoint = find(iter->begin(), iter->end(), delim);
    if (findPoint != iter->end()) {
      tmp = new std::string(iter->begin(), findPoint);
      cout << "*tmp = " << *tmp << endl;
      prtcls.push_back(*tmp);
      if (findPoint < iter->end() - 1) {
	lastStr = new std::string(findPoint + 1, iter->end());
	cout << "*lastStr = " << *lastStr << endl;
      }
      else {
	return false;
      }
    }
    else {
      return false;
    }
    
    /* zXgƃ|[gԍ擾 */
    // '/'  2 
    for (unsigned int i = 0; i < 2; i++) {
      delim = '/';
      findPoint = find(lastStr->begin(), lastStr->end(), delim);
      if (findPoint != lastStr->end()) {
	if (findPoint < lastStr->end() - 1) {
	  lastStr = new std::string(findPoint + 1, lastStr->end());
	  cout << "*lastStr = " << *lastStr << endl;
	}
	else {
	  return false;
	}
      }
      else {
	return false;
      }
    }// loop of for (unsigned int i = 0; i < 2; i++) 
    
    
    bool useDefaultPort = true;
    delim = ':';
    findPoint = find(lastStr->begin(), lastStr->end(), delim);
    if (findPoint != lastStr->end()) {
      tmp = new std::string(lastStr->begin(), findPoint);
      cout << "*tmp = " << *tmp << endl;
      useDefaultPort = false;
      hostnames.push_back(*tmp);
      if (findPoint < lastStr->end() - 1) {
	lastStr = new std::string(findPoint + 1, lastStr->end());
	cout << "*lastStr = " << *lastStr << endl;
      }
      else {
	return false;
      }
    }
    else {
      cout << "using default port: 80" << endl;
      ports.push_back(80);
    }
    
    delim = '/';
    findPoint = find(lastStr->begin(), lastStr->end(), delim);
    if (findPoint != lastStr->end()) {
      tmp = new std::string(lastStr->begin(), findPoint);
      cout << "*tmp = " << *tmp << endl;
      
      if (useDefaultPort) {
	hostnames.push_back(*tmp);
      }
      else {
	ports.push_back((unsigned int)atoi(tmp->c_str()));
      }
      if (findPoint < lastStr->end() - 1) {
	/* CGI ւ̃pX擾 */
	lastStr = new std::string(findPoint, lastStr->end());
	cgiLocations.push_back(*lastStr);
	cout << "cgiLocation = " << *lastStr << endl;
      }
      else {
	return false;
      }
    }
    else {
      return false;
    }
    
    
  }// loop of (iter)
  cout << "----- fwdPathInfo -----" << endl;
  for (unsigned int i = 0; i < fwdPath.size(); i++) {
    cout << "prtcls[" << i << "] = " << prtcls[i] << endl;
    cout << "hostnames[" << i << "] = " << hostnames[i] << endl;
    cout << "ports[" << i << "] = " << ports[i] << endl;
    cout << "cgiLocations[" << i << "] = " << cgiLocations[i] << endl;
    cout << endl;
  }
  cout << "----- fwdPathInfo -----" << endl;
  
  return true;
}// end of analyseFwdHostInfo()

//-- SOAP Header ɂ forward_path ̓e؂蕪 --//
extern
bool
analyseFwdHostInfo(std::string fwdPath,
		   std::string& prtcl, std::string& hostname,
		   unsigned int& port, std::string& cgiLocation)
{
  
  prtcl = "";
  hostname = "";
  port = 80;
  cgiLocation = "";
  
  std::string::iterator findPoint;
  std::string* tmp;
  std::string* lastStr;
  char delim;
  
  /* vgR擾 */
  delim = ':';
  findPoint = find(fwdPath.begin(), fwdPath.end(), delim);
  if (findPoint != fwdPath.end()) {
    tmp = new std::string(fwdPath.begin(), findPoint);
    cout << "*tmp = " << *tmp << endl;
    prtcl = *tmp;
    if (findPoint < fwdPath.end() - 1) {
      lastStr = new std::string(findPoint + 1, fwdPath.end());
      cout << "*lastStr = " << *lastStr << endl;
    }
    else {
      return false;
    }
  }
  else {
    return false;
  }
  
  /* zXgƃ|[gԍ擾 */
  // '/'  2 
  for (unsigned int i = 0; i < 2; i++) {
    delim = '/';
    findPoint = find(lastStr->begin(), lastStr->end(), delim);
    if (findPoint != lastStr->end()) {
      if (findPoint < lastStr->end() - 1) {
	lastStr = new std::string(findPoint + 1, lastStr->end());
	cout << "*lastStr = " << *lastStr << endl;
      }
	else {
	  return false;
	}
    }
      else {
	return false;
      }
  }// loop of for (unsigned int i = 0; i < 2; i++) 
  
  
  bool useDefaultPort = true;
  delim = ':';
  findPoint = find(lastStr->begin(), lastStr->end(), delim);
  if (findPoint != lastStr->end()) {
    tmp = new std::string(lastStr->begin(), findPoint);
    cout << "*tmp = " << *tmp << endl;
    useDefaultPort = false;
    hostname = *tmp;
    if (findPoint < lastStr->end() - 1) {
      lastStr = new std::string(findPoint + 1, lastStr->end());
      cout << "*lastStr = " << *lastStr << endl;
    }
    else {
      return false;
    }
  }
  else {
    cout << "using default port: 80" << endl;
    port = 80;
  }
  
  delim = '/';
  findPoint = find(lastStr->begin(), lastStr->end(), delim);
  if (findPoint != lastStr->end()) {
    tmp = new std::string(lastStr->begin(), findPoint);
    cout << "*tmp = " << *tmp << endl;
    
    if (useDefaultPort) {
      hostname = *tmp;
    }
    else {
      port = (unsigned int)atoi(tmp->c_str());
    }
    if (findPoint < lastStr->end() - 1) {
      /* CGI ւ̃pX擾 */
      lastStr = new std::string(findPoint, lastStr->end());
      cgiLocation = *lastStr;
      cout << "cgiLocation = " << *lastStr << endl;
    }
    else {
      return false;
    }
  }
  else {
    return false;
  }
  
  
  
  cout << "----- fwdPathInfo -----" << endl;
  cout << "prtcl = " << prtcl << endl;
  cout << "hostname = " << hostname << endl;
  cout << "port = " << port << endl;
  cout << "cgiLocation = " << cgiLocation << endl;
  cout << endl;
  
  cout << "----- fwdPathInfo -----" << endl;
  
  return true;
}





