/* -*- mode: c++; -*-
 *-----------------------------------------------------------------------------
 * $RCSfile: InetSocketSrvDrv.cpp,v $
 *
 * See Copyright for the status of this software.
 *
 * The OpenSOAP Project
 * http://opensoap.jp/
 *-----------------------------------------------------------------------------
 */

#include <string>
#include <vector>

#include <sys/time.h>
#include <unistd.h>

/* SSML Information Function library */
#include "SSMLInfoFunc.h"
#include "SSMLInfo.h"

/* server common library */
#include "ServerCommon.h"
#include "ChannelDescriptor.h"
#include "FIFOSrvChnlSelector.h"
#include "ResSpoolPushChnlSelector.h"
#include "FwdQueuePushChnlSelector.h"
#include "FwdInfoFunc.h"

/* srvdrv library */
#include "InetSocketSrvDrv.h"

/* common library */
#include "connection.h"
#include "FileIDFunc.h"
#include "SOAPMessage.h"
#include "StringUtil.h"
#include "Timeout.h"



using namespace OpenSOAP;

//----- Constructor -----//
InetSocketSrvDrv::InetSocketSrvDrv(const ChannelDescriptor& chnlDesc)
  : chnlDesc_(chnlDesc)
{
  cerr << "SrvDrv::SrvDrv" << endl;
  
  pthread_mutex_init(&_mt, NULL);
  downside_timeout = 5;
  _isAsync = false;

  bool localURLExist = getBackwardURLs(localURLs);
  if (!localURLExist) {
    cerr << "local URL not Exist" << endl;
  }
  
}// end of InetSocketSrvDrv ()

//----- Destructor -----//
InetSocketSrvDrv::~InetSocketSrvDrv ()
{
  cerr << "SrvDrv::~SrvDrv" << endl;
  pthread_mutex_destroy(&_mt);
}// end of ~InetSocketSrvDrv ()

//----- [J̓Iy[Vs -----//
bool
InetSocketSrvDrv::execLocalSyncOperation()
{
  
  struct timeval tv_start, tv_end;
  struct timezone tz;
  
  gettimeofday(&tv_start, &tz);
  cerr << "tv_start.tv_sec  = " << tv_start.tv_sec << endl;
  cerr.flush();
  
  // for debug
  cerr << "InetSocketSrvDrv::execLocalSyncOperation() ttl=(" << ttl << ")" << endl;
  
  // for debug
  cout << "InetSocketSrvDrv::execLocalSyncOperation() create thread..." << endl;
  
  //ړȈsȂXbh𐶐
  pthread_t thread_t;
  int status = pthread_create(&thread_t, NULL, localSyncThrFunc, this);
  
  // for debug
  cout << "InetSocketSrvDrv::execLocalSyncOperation() "
       << "thread create done... status=(" << status << ")" << endl;
  
  //^C}[IuWFNg
  OpenSOAP::Timeout timeout(ttl, thread_t);
  //^CAEgĎJn
  timeout.start();
  
  //ړȈɃXbh
  status = pthread_join(thread_t, NULL);
  
  if (status != 0) {
    cerr << "InetSocketSrvDrv::execLocalSyncOperation() pthread join error" << endl;
  }
  
  //ړIIɂ^C}[~
  timeout.finish();
  
  // for debug
  cout << "InetSocketSrvDrv::execLocalSyncOperation() bye..." << endl;
  cout << "timeout.targetThreadIsCanceled = " 
       << timeout.targetThreadIsCanceled() << endl;
  
  gettimeofday(&tv_end, &tz);
  cerr << "tv_end.tv_sec  = " << tv_end.tv_sec << endl;
  cerr.flush();
  
  return timeout.targetThreadIsCanceled();
}

//----- [JœIy[Vs߂̃Xbh  -----//
void* 
InetSocketSrvDrv::localSyncThrFunc (void* arg) 
{
  InetSocketSrvDrv* that = (InetSocketSrvDrv*)arg;
  
  // for debug  
  cout << "InetSocketSrvDrv::localSyncThrFunc: thr_func wait..." << endl;

  std::string msgFromService;
  
  // T[rX̕ԓ҂
  std::string hostname = that->ssmlInfo.getHostname();
  unsigned int port = that->ssmlInfo.getPort();
  
  std::cerr << "InetSocketSrvDrv::localSyncThrFunc: "
	    << "hostname = " << hostname << std::endl;
  std::cerr << "InetSocketSrvDrv::localSyncThrFunc: "
	    << "port = " << port << std::endl;
  
  int fd = openSocket(hostname, port);
  std::cerr << "InetSocketSrvDrv::localSyncThrFunc: "
	    << "fd = " << fd << endl;
  if (0 > fd) {
    cerr << "InetSocketSrvDrv::localSyncThrFunc: socket open failed" << endl;
    msgFromService 
      = createSOAPFaultMessage("Service.ConnectionNotEstablished",
			       "connection not established",
			       "server",
			       that->ssmlInfo.getOperationName() 
			       + " not respond");
    that->fileIDOfMsgFromService = convertToFileID(msgFromService);    
    return NULL;
  }
  
  int lock_status = pthread_mutex_lock (&(that->_mt));
  if (0 != lock_status) {
    cerr << "[[Error!!]] void* InetSocketSrvDrv::localSyncThrFunc(void*)" << endl;
    cerr << "            pthread_mutex_lock() failed." << endl;
    cerr.flush();
    return NULL;
  }
  std::string msgFromMsgDrv = revertFromFileID(that->fileIDOfMsgFromMsgdrv);
  
  std::string msgToService = attachHeaderForService(msgFromMsgDrv);
  
  if (0 > sendSocket(fd, msgToService)) {
    cerr << "InetSocketSrvDrv::localSyncThrFunc "
	 << "write failed to InetSoeketService" << endl;
    return NULL;
  }
  cerr << "** SrvDrv ** WRITE TO InetSocketSrv=[" 
       << msgToService << "]" << endl;
  
  
  if (0 > recvSocket(fd, msgFromService)) {
    cerr << "InetSocketSrvDrv::localSyncThrFunc "
	 << "read failed from InetSoeketService" << endl;
    return NULL;
  }
  cerr << "** SrvDrv ** READ From InetSocketSrv=[" 
       << msgFromService << "]" << endl;
  
  if (msgFromService.length() == 0) {
    cerr << "Received String from Service is Empty" << endl;
    msgFromService = makeResponseIsEmptyMessage();
  }
  msgFromService = removeHeaderForService(msgFromService);
  
  // attach msg ID in the case of async response
  std::string msgIDContent = getMsgIDContent(msgFromMsgDrv);
  if (msgIDContent.length() > 0) {
    attachOrUpdateMsgID(msgFromService, msgIDContent);
  }
  
  // attach backward path, 
  // if this server is the final receiver of async fwd message
  bool itIsAsyncFwdFinalReceiver = 
    isItAsyncFwdFinalReceiver(msgFromMsgDrv, that->localURLs);
  if (itIsAsyncFwdFinalReceiver) {
    attachBackwardPathToResponse(msgFromService, msgFromMsgDrv);
    attachTTLToResponse(msgFromService, msgFromMsgDrv);
  }
  
  that->fileIDOfMsgFromService = convertToFileID(msgFromService);
  
  int unlock_status = pthread_mutex_unlock (&(that->_mt));
  if (0 != unlock_status) {
    cerr << "[[Error!!]] void* InetSocketSrvDrv::localSyncThrFunc(void*)" << endl;
    cerr << "            pthread_mutex_unlock() failed." << endl;
    cerr.flush();
    return NULL;
  }
  
  cerr << "InetSocketSrvDrv::localSyncThrFunc: close socket" << endl;
  close(fd);
  
  return NULL;
}

void
InetSocketSrvDrv::run()
{
  
  cerr << "SVR: read ready." << endl;
  
  if (0 > chnlDesc_.read(fileIDOfMsgFromMsgdrv)) {
    cerr << TAG_ERR << "SrvDrv::run: read failed" << endl;
    return;
  }
  
  cerr << "$$ SrvDrv $$ READ From MsgDrv=[" 
       << fileIDOfMsgFromMsgdrv << "]" << endl;
  
  soapMsgFromMsgdrv = revertFromFileID(fileIDOfMsgFromMsgdrv);
  
  cerr << "extract Operation and Parameter" << endl;
  std::vector<std::string> operation_names;
  std::vector<std::string> param_names;
  std::vector<std::string> param_values;
  
  
  extractOperationAndParam (soapMsgFromMsgdrv, operation_names, 
                            param_names, param_values);
  
  ssmlInfo.setOperationName(operation_names[0]);
  bool operationExist = getSSMLInfo(ssmlInfo);
  
  bool soapMsgIsValid = false;
  bool ttlHeaderExist = false;
  std::string ttlType;
  
  soapMsgIsValid 
    = getTTLInfo(soapMsgFromMsgdrv, ttlHeaderExist, ttl, ttlType);
  cerr << "InetSocketSrvDrv::run:ttlHeaderExist = " << ttlHeaderExist << endl;
  cerr << "InetSocketSrvDrv::run:ttl = " << ttl << endl;
  
  if (!soapMsgIsValid) {
    std::string invalidReqErrMsg = makeInvalidRequestMessage();
    std::string fileIDOfInvalidReqErrMsg = convertToFileID(invalidReqErrMsg);
    chnlDesc_.write(fileIDOfInvalidReqErrMsg); 
    return;
  }
  
  if (!ttlHeaderExist) {
    if (_isAsync) {
      ttl = ssmlInfo.getAsyncTTL();
    }
    else {
      ttl = ssmlInfo.getSyncTTL();
    }
  }
  
  bool targetThreadIsCanceled = false;
   
  std::vector<std::string> soapServerEndPoints;
  getBackwardURLs(soapServerEndPoints);
  
  _itIsAsyncFwdFinalReceiver 
    = isItAsyncFwdFinalReceiver(soapMsgFromMsgdrv, soapServerEndPoints);
  
  if (operationExist) {
    targetThreadIsCanceled = execLocalSyncOperation();
  }
  
  if (targetThreadIsCanceled) {
    // ^CAEg thr_func LZďIꍇB
    soapMsgFromService = makeTimeoutMessage();
    fileIDOfMsgFromService = convertToFileID(soapMsgFromService);
    if (0 > chnlDesc_.write(fileIDOfMsgFromService)) {
      cerr << TAG_ERR << "SrvDrv::run: write failed" << endl;
      return;
    }
  }
  
  if (_isAsync) {
    if (_itIsAsyncFwdFinalReceiver) {
      // 񓯊]̏ꍇ̓tH[hL[ɑB
      cerr << "Response of Async Fwd is send to FwdQueue." << endl;
      asyncFwdResponse();
    }
    else {
      // 񓯊̏ꍇ̓X|XXv[ɑB
      cerr << "Response of Async is send to ResSpool." << endl;
      asyncResponse();
    }
  }
  else {
    // ̏ꍇ̓bZ[WhCoɃX|XԂB
    if (0 > chnlDesc_.write(fileIDOfMsgFromService)) {
      cerr << TAG_ERR << "SrvDrv::run: write failed" << endl;
      return;
    }
  }
  
}


void 
InetSocketSrvDrv::setAsyncFlag(bool isAsync)
{
  _isAsync = isAsync;
}

bool
InetSocketSrvDrv::asyncResponse()
{
  // Response Spool Ɍʂ𑗂B
  
  ChannelDescriptor resSpoolPushChnlDesc;
  ChannelSelector* resSpoolPushChnlSelector;
  std::vector<std::string> backwardURLs;
  getBackwardURLs(backwardURLs);
  if (isItAsyncFwdFinalReceiver(soapMsgFromMsgdrv, backwardURLs)) {
    resSpoolPushChnlSelector = new FwdQueuePushChnlSelector();
  }
  else {
    resSpoolPushChnlSelector = new ResSpoolPushChnlSelector();
  }
  if (0 != resSpoolPushChnlSelector->open(resSpoolPushChnlDesc)) {
    cerr << "InetSocketSrvDrv::asyncResponse: Chnl open failed" << endl;
    return false;
  }
  
  cout << "send to response spool" << endl;
  if (0 > resSpoolPushChnlDesc.write(fileIDOfMsgFromService)) {
    cerr << "InetSocketSrvDrv::asyncResponse: write failed to Service" << endl;
    return false;
  }
  
  std::string fileIDOfMsgFromResSMgr;
  if (0 > resSpoolPushChnlDesc.read(fileIDOfMsgFromResSMgr)) {
    cerr << "InetSocketSrvDrv::asyncResponse: read failed from Service" << endl;
    return false;
  }
  
  cerr << "** InetSocketSrvDrv ** READ From Service=[" 
       << fileIDOfMsgFromResSMgr << "]" << endl;
  
  cout << "send to request queue" << endl;
  
  resSpoolPushChnlSelector->close(resSpoolPushChnlDesc);
  delete resSpoolPushChnlSelector;
  
  
  std::string msgToReqQMgr = "success";
  std::string fileIDOfMsgToReqQMgr = convertToFileID(msgToReqQMgr);
  if (0 > chnlDesc_.write(fileIDOfMsgToReqQMgr)) {
    cerr << TAG_ERR << "InetSocketSrvDrv:asyncFwdResponse: write failed" << endl;
    return false;
  }
  
  return true;
}

//----- 񓯊]̍ŏIT[o[̏ꍇ̃X|Xs -----//
bool
InetSocketSrvDrv::asyncFwdResponse()
{
  // 񓯊]̍Ō̃T[o[̏ꍇ
  // Fwd Queue Ɍʂ𑗂B
  
  ChannelDescriptor fwdQueuePushChnlDesc;
  ChannelSelector* fwdQueuePushChnlSelector = new FwdQueuePushChnlSelector();
  if (0 != fwdQueuePushChnlSelector->open(fwdQueuePushChnlDesc)) {
    cerr << "InetSocketSrvDrv::asyncFwdResponse:: Chnl open failed" << endl;
    return false;
  }
  
  cout << "send to response spool" << endl;
  if (0 > fwdQueuePushChnlDesc.write(fileIDOfMsgFromService)) {
    cerr << "InetSocketSrvDrv::asyncFwdResponse: write failed to Service" << endl;
    return false;
  }
  
  std::string fileIDOfMsgFromResSMgr;
  if (0 > fwdQueuePushChnlDesc.read(fileIDOfMsgFromResSMgr)) {
    cerr << "InetSocketSrvDrv::asyncFwdResponse: read failed from Service" << endl;
    return false;
  }
  
  cerr << "** InetSocketSrvDrv ** READ From Service=[" 
       << fileIDOfMsgFromResSMgr << "]" << endl;
  
  cout << "send to request queue" << endl;
  
  fwdQueuePushChnlSelector->close(fwdQueuePushChnlDesc);
  delete fwdQueuePushChnlSelector;
  
  
  std::string msgToReqQMgr = "success";
  std::string fileIDOfMsgToReqQMgr = convertToFileID(msgToReqQMgr);
  if (0 > chnlDesc_.write(fileIDOfMsgToReqQMgr)) {
    cerr << TAG_ERR << "InetSocketSrvDrv::asyncFwdResponse: write failed" << endl;
    return false;
  }
  
  return true;
}

