/* -*- mode: c++; -*-
 *-----------------------------------------------------------------------------
 * $RCSfile: MsgDrv.cpp,v $
 *
 * See Copyright for the status of this software.
 *
 * The OpenSOAP Project
 * http://opensoap.jp/
 *-----------------------------------------------------------------------------
 */

#include <iostream>
#include <sys/time.h>
#include <unistd.h>

#include "ServerCommon.h"
#include "ChannelDescriptor.h"
#include "SrvDrvChnlSelector.h"
#include "ForwarderChnlSelector.h"
#include "FwdQueuePushChnlSelector.h"
#include "ReqQueuePushChnlSelector.h"
#include "ResSpoolPopChnlSelector.h"
#include "ResSpoolPushChnlSelector.h"
#include "TTLTablePushChnlSelector.h"
#include "TTLTableRefChnlSelector.h"
#include "MsgDrv.h"

/* common library */
#include "connection.h"
#include "FileIDFunc.h"
#include "SOAPMessage.h"
#include "Timeout.h"

/* SSML Information Manager library */
#include "SSMLInfoFunc.h"
#include "SSMLInfo.h"

/* Forwarder Information Function library */
#include "FwdInfoFunc.h"

using namespace OpenSOAP;

//----- Constructor -----//
MsgDrv::MsgDrv(const ChannelDescriptor& chnlDesc)
  : chnlDesc_(chnlDesc)
{
  cerr << "MsgDrv::MsgDrv" << endl;
  
  pthread_mutex_init(&_mt, NULL);
  downsideTimeout = 5;
  
  std::string  forwardHostname;
  unsigned int forwardPort; 
  std::string  forwardCGILocation;
  
  defaultForwarding 
    = getForwarderInfo(forwardHostname, forwardPort, 
		       forwardCGILocation);
  
  bool localURLExist = getBackwardURLs(localURLs);
  if (!localURLExist) {
    cerr << "local URL not Exist" << endl;
  }
  
}// end of MsgDrv (const ChannelDescriptor&)

//----- Destructor -----//
MsgDrv::~MsgDrv ()
{
  cerr << "MsgDrv::~MsgDrv" << endl;
  
  pthread_mutex_destroy(&_mt);
}// end of ~MsgDrv ()

//----- SrvDrv ւ̑Ms߂̃Xbh  -----//
void* 
MsgDrv::localSyncThrFunc (void* arg)
{
  MsgDrv* that = (MsgDrv*)arg;
  
  // for debug  
  cout << "MsgDrv::localSyncThrFunc() thr_func wait..." << endl;
  
  // Service Driver ̕ԓ҂
  
  ChannelDescriptor srvDrvChnlDesc;
  ChannelSelector* srvDrvChnlSelector = new SrvDrvChnlSelector(false);
  if (0 != srvDrvChnlSelector->open(srvDrvChnlDesc)) {
    cerr << "MsgDrv::run: SrvDrvChnl open failed" << endl;
    return NULL;
  }
  
  int lock_status = pthread_mutex_lock (&(that->_mt));
  if (0 != lock_status) {
    cerr << "[[Error!!]] void* MsgDrv::localSyncThrFunc(void*)" << endl;
    cerr << "            pthread_mutex_lock() failed." << endl;
    cerr.flush();
    return NULL;
  }
  
  if (0 > srvDrvChnlDesc.write(that->fileIDOfMsgFromIf)) {
    cerr << "MsgDrv::run: write failed to SrvDrv" << endl;
    return NULL;
  }
  
  if (0 > srvDrvChnlDesc.read(that->fileIDOfMsgFromSrvdrv)) {
    cerr << "MsgDrv::run: read failed from SrvDrv" << endl;
    return NULL;
  }
  cerr << "** MsgDrv ** READ From SrvDrv=[" 
       << that->fileIDOfMsgFromSrvdrv << "]" << endl;
  
  int unlock_status = pthread_mutex_unlock (&(that->_mt));
  if (0 != unlock_status) {
    cerr << "[[Error!!]] void* MsgDrv::localSyncThrFunc(void*)" << endl;
    cerr << "            pthread_mutex_unlock() failed." << endl;
    cerr.flush();
    return NULL;
  }
  
  srvDrvChnlSelector->close(srvDrvChnlDesc);
  delete srvDrvChnlSelector;
  
  that->chnlDesc_.write(that->fileIDOfMsgFromSrvdrv); 
  
  return 0;
}

//----- tH[hs߂̃Xbh  -----//
void* 
MsgDrv::fwdThrFunc (void* arg)
{
  MsgDrv* that = (MsgDrv*)arg;
  
  // for debug  
  cout << "MsgDrv::fwdThrFunc() wait..." << endl;
  
  // Forwarder ̕ԓ҂
  
  ChannelDescriptor fwderChnlDesc;
  ChannelSelector* fwderChnlSelector = new ForwarderChnlSelector();
  if (0 != fwderChnlSelector->open(fwderChnlDesc)) {
    cerr << "MsgDrv::run: FwderChnl open failed" << endl;
    return NULL;
  }
  
  int lock_status = pthread_mutex_lock (&(that->_mt));
  if (0 != lock_status) {
    cerr << "[[Error!!]] void* MsgDrv::fwdThrFunc(void*)" << endl;
    cerr << "            pthread_mutex_lock() failed." << endl;
    cerr.flush();
    return NULL;
  }
  
  if (0 > fwderChnlDesc.write(that->fileIDOfMsgFromIf)) {
    cerr << "MsgDrv::run: write failed to Fwder" << endl;
    return NULL;
  }
  
  if (0 > fwderChnlDesc.read(that->fileIDOfMsgFromSrvdrv)) {
    cerr << "MsgDrv::run: read failed from Fwder" << endl;
    return NULL;
  }
  cerr << "** MsgDrv ** READ From Fwder=[" 
       << that->fileIDOfMsgFromSrvdrv << "]" << endl;
  
  int unlock_status = pthread_mutex_unlock (&(that->_mt));
  if (0 != unlock_status) {
    cerr << "[[Error!!]] void* MsgDrv::fwdThrFunc(void*)" << endl;
    cerr << "            pthread_mutex_unlock() failed." << endl;
    cerr.flush();
    return NULL;
  }
  
  
  fwderChnlSelector->close(fwderChnlDesc);
  delete fwderChnlSelector;
  
  that->chnlDesc_.write(that->fileIDOfMsgFromSrvdrv); 
  
  return NULL;
}

//----- SrvDrv ւ̑M܂̓tH[hs -----//
void* 
MsgDrv::sendToSrvDrvOrFwd()
{
  
  struct timeval tv_start, tv_end;
  struct timezone tz;
  
  gettimeofday(&tv_start, &tz);
  cerr << "tv_start.tv_sec  = " << tv_start.tv_sec << endl;
  cerr.flush();
  
  // for debug
  cerr << "MsgDrv::sendToSrvDrvOrFwd() ttl =(" << ttl << ")" << endl;
  
  // for debug
  cout << "MsgDrv::sendToSrvDrvOrFwd() create thread..." << endl;
  
  //ړȈsȂXbh𐶐
  pthread_t thread_t;
  
  int status = 0;
  if (operationExist) {
    status = pthread_create(&thread_t, NULL, localSyncThrFunc, this);
  }
  else {
    ttl = 5;
    status = pthread_create(&thread_t, NULL, fwdThrFunc, this);
  }
  
  // this debug
  cout << "MsgDrv::sendToSrvDrvOrFwd() " 
       << "thread create done... status=(" << status << ")" << endl;
  
  //^C}[IuWFNg
  OpenSOAP::Timeout timeout(ttl, thread_t);
  //^CAEgĎJn
  timeout.start();
  
  //ړȈɃXbh
  status = pthread_join(thread_t, NULL);
  
  if (status != 0) {
    cerr << "MsgDrv::sendToSrvDrvOrFwd() pthread join error" << endl;
  }
  
  //ړIIɂ^C}[~
  timeout.finish();
  
  // for debug
  cout << "MsgDrv::sendToSrvDrvOrFwd() bye..." << endl;
  cout << "timeout.targetThreadIsCanceled = " 
       << timeout.targetThreadIsCanceled() << endl;

  gettimeofday(&tv_end, &tz);
  cerr << "tv_end.tv_sec  = " << tv_end.tv_sec << endl;
  cerr.flush();
  
  return (void*)timeout.targetThreadIsCanceled();
  
}


void
MsgDrv::popFromResponseSpool(std::string fileIDOfMsgID)
{
  ChannelDescriptor resSpoolPopChnlDesc;
  ChannelSelector* resSpoolPopChnlSelector = new ResSpoolPopChnlSelector();
  if (0 != resSpoolPopChnlSelector->open(resSpoolPopChnlDesc)) {
    cerr << "MsgDrv::popFromResponseSpool() fifo open failed" << endl;
    std::string errorMsg = makeInvalidRequestMessage();
    std::string fileIDOfErrorMsg = convertToFileID(errorMsg);
    chnlDesc_.write(fileIDOfErrorMsg);
  }
  
  cout << "send msgID to ress" << endl;
  if (0 > resSpoolPopChnlDesc.write(fileIDOfMsgID)) {
    cerr << "MsgDrv::popFromResponseSpool: "
	 << "write failed to ResSpoolPop" << endl;
    return;
  }
  cerr << "** MsgDrv ** WRITE To ResSpoolPop=[" 
       << fileIDOfMsgID << "]" << endl;
  
  
  std::string fileIDOfMsgFromResS;
  cout << "recv from ttl table" << endl;  
  if (0 > resSpoolPopChnlDesc.read(fileIDOfMsgFromResS)) {
    cerr << "MsgDrv::popFromResponseSpool: "
	 << "read failed from ResSpoolPop" << endl;
    return;
  }
  cerr << "** MsgDrv ** READ From ResSpoolPop=[" 
       << fileIDOfMsgFromSrvdrv << "]" << endl;
  
  
  resSpoolPopChnlSelector->close(resSpoolPopChnlDesc);
  delete resSpoolPopChnlSelector;
  
  chnlDesc_.write(fileIDOfMsgFromResS);   

  std::string msgFromResS;
  msgFromResS = revertFromFileID(fileIDOfMsgFromResS);
  cout << "msg from ress = " << msgFromResS << endl;
  
}

//----- TTL Table ɋL^ -----/
bool
MsgDrv::pushToTTLTable()
{
  ChannelSelector *ttlTablePushChnlSelector 
    = new TTLTablePushChnlSelector();
  ChannelDescriptor ttlTablePushChnlDesc;
  
  if (0 != ttlTablePushChnlSelector->open(ttlTablePushChnlDesc)) {
    cerr << "MsgDrv::pushToTTLTable: ttlTablePushChnl open failed" << endl;
    return false;
  }
  
  cout << "send to ttl table" << endl;
  if (0 > ttlTablePushChnlDesc.write(fileIDOfMsgFromIf)) {
    cerr << "MsgDrv::pushToTTLTable: write failed to TTLTablePush" << endl;
    return false;
  }
  
  std::string fileIDOfMsgFromTTLTable;
  cout << "recv from ttl table" << endl;  
  if (0 > ttlTablePushChnlDesc.read(fileIDOfMsgFromTTLTable)) {
    cerr << "MsgDrv::pushToTTLTable: read failed from TTLTablePush" << endl;
    return false;
  }
  cerr << "** MsgDrv ** READ From TTLTable=[" 
       << fileIDOfMsgFromTTLTable << "]" << endl;
  
  ttlTablePushChnlSelector->close(ttlTablePushChnlDesc);
  delete ttlTablePushChnlSelector;
  
  return true;
}// end of pushToTTLTable()

bool 
MsgDrv::pushToQueue(bool fwdPathExist, bool fwdHopExist,
		    bool defaultForwarding)
{
  // bZ[W ID ̕t
  std::string msgID = attachMsgID(soapMsgFromIf, fileIDOfMsgFromIf, 
				  operationNames[0]);
  
  // bZ[W ID tꂽV SOAP bZ[W̃t@C ID 擾
  fileIDOfMsgFromIf = convertToFileID(soapMsgFromIf);
  
  std::string fileIDOfMsgID = convertToFileID(msgID);
  chnlDesc_.write(fileIDOfMsgID);
  
  pushToTTLTable();
  
  ChannelDescriptor queuePushChnlDesc;
  ChannelSelector* queuePushChnlSelector;
  if (fwdPathExist || fwdHopExist) {
    bool itIsAsyncFinalReceiver 
      = isItAsyncFwdFinalReceiver(soapMsgFromIf, localURLs);
    if (itIsAsyncFinalReceiver) {
      queuePushChnlSelector = new ReqQueuePushChnlSelector();
    }
    else {
      queuePushChnlSelector = new FwdQueuePushChnlSelector();
    }
  }
  else {
    queuePushChnlSelector = new ReqQueuePushChnlSelector();
  }
  
  if (0 != queuePushChnlSelector->open(queuePushChnlDesc)) {
    cerr << "MsgDrv::pushToReqQueue: queuePushChnl open failed" << endl;
    return false;
  }
  
  cout << "send to reqq" << endl;
  if (0 > queuePushChnlDesc.write(fileIDOfMsgFromIf)) {
    cerr << "MsgDrv::pushToReqQueue: write failed to ReqQueuePush" << endl;
    return false;
  }
  
  std::string fileIDOfMsgFromQueue;
  cout << "recv from reqq" << endl;  
  if (0 > queuePushChnlDesc.read(fileIDOfMsgFromQueue)) {
    cerr << "MsgDrv::pushToReqQueue: read failed from ReqQueuePush" << endl;
    return false;
  }
  cerr << "** MsgDrv ** READ From ReqQueuePush=[" 
       << fileIDOfMsgFromQueue << "]" << endl;
  
  
  queuePushChnlSelector->close(queuePushChnlDesc);
  delete queuePushChnlSelector;
  
  chnlDesc_.write(fileIDOfMsgFromSrvdrv); 
  
  std::string msgFromQueue;
  msgFromQueue = revertFromFileID(fileIDOfMsgFromQueue);
  cout << "msg from reqq = " << msgFromQueue << endl;
  
  return false;
}

//----- TTL Table 烁bZ[W ID ɑΉ -----/
//----- backward_path T -----/
/* ̃\bh Forwarder ɂ̂łƂ瓝 */
bool
MsgDrv::refToTTLTable(const std::string& msgIDContent, 
		      std::string& backwardPath)
{
  ChannelSelector *ttlTableRefChnlSelector 
    = new TTLTableRefChnlSelector();
  ChannelDescriptor ttlTableRefChnlDesc;
  
  if (0 != ttlTableRefChnlSelector->open(ttlTableRefChnlDesc)) {
    cerr << "MsgDrv::refToTTLTable: ttlTableRefChnl open failed" << endl;
    return false;
  }
  
  cout << "send to ttl table" << endl;
  if (0 > ttlTableRefChnlDesc.write(msgIDContent)) {
    cerr << "MsgDrv::refToTTLTable: write failed to TTLTableRef" << endl;
    return false;
  }
  
  
  cout << "recv from ttl table" << endl;  
  if (0 > ttlTableRefChnlDesc.read(backwardPath)) {
    cerr << "MsgDrv::refToTTLTable: read failed from TTLTableRef" << endl;
    return false;
  }
  cerr << "** MsgDrv ** READ From TTLTable=[" 
       << backwardPath << "]" << endl;
  
  ttlTableRefChnlSelector->close(ttlTableRefChnlDesc);
  delete ttlTableRefChnlSelector;
  
  
  return true;
}// end of refToTTLTable()

//----- X|XX|XbZ[Wɂ -----//
bool
MsgDrv::pushToResponseSpool(std::string fileIDOfMsgFromTransIF)
{
  ChannelDescriptor resSpoolPushChnlDesc;
  ChannelSelector* resSpoolPushChnlSelector;
  
  resSpoolPushChnlSelector = new ResSpoolPushChnlSelector();
  
  if (0 != resSpoolPushChnlSelector->open(resSpoolPushChnlDesc)) {
    cerr << "MsgDrv::PushToResponseSpool: Chnl open failed" << endl;
    return false;
  }
  
  cout << "send to response spool" << endl;
  if (0 > resSpoolPushChnlDesc.write(fileIDOfMsgFromTransIF)) {
    cerr << "MsgDrv::PushToResponseSpool: write failed to Service" << endl;
    return false;
  }
  
  std::string fileIDOfMsgFromResSMgr;
  if (0 > resSpoolPushChnlDesc.read(fileIDOfMsgFromResSMgr)) {
    cerr << "MsgDrv::PushToResponseSpool: read failed from Service" << endl;
    return false;
  }
  
  cerr << "** MsgDrv ** READ From ResSpool=[" 
       << fileIDOfMsgFromResSMgr << "]" << endl;
  
  resSpoolPushChnlSelector->close(resSpoolPushChnlDesc);
  delete resSpoolPushChnlSelector;
  
  
  cout << "send to trans if" << endl;
  std::string msgToReqQMgr = "success";
  std::string fileIDOfMsgToReqQMgr = convertToFileID(msgToReqQMgr);
  if (0 > chnlDesc_.write(fileIDOfMsgToReqQMgr)) {
    cerr << TAG_ERR << "MsgDrv::pushToResponseSpool: write failed" << endl;
    return false;
  }
  
  return true;
  
}// end of pushToResponseSpool(std::string fileIDOfMsgFromSrvdrv)

bool
MsgDrv::run()
{
  cerr << "MsgDrv::run()" << endl;
  
  // TransIF 瑗Ăf[^ (t@C ID) ǂݍށB
  if (0 > chnlDesc_.read(fileIDOfMsgFromIf)) {
    cerr << TAG_ERR << "run: read failed" << endl;
    return false;
  }
  
  cerr << "** MsgDrv ** READ From TransIF=[" 
       << fileIDOfMsgFromIf << "]" << endl;
  
  // t@C ID t@C̓eɕϊB
  soapMsgFromIf = revertFromFileID (fileIDOfMsgFromIf);
  
  // SOAP bZ[W̎ނ̃`FbN
  bool soapMsgIsValid;
  
  // bZ[W ID H
  bool itIsMsgID = false;
  soapMsgIsValid = isItMsgID(soapMsgFromIf, itIsMsgID);
  cerr << "itIsMsgID = " << itIsMsgID << endl;
  if (!soapMsgIsValid) {
    soapMsgFromSrvdrv = makeInvalidRequestMessage();
    fileIDOfMsgFromSrvdrv = convertToFileID(soapMsgFromSrvdrv);
    chnlDesc_.write(fileIDOfMsgFromSrvdrv); 
    return false;
  }
  if (itIsMsgID) {
    std::string fileIDOfMsgID = fileIDOfMsgFromIf;
    popFromResponseSpool(fileIDOfMsgID);
    return true;
  }
  
  // <response>  true ŁA
  // <backward_path>  TTLMgr ɓo^ĂȂƂ
  // X|XXv[ɑD
  bool itIsResMsg = isItResponseMessage(soapMsgFromIf);
  std::string msgIDContent = getMsgIDContent(soapMsgFromIf);
  std::string backwardPath;
  if (msgIDContent.size() > 0) {
    refToTTLTable(msgIDContent, backwardPath);
    if (itIsResMsg && backwardPath == "") {
      cout << "pushToResponseSpool()" << endl;
      pushToResponseSpool(fileIDOfMsgFromIf);
      return true;
    }
  }
  
  // Iy[V͉H
  cerr << "extract Operation and Parameter" << endl;
  
  soapMsgIsValid 
    = extractOperationAndParam (soapMsgFromIf, operationNames, 
				paramNames, paramValues);
  if (!soapMsgIsValid) {
    std::string invalidReqErrMsg = makeInvalidRequestMessage();
    std::string fileIDOfInvalidReqErrMsg = convertToFileID(invalidReqErrMsg);
    chnlDesc_.write(fileIDOfInvalidReqErrMsg); 
    return false;
  }
  
  /* Does async header exist ? */
  // 񓯊wb_tĂ邩H
  bool asyncHeaderExist, isAsync;
  
  soapMsgIsValid 
    = getAsyncInfoFromHeader(soapMsgFromIf, asyncHeaderExist, isAsync);
  if (!soapMsgIsValid) {
    std::string invalidReqErrMsg = makeInvalidRequestMessage();
    std::string fileIDOfInvalidReqErrMsg = convertToFileID(invalidReqErrMsg);
    chnlDesc_.write(fileIDOfInvalidReqErrMsg); 
    return false;
  }
  
  /* Does fwd header exist ? */
  // ]wb_tĂ邩H
  bool fwdPathExist, fwdHopExist;
  std::vector<std::string> fwdPath;
  unsigned int fwdHop;
  
  soapMsgIsValid 
    = getFwdInfoFromHeader(soapMsgFromIf, fwdPathExist, fwdPath,
			   fwdHopExist, fwdHop);
  if (!soapMsgIsValid) {
    std::string invalidReqErrMsg = makeInvalidRequestMessage();
    std::string fileIDOfInvalidReqErrMsg = convertToFileID(invalidReqErrMsg);
    chnlDesc_.write(fileIDOfInvalidReqErrMsg); 
    return false;
  }
  
  cerr << "fwdPathExist = " << fwdPathExist << endl;
  cerr << "fwdHopExist = " << fwdHopExist << endl;
  
  /* Does ttl header exist ? */
  bool ttlHeaderExist;
  ttl = 0;
  std::string ttlType;
  
  soapMsgIsValid 
    = getTTLInfo(soapMsgFromIf, ttlHeaderExist, ttl, ttlType);
  if (!soapMsgIsValid) {
    std::string invalidReqErrMsg = makeInvalidRequestMessage();
    std::string fileIDOfInvalidReqErrMsg = convertToFileID(invalidReqErrMsg);
    chnlDesc_.write(fileIDOfInvalidReqErrMsg); 
    return false;
  }
  
  
//    bool isFwdByPathName;
//    std::vector<std::string> fwdPathName;
//    bool isFwdByHopCount;
//    unsigned int fwdHopCountValue;
  
  //* SSML p[^𒲂ׂ */
  OpenSOAP::SSMLInfo ssmlInfo;
  ssmlInfo.setOperationName(operationNames[0]);
  operationExist = getSSMLInfo(ssmlInfo);
  
  cerr << "MsgDrv::run::"
       << "ssmlInfo.getSyncTTL() = " << ssmlInfo.getSyncTTL() << endl;
  
  bool targetThreadIsCanceled = false;
  
  if (!operationExist && !defaultForwarding && !fwdPathExist) {
    // Iy[V݂A
    // ftHg̃tH[hȂ
    // tH[hwb_Ȃꍇ
    std::string oprNotFoundMsg = makeOperationNotFoundMessage();
    std::string fileIDOfOprNotFoundMsg = convertToFileID(oprNotFoundMsg);
    chnlDesc_.write(fileIDOfOprNotFoundMsg); 
    return false;
  }
  
  if (!asyncHeaderExist || !isAsync) {
    /* async wb_ȂD܂ async wb_ false ̏ꍇ */
    /* (̏ꍇ) */
    if (operationExist || defaultForwarding || fwdPathExist) {
      // Iy[V݂D܂̓ftHg̃tH[h悪ꍇ
      if (!ttlHeaderExist) {
	ttl = ssmlInfo.getSyncTTL();
      }  
      targetThreadIsCanceled = sendToSrvDrvOrFwd();
    }
    if (targetThreadIsCanceled) {
      // ^CAEg thr_func LZďIꍇB
      soapMsgFromSrvdrv = makeTimeoutMessage();
      fileIDOfMsgFromSrvdrv = convertToFileID(soapMsgFromSrvdrv);
      chnlDesc_.write(fileIDOfMsgFromSrvdrv); 
    }
  }
  else {
    /* async wb_ true ̏ꍇ */
    /* (񓯊̏ꍇ) */
    if (operationExist || defaultForwarding 
	|| fwdPathExist || fwdHopExist) {
      // Iy[V݂D܂̓ftHg̃tH[h悪ꍇ
      //        targetThreadIsCanceled = sendToSrvDrvOrFwd();
      if (!ttlHeaderExist) {
	ttl = ssmlInfo.getAsyncTTL();
      }
      targetThreadIsCanceled 
	= pushToQueue(fwdPathExist, fwdHopExist, defaultForwarding);
    }
    if (targetThreadIsCanceled) {
      // ^CAEg thr_func LZďIꍇB
      soapMsgFromSrvdrv = makeTimeoutMessage();
      fileIDOfMsgFromSrvdrv = convertToFileID(soapMsgFromSrvdrv);
      chnlDesc_.write(fileIDOfMsgFromSrvdrv); 
    }
    
  }
  return true;
  
}// end of run()
