/* -*- mode: c++; -*-
 *-----------------------------------------------------------------------------
 * $RCSfile: Queue.h,v $
 *
 * See Copyright for the status of this software.
 *
 * The OpenSOAP Project
 * http://opensoap.jp/
 *-----------------------------------------------------------------------------
 */

#ifndef QUEUE_H
#define QUEUE_H

#include <fstream.h>
#include <vector>
#include <pthread.h>
#include <string>

#define MAX_POP_THREAD_NUMBER 4
#define INITIAL_STATE 0
#define PROGRESS 1

namespace OpenSOAP {
  
  class ChannelDescriptor;
  
  class Queue {
  private:
    pthread_t pop_thread[MAX_POP_THREAD_NUMBER];
    pthread_mutex_t queue_mutex;
    pthread_cond_t queue_cond;
    
    char tableFile[100];
    ifstream tin;
    ofstream tout;
    bool isFwdQueue_;
    struct TableRecord{
      //string msgID;
      string fileID;
      int proccessFlag; // INITIAL_STATE, PROGRESS
    };
    
    vector<TableRecord> queueTable;
    ChannelDescriptor chnlDesc_;  
  public:
    Queue(char filePass[100], bool isFwdQueue);
    int push(string fileID);
    ~Queue(); 
    void setChnlDesc(const ChannelDescriptor& chnlDesc);
    bool execPushOpr();
    
  private:
    int fileOut();
    static void* pop(void* arg);
    
    bool popOutToNextDrv(string fileID);
    int queueEraseAndFileOut(string fileID);
    
  };
}
#endif /* QUEUE_H */
