/* -*- mode: c++; -*-
 *-----------------------------------------------------------------------------
 * $RCSfile: QueueManager_main.cpp,v $
 *
 * See Copyright for the status of this software.
 *
 * The OpenSOAP Project
 * http://opensoap.jp/
 *-----------------------------------------------------------------------------
 */

#include <unistd.h>
#include <fstream>
#include "QueueManager.h"
#include "ServerCommon.h"

using namespace OpenSOAP;

int 
main(int argc, char* argv[])
{
  // if argument exists, it behaves as Queue Manager of Forwarder
  bool isFwdQueue = true;
  if (argc == 1 || (argc > 1 && atoi(argv[1]) == 0)) {
    isFwdQueue = false;
  }

  // make it to daemon & write pid file
#if HAVE_DAEMON
  daemon(0, 1);
#endif
  
  pid_t pid = getpid();
  
  
  ofstream pidFile;
  if (isFwdQueue) {
    // Queue Manager for Forward
    cout << "Forward Queue Manager" << endl;
    pidFile.open(FWDQUEUE_PID_FILE.c_str());
    if (!pidFile) {
      cerr << "Error: pid file [" 
           << FWDQUEUE_PID_FILE << "] cannot open!" << endl;
      exit(1);
    }
    pidFile << pid;
    pidFile.close();
  }
  else {
    // Queue Manager for Request
    cout << "Response Queue Manager" << endl;
    pidFile.open(REQQUEUE_PID_FILE.c_str());
    if (!pidFile) {
      cerr << "Error: pid file [" 
           << REQQUEUE_PID_FILE << "] cannot open!" << endl;
      exit(1);
    }
  }
  pidFile << pid;
  pidFile.close();
  
  // create instance of QueueManager and run
  ChannelManager* queueManagerPtr = new QueueManager(isFwdQueue);
  queueManagerPtr->run();
  delete queueManagerPtr;
  return 0;
}

