/* -*- mode: c++; -*-
 *-----------------------------------------------------------------------------
 * $RCSfile: SOAPMessage.cpp,v $
 *
 * See Copyright for the status of this software.
 *
 * The OpenSOAP Project
 * http://opensoap.jp/
 *-----------------------------------------------------------------------------
 */

#include <algorithm>
#include <string>
#include <vector>

#include <unistd.h>
#include <netdb.h>

/* libxml2 */
#include <tree.h>
#include <parser.h>

#include "SOAPMessage.h"
#include "MsgAttrHandler.h"
#include "SOAPException.h"
#include "XmlModifier.h"
#include "StringUtil.h"

using namespace OpenSOAP;

std::string
createSOAPFaultMessage (std::string faultcodeStr, std::string faultstringStr,
			std::string faultactorStr, std::string detailStr)
{
    LIBXML_TEST_VERSION xmlKeepBlanksDefault(0);
    xmlDocPtr doc;
    xmlAttrPtr attr1, attr2;
    xmlNodePtr body, fault, faultcode, faultstring, faultactor, detail;
    xmlNsPtr soapEnvNs;
	
    /* make xml element */ 
    doc = xmlNewDoc((xmlChar*)("1.0"));
    attr1 = xmlNewDocProp(doc, (xmlChar*)("encoding"), 
			  (xmlChar*)("UTF-8")); /* ??? */
    attr2 = xmlNewDocProp(doc, (xmlChar*)("standalone"), 
			  (xmlChar*)("no")); /* ??? */
    
    /* make Envelope element */
    doc->children = xmlNewDocNode(doc, NULL, 
				  (xmlChar*)("Envelope"), NULL);
    xmlSetProp(doc->children, (xmlChar*)("encodingStyle"),
	       (xmlChar*)("http://schemas.xml.soap.org/soap/encoding/"));
    soapEnvNs = xmlNewNs(doc->children,
			 (xmlChar*)("http://schemas.xmlsoap.org/soap/envelope/"),
			 (xmlChar*)("SOAP-ENV"));
    xmlSetNs(doc->children, soapEnvNs);
    
    /* make Body element */
    body = xmlNewChild(doc->children, NULL, (xmlChar*)("Body"), NULL);
    xmlSetNs(body, soapEnvNs);
    
    /* make Fault element */
    fault = xmlNewChild(body, NULL, (xmlChar*)("Fault"), NULL);
    
    /* make faultcode element */
    faultcode = xmlNewChild(fault, NULL, (xmlChar*)("faultcode"),
			    (xmlChar*)faultcodeStr.c_str());
    
    /* make faultstring element */
    faultstring = xmlNewChild(fault, NULL, (xmlChar*)("faultstring"),
			      (xmlChar*)faultstringStr.c_str());

    /* make faultactor element */
    if (faultactorStr.length() >0) {
      faultactor = xmlNewChild(fault, NULL, (xmlChar*)("faultactor"),
			       (xmlChar*)faultactorStr.c_str());
    }
    
    /* make detail element */
    detail = xmlNewChild(fault, NULL, (xmlChar*)("detail"),
			 (xmlChar*)detailStr.c_str());
    
    /* output doc tree */
    xmlChar* mem_ptr;
    int size;
    xmlDocDumpMemory(doc, &mem_ptr, &size);
    
    std::string soapMsg ((char*)mem_ptr, size);
    
    /* free memory */
    xmlFreeDoc(doc);
    
    return soapMsg;
}


/*
  in: 
  ret: bool, operation ݂邩
  throw: SOAPMsgIsInvalidException
 */
bool
extractOperationAndParam (std::string soapMsg,
			  std::vector<std::string>& operationNames,
			  std::vector<std::string>& paramNames,
			  std::vector<std::string>& paramValues) 
{
  
  cerr << "=====  extractOperationAndParam() begin =====" << endl;
  
#ifdef DEBUG  
  cerr << "soapMsg = [" << endl;
  cerr << soapMsg << endl;
  cerr << "]" << endl;
#endif
  
  bool operationExist = false;
  std::string query;
  MsgAttrHandler msgAttrHndl(soapMsg);
  std::vector<std::string> values;
  std::vector<std::string>::iterator iter;
  
  operationNames.clear();
  paramNames.clear();
  paramValues.clear();  
  
  // Iy[V擾B
  query = "/Envelope/Body/?";
  if (0 > msgAttrHndl.queryXml(query, values)) {
    cerr << "*** ERROR ***" << endl;
    return operationExist;
  }
  
  cerr << "SOAPMessage::isItMsgID:values[0] = " << values[0] << endl;
  operationNames.push_back(values[0]);
  values.clear();
  
  // p[^擾B
  bool paramsExist = true;
  query = "/Envelope/Body/" + operationNames[0] +  "/??";
  if (0 > msgAttrHndl.queryXml(query, values)) {
    cerr << "*** ERROR ***" << endl;
    return paramsExist;
  }
  
  if (paramsExist) {
    for (iter = values.begin(); iter != values.end(); iter++) {
      cerr << "SOAPMessage::isItMsgID:values = " << *iter << endl;
      paramNames.push_back(*iter);    
    }
    values.clear();
    
    // p[^̒l擾B
    
    for (iter = paramNames.begin(); iter != paramNames.end(); iter++) {
      query = "/Envelope/Body/" + operationNames[0] +  "/"
	+ *iter + "=?";
      if (0 > msgAttrHndl.queryXml(query, values)) {
	cerr << "*** ERROR ***" << endl;
	return operationExist;
      }
      paramValues.push_back(values[0]);
      values.clear();
    }
  }// end of if (paramsExist)
  
  operationExist = true;  
  cerr << "=====  extractOperationAndParam() end =====" << endl;
  
  return operationExist;
}// end of extractOperationAndParam()

/*
  Messaage ID ł邩ׂB
 */
bool
isItMsgID(const std::string soapMsg, bool& itIsMsgID)
{
  cerr << "=====  isItMsgID() begin =====" << endl;
  itIsMsgID = false;
  
  std::string query;
  MsgAttrHandler msgAttrHndl(soapMsg);
  std::vector<std::string> values;

  query = "/Envelope/Header/opensoap-header-block/message_id=?";
  if (0 > msgAttrHndl.queryXml(query, values)) {
    cerr << "*** ERROR ***" << endl;
    return true;
  }
  
  cerr << "SOAPMessage::isItMsgID:values[0] = " << values[0] << endl;
  
  values.clear();
  
  query = "/Envelope/Body/?";
  if (0 > msgAttrHndl.queryXml(query, values)) {
    cerr << "*** ERROR ***" << endl;
    // <Body/>  ̏ꍇ SOAP bZ[W ID Ȃ̂ true ԂB
    itIsMsgID = true;
    return true;
  }
  
  cerr << "SOAPMessage::isItMsgID:values[0] = " << values[0] << endl;
  if (values[0] == "") {
    // <message_id>  <Body> ̏ꍇA
    // ̃bZ[W̓bZ[W ID ƌȂB
    itIsMsgID = true;
  }
  
  cerr << "=====  isItMsgID() end =====" << endl;  
  
  return true;
}

/*
  E񓯊Ɋւׂ
 */
bool
getAsyncInfoFromHeader(const std::string soapMsg,
		       bool& asyncHeaderExist, bool& isAsync)
{
  cerr << "=====  getAsyncInfoFromHeader() begin =====" << endl;
  asyncHeaderExist = false;
  isAsync = false;
  
  std::string query = "/Envelope/Header/opensoap-header-block/async=?";
  MsgAttrHandler msgAttrHndl(soapMsg);
  std::vector<std::string> values;
  if (0 > msgAttrHndl.queryXml(query, values)) {
    cerr << "*** ERROR ***" << endl;
    isAsync = false;
    return true;
  }
  
  cerr << "SOAPMessage::getAsyncInfoFromHeader:values[0] = " << values[0] << endl;
  
  if (values[0] == "true") {
    asyncHeaderExist = true;
    isAsync = true;
  }
  else if (values[0] == "false") {
    asyncHeaderExist = true;
    isAsync = false;
  }
  else {
    return true;
  }
  
  cerr << "=====  getAsyncInfoFromHeader() end =====" << endl;
  
  return true;
}// end of getAsyncInfoFromHeader()

/*
  ]Ɋւׂ
 */
bool
getFwdInfoFromHeader(const std::string soapMsg,
		     bool& fwdPathExist, std::vector<std::string>& fwdPath,
		     bool& hopCountExist, unsigned int& hopCount)
{
  cerr << "=====  getFwdInfoFromHeader() begin =====" << endl;
  fwdPathExist = false;
  hopCountExist = false;
  fwdPath.clear();
  
  std::string query;
  MsgAttrHandler msgAttrHndl(soapMsg);
  std::vector<std::string> values;
  std::vector<std::string>::iterator iter;
  
  query = "/Envelope/Header/opensoap-header-block/forwarder/forward_path=??";  
  if (0 <= msgAttrHndl.queryXml(query, values)) {
    fwdPathExist = true;
    for (iter = values.begin(); iter != values.end(); iter++) {
      fwdPath.push_back(*iter);
      cerr << "SOAPMessage::getFwdInfoFromHeader:forward_path = " 
	   << *iter << endl;    
    }
  }
  
  values.clear();
  query = "/Envelope/Header/opensoap-header-block/forwarder/hopcount=?";  
  if (0 <= msgAttrHndl.queryXml(query, values)) {
    hopCountExist = true;    
    hopCount = atoi(values[0].c_str());
      cerr << "SOAPMessage::getFwdInfoFromHeader:hopcount = " 
	   << values[0] << endl;
  }
  
  
  cerr << "=====  getFwdInfoFromHeader() end =====" << endl;
  
  return true;
}// end of getFwdInfoFromHeader()

/*
   Received Path Ɋւ𒲂ׂ
 */
bool
getReceivedPathInfo(const string soapMsg, 
		    bool& receivedPathExist, 
		    std::vector<std::string>& receivedPath) 
{
  cerr << "=====  getReceivedInfo() begin =====" << endl;
  receivedPathExist = false;
  receivedPath.clear();
  
  std::string query;
  MsgAttrHandler msgAttrHndl(soapMsg);
  std::vector<std::string> values;
  std::vector<std::string>::iterator iter;
  
  query = "/Envelope/Header/opensoap-header-block/received_path/url=??";  
  if (0 <= msgAttrHndl.queryXml(query, values)) {
    receivedPathExist = true;
    for (iter = values.begin(); iter != values.end(); iter++) {
      receivedPath.push_back(*iter);
      cerr << "SOAPMessage::getReceivedInfo:values = " 
	   << *iter << endl;    
    }
  }
  cerr << "=====  getReceivedInfo() end =====" << endl;
  
  return true;
}

/*
  TTL Ɋւׂ
 */
bool
getTTLInfo(const std::string soapMsg,
	   bool& ttlHeaderExist, unsigned int& ttl, std::string& ttlType)
{
  cerr << "=====  getTTLInfo() begin =====" << endl;
  
  ttlHeaderExist = false;
  
  std::string query;
  MsgAttrHandler msgAttrHndl(soapMsg);
  std::vector<std::string> values;

  query = "/Envelope/Header/opensoap-header-block/ttl=?";
  if (0 > msgAttrHndl.queryXml(query, values)) {
    cerr << "*** ERROR ***" << endl;
    cerr << "[" <<query << "] is not exist." << endl;
    ttl = 0;
    ttlType = "";
    ttlHeaderExist = false;
    return true;
  }
  cerr << "SOAPMessage::getTTLInfo: values[0] = " << values[0] << endl; 
  ttl = atoi(values[0].c_str());
  ttlHeaderExist = true;
  values.clear();
  
  query = "/Envelope/Header/opensoap-header-block/ttl,type=?";
  query = "/Envelope/Header/opensoap-header-block/ttl=?";
  if (0 > msgAttrHndl.queryXml(query, values)) {
    cerr << "*** ERROR ***" << endl;
    cerr << "[" <<query << "] is not exist." << endl;
    ttl = 0;
    ttlType = "";
    ttlHeaderExist = false;
    return true;
  }
  cerr << "SOAPMessage::getTTLInfo: values[0] = " << values[0] << endl; 
  ttlType = values[0];
  
  cerr << "=====  getTTLInfo() end =====" << endl;
  
  return true;
}// end of getTTLInfo()


//----- Iy[V^CAEgƂ Fault bZ[W쐬 -----//
std::string
makeTimeoutMessage()
{
  
  std::string faultcodeStr = "SOAP-ENV:Server.Timeout";
  std::string faultstringStr = "Time Out occured in Server";
  std::string faultactorStr = "";

  std::string serverName;
  size_t len = 128;  
  char* hostname = new char[len];
  int ret = gethostname(hostname, len);
  if (ret < 0) {
    serverName = "localhost";
  }
  else {
    hostent* hostEntPtr;
    hostEntPtr = gethostbyname(hostname);
    cerr << "hostEntPtr->h_name = " << hostEntPtr->h_name << endl;
    serverName = hostEntPtr->h_name;
  }
  delete[] hostname;
  
  std::string official_hostname = serverName;
  
  std::string detailStr = std::string("This FAULT is created by ") 
    + official_hostname + std::string(".");
  
  std::string soapMsgToIf
    = createSOAPFaultMessage(faultcodeStr, faultstringStr,
			     faultactorStr, detailStr);
  
  return soapMsgToIf;
}


//----- Iy[VȂƂ Fault bZ[W쐬 -----//
std::string
makeOperationNotFoundMessage()
{
  std::string faultcodeStr = "SOAP-ENV:Server.OperationNotFound";
  std::string faultstringStr = "Request Operation is not Treated in Server";
  std::string faultactorStr = "";
  
  std::string serverName;
  size_t len = 128;  
  char* hostname = new char[len];
  int ret = gethostname(hostname, len);
  if (ret < 0) {
    serverName = "localhost";
  }
  else {
    hostent* hostEntPtr;
    hostEntPtr = gethostbyname(hostname);
    cerr << "hostEntPtr->h_name = " << hostEntPtr->h_name << endl;
    serverName = hostEntPtr->h_name;
  }
  delete[] hostname;
  
  std::string official_hostname = serverName;
  
  std::string detailStr = std::string("This FAULT is created by ") 
    + official_hostname + std::string(".");
  
  std::string soapMsgToIf
    = createSOAPFaultMessage(faultcodeStr, faultstringStr,
			     faultactorStr, detailStr);
  return soapMsgToIf;
}

//----- SOAP NGXgsłƂ Fault bZ[W쐬 -----//
std::string
makeInvalidRequestMessage()
{
  std::string faultcodeStr = "SOAP-ENV:Server.RequestInvalid";
  std::string faultstringStr = "Request from Client is Invalid";
  std::string faultactorStr = "";
  
  std::string serverName;
  size_t len = 128;  
  char* hostname = new char[len];
  int ret = gethostname(hostname, len);
  if (ret < 0) {
    serverName = "localhost";
  }
  else {
    hostent* hostEntPtr;
    hostEntPtr = gethostbyname(hostname);
    cerr << "hostEntPtr->h_name = " << hostEntPtr->h_name << endl;
    serverName = hostEntPtr->h_name;
  }
  delete[] hostname;
  
  std::string official_hostname = serverName;
  
  std::string detailStr = std::string("This FAULT is created by ") 
    + official_hostname + std::string(".");
  
  std::string soapMsgToIf
    = createSOAPFaultMessage(faultcodeStr, faultstringStr,
			     faultactorStr, detailStr);
  return soapMsgToIf;
}

//----- create Falut Messgage of No Entry -----//
std::string
makeNoEntryMessage()
{
  std::string faultcodeStr = "SOAP-ENV:Server.NoEntry";
  std::string faultstringStr = "No Corresponding Entry in Response Spool";
  std::string faultactorStr = "";
  
  std::string serverName;
  size_t len = 128;  
  char* hostname = new char[len];
  int ret = gethostname(hostname, len);
  if (ret < 0) {
    serverName = "localhost";
  }
  else {
    hostent* hostEntPtr;
    hostEntPtr = gethostbyname(hostname);
    cerr << "hostEntPtr->h_name = " << hostEntPtr->h_name << endl;
    serverName = hostEntPtr->h_name;
  }
  delete[] hostname;
  
  std::string official_hostname = serverName;
  
  std::string detailStr = std::string("This FAULT is created by ") 
    + official_hostname + std::string(".");
  
  std::string soapMsgToIf
    = createSOAPFaultMessage(faultcodeStr, faultstringStr,
			     faultactorStr, detailStr);
  return soapMsgToIf;
}

//----- create Falut Messgage in the case of -----//
//----- Response from Serive is Empty        -----//
std::string
makeResponseIsEmptyMessage()
{
  std::string faultcodeStr = "SOAP-ENV:Server.ResponseIsEmpty";
  std::string faultstringStr = "Response from Service is Empty";
  std::string faultactorStr = "";
  
  std::string serverName;
  size_t len = 128;  
  char* hostname = new char[len];
  int ret = gethostname(hostname, len);
  if (ret < 0) {
    serverName = "localhost";
  }
  else {
    hostent* hostEntPtr;
    hostEntPtr = gethostbyname(hostname);
    cerr << "hostEntPtr->h_name = " << hostEntPtr->h_name << endl;
    serverName = hostEntPtr->h_name;
  }
  delete[] hostname;
  
  std::string official_hostname = serverName;
  
  std::string detailStr = std::string("This FAULT is created by ") 
    + official_hostname + std::string(".");
  
  std::string soapMsgToIf
    = createSOAPFaultMessage(faultcodeStr, faultstringStr,
			     faultactorStr, detailStr);
  return soapMsgToIf;
}

//----- SOAP bZ[WɃbZ[W ID tƂ -----//
//----- bZ[W ID Ԃ -----//
//----- MsgDrv Ŏgp -----//
std::string
attachMsgID(std::string& soapMsg, const std::string& fileIDOfMsgFromIf,
	    const std::string& operationName)
{
  cerr << "===== attachMsgID() begin =====" << endl;
  
  // ɃbZ[W ID Ăꍇ͂ȂɂtȂ
  std::string msgIDContent;
  msgIDContent = getMsgIDContent(soapMsg);
  if (msgIDContent.size() > 0) {
    return msgIDContent;
  }
  
  std::string dot = ".";
  
  std::string serverName = getLocalhostName();
  
  std::string receivedDateTime = fileIDOfMsgFromIf;
  
  msgIDContent = operationName + dot 
    + serverName + dot + receivedDateTime;
  
  //----- SOAP bZ[W ID t -----//
  xmlDocPtr soapMsgDoc = NULL;
  xmlNsPtr ns;
  xmlNodePtr node1, node2;
  
  /* build an XML tree from a std::string */
  
  char *soapMsgChar = new char[soapMsg.length()];
  memcpy (soapMsgChar, soapMsg.c_str(), soapMsg.length());
  
  soapMsgDoc = xmlParseMemory(soapMsgChar, soapMsg.length());
  if (soapMsgDoc == NULL) {
    fprintf(stderr, "parsing soapMsg =  '%s' failed\n", soapMsgChar);
    delete[] soapMsgChar;
    return NULL;
  }
  delete[] soapMsgChar;
  
  /* check document (root node) */
  node1 = xmlDocGetRootElement(soapMsgDoc);
  if (node1 == NULL) {
    fprintf(stderr, "empty document\n");
    xmlFreeDoc(soapMsgDoc);
    return NULL;
  }
  cout << "node1->name = " << node1->name << endl;
  
  /* check namespace */
  ns = xmlSearchNsByHref(soapMsgDoc, node1,
			 (xmlChar*)("http://schemas.xmlsoap.org/soap/envelope/"));
  if (ns == NULL) {
    fprintf(stderr, "wrong namespace of Envelope\n");
    xmlFreeDoc(soapMsgDoc);
    return NULL;
  }
  
  /* check document type (root node) */
  if (strcmp((const char*)(node1->name), "Envelope") != 0) {
    fprintf(stderr, "wrong document type");
    xmlFreeDoc(soapMsgDoc);
    return NULL;
  }
  
  node2 = node1->xmlChildrenNode;
  if (node2 == NULL) {
    cerr << "node1->xmlChildrenNode == NULL" << endl;
    return NULL;
  }
  cerr << "node2->name = " << node2->name << endl;
  
  xmlNewChild(node2->xmlChildrenNode, NULL, (xmlChar*)("message_id"),
	      (xmlChar*)(msgIDContent.c_str()));
  
  
  //----- bZ[W ID ̍쐬 -----//
  LIBXML_TEST_VERSION xmlKeepBlanksDefault(0);
  xmlDocPtr msgIDDoc;
  xmlAttrPtr attr1, attr2;
  xmlNodePtr body;
  xmlNsPtr soapEnvNs;
  
  /* make xml element */ 
  msgIDDoc = xmlNewDoc((xmlChar*)("1.0"));
  attr1 = xmlNewDocProp(msgIDDoc, (xmlChar*)("encoding"), 
			(xmlChar*)("UTF-8")); /* ??? */
  attr2 = xmlNewDocProp(msgIDDoc, (xmlChar*)("standalone"), 
			(xmlChar*)("no")); /* ??? */
  
  /* make Envelope element */
  msgIDDoc->children = xmlNewDocNode(msgIDDoc, NULL, 
				     (xmlChar*)("Envelope"), NULL);
  xmlSetProp(msgIDDoc->children, (xmlChar*)("encodingStyle"),
	     (xmlChar*)("http://schemas.xml.soap.org/soap/encoding/"));
  soapEnvNs = xmlNewNs(msgIDDoc->children,
		       (xmlChar*)("http://schemas.xmlsoap.org/soap/envelope/"),
		       (xmlChar*)("SOAP-ENV"));
  xmlSetNs(msgIDDoc->children, soapEnvNs);
  
  
  /* make Header element */
  xmlNodePtr header;
  header = xmlNewChild(msgIDDoc->children, NULL, (xmlChar*)("Header"), NULL);
  
  xmlNodePtr extBlock 
    = xmlNewChild(header, NULL, (xmlChar*)("opensoap-header-block"), NULL);
  xmlNsPtr opensoapExtNs 
    = xmlNewNs(extBlock, (xmlChar*)("http://header.opensoap.jp/1.0/"),
	       (xmlChar*)("opensoap-header"));
  xmlSetNs(extBlock, opensoapExtNs);
  
  xmlNewChild(extBlock, NULL, (xmlChar*)("message_id"),
	      (xmlChar*)(msgIDContent.c_str()));
  
/* make Body element */
  body = xmlNewChild(msgIDDoc->children, NULL, (xmlChar*)("Body"), NULL);
  xmlSetNs(body, soapEnvNs);
  
  /* output doc tree */
  xmlChar* mem_ptr;
  int size;
  
  xmlDocDumpMemory(soapMsgDoc, &mem_ptr, &size);
  soapMsg = std::string((char*)mem_ptr, size);
  
  cout << "SOAPMessage::attachMsgID:New soapMsg = " << soapMsg << endl;
  
  xmlDocDumpMemory(msgIDDoc, &mem_ptr, &size);
  std::string msgID = std::string((char*)mem_ptr, size);
  
  /* free memory */
  xmlFreeDoc(soapMsgDoc);
  xmlFreeDoc(msgIDDoc);
  
  cerr << "===== attachMsgID() end =====" << endl;
  
  return msgID;
}

//----- SOAP bZ[WɃbZ[W ID tƂ -----//
//----- bZ[W ID Ԃ -----//
//----- SrvDrv Ŏgp -----//
std::string
attachOrUpdateMsgID(std::string& soapMsg, std::string msgIDContent)
{
  cerr << "===== attachOrUpdateMsgID() begin =====" << endl;
  
  XmlModifier xmlMod(soapMsg);
  XmlModifier::NameSpaceDef nsDef;
  std::string fmt;
  std::string val;
  int result;
  std::string newMsg;
  
  // ɃbZ[W ID Ăꍇ͍XV, 
  // ȂɂȂꍇ͕tB
  std::string msgIDContentCurrent;
  msgIDContentCurrent = getMsgIDContent(soapMsg);
  if (msgIDContentCurrent.size() > 0) {
    
    
  }
  else {
    // add header 
    fmt = "/Envelope/?";
    val = "Header";
    result = xmlMod.attach(fmt, val, NULL);
    if (result != 0) {
      cerr << "attach failed. code=(" << result << ")" << endl;
    }
    newMsg = xmlMod.toString();
    cout << "####################################" << endl;
    cout << "newMsg=[" << newMsg << "]" << endl;
    cout << "####################################" << endl;
    
    // add header block
    fmt = "/Envelope/Header/?";
    val = "opensoap-header-block";
    nsDef.href = "http://header.opensoap.jp/1.0/";
    nsDef.prefix = "opensoap-header";
    result = xmlMod.attach(fmt, val, &nsDef);
    if (result != 0) {
      cerr << "attach failed. code=(" << result << ")" << endl;
    }
    newMsg = xmlMod.toString();
    cout << "####################################" << endl;
    cout << "newMsg=[" << newMsg << "]" << endl;
    cout << "####################################" << endl;

    // add message ID
    fmt = "/Envelope/Header/opensoap-header-block/message_id=?";
    val = msgIDContent;
    result = xmlMod.attach(fmt, val, &nsDef);
    if (result != 0) {
      cerr << "attach failed. code=(" << result << ")" << endl;
    }
    newMsg = xmlMod.toString();
    cout << "####################################" << endl;
    cout << "newMsg=[" << newMsg << "]" << endl;
    cout << "####################################" << endl;
    soapMsg = newMsg;
  }
  
  cerr << "===== attachOrUpdateMsgID() end =====" << endl;
  
  return msgIDContent;
}


//----- attach backward path to response message,                 -----//
//----- if this server is the final receiver of async fwd message -----//
//----- This is used by SrvDrv -----//
bool
attachBackwardPathToResponse(std::string& msgFromService, 
			     std::string msgFromMsgDrv)
{
  
  cerr << "===== attachBackwardPathToResponse() begin =====" << endl;
  
  XmlModifier xmlMod(msgFromService);
  XmlModifier::NameSpaceDef nsDef;
  std::string fmt;
  std::string val;
  int result;
  std::string newMsg;
  
  std::string query;
  MsgAttrHandler msgAttrHndlRes(msgFromService);
  MsgAttrHandler msgAttrHndlReq(msgFromMsgDrv);
  std::vector<std::string> values;
  
  // check if response has backward path
  values.clear();
  query = "/Envelope/Header/opensoap-header-block/backward_path=??";
  if (0 > msgAttrHndlRes.queryXml(query, values)) {
    cerr << "*** ERROR ***" << endl;
    cerr << " query = [" << query << "] is not found."<< endl;
  }
  
  // if there is no backward path in Response message,
  // add backward path of request message,
  // otherwise do nothing
  if(values.size() > 0) {
    cerr << "Response Message already has <backward_path>." << endl;
    return true;
  }
  
  // retain backward_path from Request message
  values.clear();
  query = "/Envelope/Header/opensoap-header-block/backward_path=??";
  if (0 > msgAttrHndlReq.queryXml(query, values)) {
    cerr << "*** ERROR ***" << endl;
    cerr << " query = [" << query << "] is not found."<< endl;
  }

  // return error, if no backward_path in Request message
  if (values.size() == 0) {
    cerr << "Request Message has no <backward_path>." << endl;
    return false;
  }
  
  // add backward_path
  fmt = "/Envelope/Header/opensoap-header-block/backward_path=?";
  for(std::vector<std::string>::iterator iter = values.begin();
      iter != values.end(); iter++) {
    val = *iter;
    result = xmlMod.attach(fmt, val, NULL);
    if (result != 0) {
      cerr << "attach failed. code=(" << result << ")" << endl;
      return false;
    }
  }// loop of for(iter)
  
  newMsg = xmlMod.toString();
  cout << "####################################" << endl;
  cout << "newMsg=[" << newMsg << "]" << endl;
  cout << "####################################" << endl;
  
  // add <response_msg> with value of "true"
  fmt = "/Envelope/Header/opensoap-header-block/response_msg=?";
  val = "true";
  result = xmlMod.attach(fmt, val, NULL);
  if (result != 0) {
    cerr << "attach failed. code=(" << result << ")" << endl;
    return false;
  }
  
  newMsg = xmlMod.toString();
  cout << "####################################" << endl;
  cout << "newMsg=[" << newMsg << "]" << endl;
  cout << "####################################" << endl;

  msgFromService = newMsg;
  
  cerr << "===== attachBackwardPathToResponse() end =====" << endl;
  
  return true;
}// end of attachBackwardPathToResponse()

//----- attach TTL to response message,                 -----//
//----- if this server is the final receiver of async fwd message -----//
//----- This is used by SrvDrv -----//
bool
attachTTLToResponse(std::string& msgFromService, 
		    std::string msgFromMsgDrv)
{
  
  cerr << "===== attachTTLToResponse() begin =====" << endl;
  
  XmlModifier xmlMod(msgFromService);
  XmlModifier::NameSpaceDef nsDef;
  std::string fmt;
  std::string val;
  int result;
  std::string newMsg;
  
  std::string query;
  MsgAttrHandler msgAttrHndlRes(msgFromService);
  MsgAttrHandler msgAttrHndlReq(msgFromMsgDrv);
  std::vector<std::string> values;
  
  // check if response has ttl
  values.clear();
  query = "/Envelope/Header/opensoap-header-block/ttl=?";
  if (0 > msgAttrHndlRes.queryXml(query, values)) {
    cerr << "*** ERROR ***" << endl;
    cerr << " query = [" << query << "] is not found."<< endl;
  }
  
  // if there is no ttl path in Response message,
  // add backward path of request message,
  // otherwise do nothing
  if (values.size() > 0) {
    cerr << "Response Message already has <ttl>." << endl;
    return true;
  }
  
  // retain ttl from Request message
  values.clear();
  query = "/Envelope/Header/opensoap-header-block/ttl=?";
  if (0 > msgAttrHndlReq.queryXml(query, values)) {
    cerr << "*** ERROR ***" << endl;
    cerr << " query = [" << query << "] is not found."<< endl;
  }
  
  // if no ttl in Request message, ttl becomes defalut value
  if (values.size() == 0) {
    cerr << "Request Message has no <ttl>." << endl;
    values[0] = 10;
  }
  
  // add ttl
  fmt = "/Envelope/Header/opensoap-header-block/ttl=?";
  val = values[0];
  result = xmlMod.attach(fmt, val, NULL);
  if (result != 0) {
    cerr << "attach failed. code=(" << result << ")" << endl;
    return false;
  }
  
  newMsg = xmlMod.toString();
  cout << "####################################" << endl;
  cout << "newMsg=[" << newMsg << "]" << endl;
  cout << "####################################" << endl;
  
  // retain type of ttl from Request message
  values.clear();
  query = "/Envelope/Header/opensoap-header-block/ttl,type=?";
  if (0 > msgAttrHndlReq.queryXml(query, values)) {
    cerr << "*** ERROR ***" << endl;
    cerr << " query = [" << query << "] is not found."<< endl;
  }
  
  // if no type of ttl in Request message, type of ttl becomes defalut value
  if (values.size() == 0) {
    cerr << "Request Message has no <ttl type=>." << endl;
    values[0] = "second";
  }
  
  // add type of ttl
  fmt = "/Envelope/Header/opensoap-header-block/ttl,type=?";
  val = values[0];
  result = xmlMod.attach(fmt, val);
  if (result != 0) {
    cerr << "attach failed. code=(" << result << ")" << endl;
    return false;
  }
  
  newMsg = xmlMod.toString();
  cout << "####################################" << endl;
  cout << "newMsg=[" << newMsg << "]" << endl;
  cout << "####################################" << endl;
  
  

  msgFromService = newMsg;
  
  cerr << "===== attachTTLToResponse() end =====" << endl;
  
  return true;
}// end of attachTTLToResponse()

//----- SOAP bZ[W received_path t -----//
bool
attachReceivedPath(std::string& soapMsg, 
		   const std::string& fileIDOfMsgFromIf,
		   const std::string& url)
{
  cerr << "===== attachReceivedPath() begin =====" << endl;
  
  std::string receivedDateTime = fileIDOfMsgFromIf;
  
  //-----------------------------------------
  // XmlModifier
  //-----------------------------------------
  XmlModifier xmlMod(soapMsg);
  XmlModifier::NameSpaceDef nsDef;
  std::string fmt;
  std::string val;
  std::string newMsg;
  
  int result = 0;
  //vfǉ
  //   fmt="/XXX/?" + val="AAA"  fmt="/XXX/AAA=?" + val=""ƓD
  //Namespacewǉ(eQ)
  //----------------------------------
  fmt = "/Envelope/Header/opensoap-header-block/?";
  val = "received_path";
  
  //namespacewDNULLw肷,ǉvf̐eNSpD
  result = xmlMod.attach(fmt, val, NULL);
  if (result != 0) {
    cerr << "attach failed. code=(" << result << ")" << endl;
  }
  newMsg = xmlMod.toString();
  cout << "####################################" << endl;
  cout << "newMsg=[" << newMsg << "]" << endl;
  cout << "####################################" << endl;
  
  //----------------------------------
  //Pǉ
  //----------------------------------
  fmt = "/Envelope/Header/opensoap-header-block/received_path/url=?";
  val = url;
  
  result = xmlMod.attach(fmt, val);
  if (result != 0) {
    cerr << "attach failed. code=(" << result << ")" << endl;
  }
  newMsg = xmlMod.toString();
  cout << "####################################" << endl;
  cout << "newMsg=[" << newMsg << "]" << endl;
  cout << "####################################" << endl;

  //----------------------------------
  //Pǉ
  //----------------------------------
  fmt = "/Envelope/Header/opensoap-header-block/received_path/time=?";
  val = fileIDOfMsgFromIf;
  
  result = xmlMod.attach(fmt, val);
  if (result != 0) {
    cerr << "attach failed. code=(" << result << ")" << endl;
  }
  newMsg = xmlMod.toString();
  cout << "####################################" << endl;
  cout << "newMsg=[" << newMsg << "]" << endl;
  cout << "####################################" << endl;
  

  soapMsg = newMsg;
  
  cerr << "===== attachReceivedPath() end =====" << endl;
  
  return true;
}

//----- SOAP bZ[W backward_path tEXV -----//
bool
updateBackwardPath(std::string& soapMsg, 
		   const std::string& backwardURL)
{
  cerr << "===== updateBackwardPath() begin =====" << endl;
  
  //-----------------------------------------
  // XmlModifier
  //-----------------------------------------
  XmlModifier xmlMod(soapMsg);
  XmlModifier::NameSpaceDef nsDef;
  std::string fmt;
  std::string val;
  std::string newMsg;
  
  int result = 0;
  //----------------------------------
  //XV
  //----------------------------------
  fmt = "/Envelope/Header/opensoap-header-block/backward_path=?";
  val = backwardURL;
  
  result = xmlMod.update(fmt, val);
  if (result != 0) {
    // XV backward_path Ȃꍇ͍쐬
    cerr << "attach failed. code=(" << result << ")" << endl;
    
    fmt = "/Envelope/Header/opensoap-header-block/backward_path=?";
    val = backwardURL;
    
    result = xmlMod.attach(fmt, val);
    if (result != 0) {
      cerr << "attach failed. code=(" << result << ")" << endl;
    }
  }
  
  newMsg = xmlMod.toString();
  cout << "####################################" << endl;
  cout << "newMsg=[" << newMsg << "]" << endl;
  cout << "####################################" << endl;
  
  soapMsg = newMsg;
  cerr << "===== updateBackwardPath() end =====" << endl;
  
  return true;
}// end of updateBackwardPath()

//----- Decrease value of <hopcount> in SOAP Message -----//
extern
bool
decreaseHopCount(std::string& soapMsg)
{
  cerr << "===== decreaseHopCount begin =====" << endl;

  std::string query;
  MsgAttrHandler msgAttrHndl(soapMsg);
  std::vector<std::string> values;
  
  // check if response has ttl
  values.clear();
  query = "/Envelope/Header/opensoap-header-block/forwarder/hopcount=?";
  if (0 > msgAttrHndl.queryXml(query, values)) {
    cerr << "*** ERROR ***" << endl;
    cerr << " query = [" << query << "] is not found."<< endl;
    return false;
  }
  
  int currentHopCount = atoi(values[0].c_str());
  
  //-----------------------------------------
  // XmlModifier
  //-----------------------------------------
  XmlModifier xmlMod(soapMsg);
  XmlModifier::NameSpaceDef nsDef;
  std::string fmt;
  std::string val;
  std::string newMsg;
  
  int result = 0;
  //----------------------------------
  //XV
  //----------------------------------
  fmt = "/Envelope/Header/opensoap-header-block/forwarder/hopcount=?";
  val = StringUtil::toString(currentHopCount - 1);
  
  result = xmlMod.update(fmt, val);
  if (result != 0) {
    return false;
  }
  
  newMsg = xmlMod.toString();
  cout << "####################################" << endl;
  cout << "newMsg=[" << newMsg << "]" << endl;
  cout << "####################################" << endl;
  
  soapMsg = newMsg;
  cerr << "===== decreaseHopCount end =====" << endl;
  
  return true;
}// end of decreaseHopCount


//----- SOAP bZ[WɃbZ[W ID t -----//
//----- T[rXŎgp -----//
bool
attachMsgID(std::string& soapMsg, const std::string& msgIDContent)
{
  cerr << "===== attachMsgID() begin =====" << endl;
  
  //----- SOAP bZ[W ID t -----//
    //-----------------------------------------
  // XmlModifier
  //-----------------------------------------
  XmlModifier xmlMod(soapMsg);
  XmlModifier::NameSpaceDef nsDef;
  std::string fmt;
  std::string val;
  std::string newMsg;
  
  int result = 0;
  //vfǉ
  //   fmt="/XXX/?" + val="AAA"  fmt="/XXX/AAA=?" + val=""ƓD
  //Namespacewǉ(eQ)
  //----------------------------------
  fmt = "/Envelope/?";
  val = "Header";
  
  //namespacewDNULLw肷,ǉvf̐eNSpD
  result = xmlMod.attach(fmt, val, NULL);
  if (result != 0) {
    cerr << "attach failed. code=(" << result << ")" << endl;
  }
  newMsg = xmlMod.toString();
  cout << "####################################" << endl;
  cout << "newMsg=[" << newMsg << "]" << endl;
  cout << "####################################" << endl;
  
  //----------------------------------
  // <opensoap-header:opensoap-header-block 
  //    xmlns:opensoap-header="http://header.opensoap.jp/1.0/" mustUnderstand="1">
  //----------------------------------
  //Namespace(VK)wǉ
  //----------------------------------
  
  fmt = "/Envelope/Header/?";
  val = "opensoap-header-block";
  nsDef.href = "http://header.opensoap.jp/1.0/";
  nsDef.prefix = "opensoap-header";
  
  result = xmlMod.attach(fmt, val, &nsDef);
  if (result != 0) {
    cerr << "attach failed. code=(" << result << ")" << endl;
  }
  newMsg = xmlMod.toString();
  cout << "####################################" << endl;
  cout << "newMsg=[" << newMsg << "]" << endl;
  cout << "####################################" << endl;
  
  //----------------------------------
  // <opensoap-header:opensoap-header-block 
  //    xmlns:opensoap-header="http://header.opensoap.jp/1.0/" mustUnderstand="1">
  //----------------------------------
  //wǉ
  //----------------------------------
  fmt = "/Envelope/Header/opensoap-header-block,mustUnderstand=?";
  val = "1";
  
  result = xmlMod.attach(fmt, val);
  if (result != 0) {
    cerr << "attach failed. code=(" << result << ")" << endl;
  }
  newMsg = xmlMod.toString();
  cout << "####################################" << endl;
  cout << "newMsg=[" << newMsg << "]" << endl;
  cout << "####################################" << endl;
  
  //----------------------------------
  // <opensoap-header:message_id>
  //   ServiceName.ServerName.RecieveDateTime
  // </opensoap-header:message_id>
  //----------------------------------
  //Namespace(Q)wǉ
  //----------------------------------
  fmt = "/Envelope/Header/opensoap-header-block/message_id=?";
  val = msgIDContent;
  //namespacewDNULLw肷,ǉvf̐eNSpD
  result = xmlMod.attach(fmt, val, NULL);
  if (result != 0) {
    cerr << "attach failed. code=(" << result << ")" << endl;
  }
  newMsg = xmlMod.toString();
  cout << "####################################" << endl;
  cout << "newMsg=[" << newMsg << "]" << endl;
  cout << "####################################" << endl;
  
  //----------------------------------
  // <opensoap-header:response_msg>
  //   true
  // </opensoap-header:response_msg>
  //----------------------------------
  //Namespace(Q)wǉ
  //----------------------------------
  fmt = "/Envelope/Header/opensoap-header-block/response_msg=?";
  val = "true";
  //namespacewDNULLw肷,ǉvf̐eNSpD
  result = xmlMod.attach(fmt, val, NULL);
  if (result != 0) {
    cerr << "attach failed. code=(" << result << ")" << endl;
  }
  newMsg = xmlMod.toString();
  cout << "####################################" << endl;
  cout << "newMsg=[" << newMsg << "]" << endl;
  cout << "####################################" << endl;
  
  soapMsg = newMsg;
  cerr << "===== attachMsgID() end =====" << endl;
  return true;
}

// SOAPbZ[W烁bZ[W ID ̕oB
std::string
getMsgIDContent(const std::string soapMsg)
{
  string msgIDContent = "";
  
  cerr << "=====  getMsgIDContent() begin =====" << endl;
  
  std::string query;
  MsgAttrHandler msgAttrHndl(soapMsg);
  std::vector<std::string> values;
  
  query = "/Envelope/Header/opensoap-header-block/message_id=?";
  if (0 > msgAttrHndl.queryXml(query, values)) {
    cerr << "*** ERROR ***" << endl;
    return msgIDContent;
  }
  
  cerr << "SOAPMessage::getMsgIDContent:values[0] = " << values[0] << endl;
  msgIDContent = values[0];
  
  cerr << "=====  getMsgIDContent() end =====" << endl;  
  
  return msgIDContent;
}

// NGXg SOAP bZ[WÃT[o[
// bZ[W̍ŏIł邩𔻒fB
// MsgDrv, AsyncSrvDrv g
bool
isItAsyncFwdFinalReceiver(std::string soapMsg,
			  std::vector<std::string> soapServerEndPoints)
{
  bool itIsAsyncFwdFinalReceiver = false;
  
  cerr << "=====  isItAsyncFwdFinalReceiver() begin =====" << endl;
  
  cerr << "soapMsg = [" << soapMsg << "]" << endl;
  
  std::string query;
  MsgAttrHandler msgAttrHndl(soapMsg);
  std::vector<std::string> values;
  
  query = "/Envelope/Header/opensoap-header-block/forwarder/forward_path=??";
  if (0 > msgAttrHndl.queryXml(query, values)) {
    cerr << "*** ERROR ***" << endl;
    cerr << " query = [" << query << "] is not found."<< endl;
    query = "/Envelope/Header/opensoap-header-block/forwarder/hopcount=?";
    if (0 > msgAttrHndl.queryXml(query, values)) {
      cerr << "*** ERROR ***" << endl;
      cerr << " query = [" << query << "] is not found."<< endl;
      return itIsAsyncFwdFinalReceiver;
    }
    else {
      cerr << "hopcount = " << values[0] << endl;
      if (atoi(values[0].c_str()) == 1) {
	itIsAsyncFwdFinalReceiver = true;
      }
    }
    return itIsAsyncFwdFinalReceiver;
  }
  
  cerr << "SOAPMessage::isItAsyncFwdFinalReceiver:values.end() = " 
       << *(values.end() - 1) << endl;
  std::string soapServerEndPoint;
  for (std::vector<std::string>::iterator iter = soapServerEndPoints.begin();
       iter != soapServerEndPoints.end(); iter++) {
    if (*(values.end() - 1) == *iter) {
      itIsAsyncFwdFinalReceiver = true;
      soapServerEndPoint = *iter;
      break;
    }
  }
  cout << "soapServerEndPoint = " << soapServerEndPoint << endl;
  cout << "isItAsyncFwdFinalReceiver = " 
       << itIsAsyncFwdFinalReceiver << endl;;
  
  cerr << "=====  isItAsyncFwdFinalReceiver() end =====" << endl;  
  
  return itIsAsyncFwdFinalReceiver;
}// end of isItAsyncFwdFinalReceiver()

/*
  backward_path Ɋւׂ
 */
bool
getBackwardPath(const std::string& soapMsg, std::string& backwardPath)
{
  cerr << "=====  getBackwardPath() begin =====" << endl;
  
  backwardPath = "";
  
  std::string query;
  MsgAttrHandler msgAttrHndl(soapMsg);
  std::vector<std::string> values;

  query = "/Envelope/Header/opensoap-header-block/backward_path=?";
  if (0 > msgAttrHndl.queryXml(query, values)) {
    cerr << "*** ERROR ***" << endl;
    cerr << "[" <<query << "] is not exist." << endl;
    return false;
  }
  cerr << "SOAPMessage::getTTLInfo: values[0] = " << values[0] << endl; 
  backwardPath = values[0];
  
  cerr << "=====  getBackwardPath() end   =====" << endl;
  return true;
}// end of getBackwardPath()

// bZ[WX|Xǂ𔻒f
// MsgDrv, Forwarder g
bool
isItResponseMessage(std::string soapMsg)
{
  bool itIsResponseMessage = false;
  
  cerr << "=====  isItResponseMessage() begin =====" << endl;
  
  cerr << "soapMsg = [" << soapMsg << "]" << endl;
  
  std::string query;
  MsgAttrHandler msgAttrHndl(soapMsg);
  std::vector<std::string> values;
  
  query = "/Envelope/Header/opensoap-header-block/response_msg=?";
  if (0 > msgAttrHndl.queryXml(query, values)) {
    cerr << "*** ERROR ***" << endl;
    cerr << " query = [" << query << "] is not found."<< endl;
    return itIsResponseMessage;
  }
  
  cerr << "SOAPMessage::isItResponseMessage:values[0] = " 
       << values[0] << endl;
  
  if (values[0] == "true") {
    itIsResponseMessage = true;
  }
  
  cout << "isItResponseMessage = " 
       << itIsResponseMessage << endl;;
  
  cerr << "=====  isItResponseMessage() end =====" << endl;  
  
  return itIsResponseMessage;
}// end of isItResponseMessage()

//----- Service ƒʐMƂ߂̃wb_t -----//
std::string
attachHeaderForService(std::string soapMsg)
{
  std::string newSOAPMsg;
  std::vector<std::string> operationNames;
  std::vector<std::string> paramNames;
  std::vector<std::string> paramValues;
  std::string endString = "\r\n";
  extractOperationAndParam(soapMsg, operationNames, paramNames, paramValues);
  
  newSOAPMsg = std::string("POST ") + operationNames[0]
    + std::string(" HTTP/1.1") + endString;
  newSOAPMsg += std::string("Content-Type: text/xml; charset=UTF-8") 
    + endString;
  newSOAPMsg += std::string("Content-Length: ") 
    + StringUtil::toString(soapMsg.length()) + endString;
  newSOAPMsg += endString;
  newSOAPMsg += soapMsg;
  
  return newSOAPMsg;
  
}// end of attachHeaderForService(std::string soapMsg)

//----- Service ƒʐMƂ߂̃wb_폜 -----//
std::string
removeHeaderForService(std::string recvMsg)
{
  char delim = '<';
  std::string::iterator openBraceLoc
    = find (recvMsg.begin(), recvMsg.end (), delim);
  if (openBraceLoc == recvMsg.end()) {
    cerr << "HTTP Message does not contain SOAP Message !" << endl;
    return std::string ("");
  }
  std::string soapMsg = std::string(openBraceLoc, recvMsg.end());
  
  return soapMsg;
}// end of removeHeaderForService (std::string& recvMsg)

//----- SrvDrv ƒʐMƂ߂̃wb_t -----//
std::string
attachHeaderForSrvDrv(std::string soapMsg)
{
  std::string newSOAPMsg;
  std::vector<std::string> operationNames;
  std::vector<std::string> paramNames;
  std::vector<std::string> paramValues;
  std::string endString = "\r\n";
  extractOperationAndParam(soapMsg, operationNames, paramNames, paramValues);
  
  newSOAPMsg = std::string("HTTP/1.1 200 OK") + endString;
  newSOAPMsg += std::string("Content-Type: text/xml; charset=UTF-8") 
    + endString;
  newSOAPMsg += std::string("Content-Length: ") 
    + StringUtil::toString(soapMsg.length()) + endString;
  newSOAPMsg += endString;
  newSOAPMsg += soapMsg;
  
  return newSOAPMsg;
  
}// end of attachHeaderForService(std::string soapMsg)

//----- FQDN t̃zXgԂ -----//
std::string
getLocalhostName() 
{
  char localhostname[128];
  int ret0 = gethostname(localhostname, sizeof(localhostname));
  if (ret0 == 0) {
    cerr << "localhostname = " << localhostname << endl;
  }
  else {
    cerr << "gethostname() failed !!" << endl;
  }
  std::string officialHostname = localhostname;
  
  hostent* hostEntPtr;
  hostEntPtr = gethostbyname(localhostname);
  if (hostEntPtr != NULL) {
    cerr << "hostEntPtr = " << hostEntPtr << endl;
    if (hostEntPtr->h_name != NULL) {
      cerr << "hostEntPtr->h_name = " << hostEntPtr->h_name << endl;
      officialHostname = hostEntPtr->h_name;
    }
  }
  
  return officialHostname;
}
