/* -*- mode: c++; -*-
 *-----------------------------------------------------------------------------
 * $RCSfile: SSMLInfoFunc.cpp,v $
 *
 * See Copyright for the status of this software.
 *
 * The OpenSOAP Project
 * http://opensoap.jp/
 *-----------------------------------------------------------------------------
 */

#include <iostream>

/* libxml2 */
#include <tree.h>
#include <parser.h>

#include "SSMLInfoFunc.h"
#include "SSMLAttrHandler.h"

using namespace OpenSOAP;


bool
getOperationNameAndType(SSMLInfo& ssmlInfo)
{
  cerr << "===== SSMLInfoFunc::getOperationNameAndType() begin =====" << endl;
  bool operationExist = false;
  
  // operation name ̎擾
  std::string query = "/SSML/service/operation=??"; 
  
  SSMLAttrHandler ssmlAttrHndl;
  std::vector<std::string> values;
  
  if (0 > ssmlAttrHndl.queryXml(query, values)) {
    cerr << "*** ERROR ***" << "query = [" << query << "]" << endl;
    return operationExist;
  }
  
  std::vector<std::string>::const_iterator pos;
  unsigned int operationNum = 0;
  for (pos = values.begin(); pos != values.end(); ++pos, ++operationNum) {
    cerr << "values = " << *pos << ", ";
    cerr << "ssmlInfo.operationName = " << ssmlInfo.getOperationName() << endl;
    if (*pos == ssmlInfo.getOperationName()) {
      operationExist = true;
      break;
    }
  }
  cerr << "SSMLInfoFunc::getOperationNameAndType: " 
       << "obtain operationType" << endl;
  // operation type ̎擾
  std::vector<std::string> values1;
  if (operationExist) {
    query = "/SSML/service/operation,type=??";
    if (0 > ssmlAttrHndl.queryXml(query, values1)) {
      cerr << "*** ERROR ***" << "query = [" 
	   << query << "]" << endl;
      operationExist = false;
    }
    else {
      ssmlInfo.setOperationType(values1[operationNum]);
      cerr << "values1[operationNum] = " << values1[operationNum] << endl;
    }
  }
  
  cerr << "===== SSMLInfoFunc::getOperationNameAndType() end   =====" << endl;
  return operationExist;
}

void
getSyncTTL(SSMLInfo& ssmlInfo)
{
  std::string query = "/SSML/service/connection,name=" 
    + ssmlInfo.getOperationType() + "/synchronizedTTL=?";
  SSMLAttrHandler ssmlAttrHndl;
  std::vector<std::string> values;
  
  if (0 > ssmlAttrHndl.queryXml(query, values)) {
    cerr << "*** ERROR ***" << "query = [" 
	 << query << "]" << endl;
    ssmlInfo.setSyncTTL(0);
  }
  else {
    cerr << "values[0] = "  << values[0] << endl;
    ssmlInfo.setSyncTTL(atoi(values[0].c_str()));
  }
}

void
getAsyncTTL(SSMLInfo& ssmlInfo)
{
  std::string query = "/SSML/service/connection,name=" 
    + ssmlInfo.getOperationType() + "/asynchronizedTTL=?";
  SSMLAttrHandler ssmlAttrHndl;
  std::vector<std::string> values;
  
  if (0 > ssmlAttrHndl.queryXml(query, values)) {
    cerr << "*** ERROR ***" << "query = [" 
	 << query << "]" << endl;
    ssmlInfo.setAsyncTTL(0);
  }
  else {
    cerr << "values[0] = "  << values[0] << endl;
    ssmlInfo.setAsyncTTL(atoi(values[0].c_str()));
  }
}

void
getHostname(SSMLInfo& ssmlInfo)
{
  std::string query = "/SSML/service/connection,name=" 
    + ssmlInfo.getOperationType() + "/Socket,hostname=?";
  SSMLAttrHandler ssmlAttrHndl;
  std::vector<std::string> values;
  
  if (0 > ssmlAttrHndl.queryXml(query, values)) {
    cerr << "*** ERROR ***" << "query = [" 
	 << query << "]" << endl;
    ssmlInfo.setHostname("");
  }
  else {
    cerr << "values[0] = "  << values[0] << endl;
    ssmlInfo.setHostname(values[0]);
  }
}

void
getPort(SSMLInfo& ssmlInfo)
{
  std::string query = "/SSML/service/connection,name=" 
    + ssmlInfo.getOperationType() + "/Socket,port=?";
  SSMLAttrHandler ssmlAttrHndl;
  std::vector<std::string> values;
  
  if (0 > ssmlAttrHndl.queryXml(query, values)) {
    cerr << "*** ERROR ***" << "query = [" 
	 << query << "]" << endl;
    ssmlInfo.setPort(0);
  }
  else {
    cerr << "values[0] = "  << values[0] << endl;
    ssmlInfo.setPort(atoi(values[0].c_str()));
  }
}

extern
bool
getSSMLInfo(SSMLInfo& ssmlInfo)
{
  /* parse ssml begin */

  // Iy[Vƃ^Cv̎擾
  bool operationExist = getOperationNameAndType(ssmlInfo);
  
  // TTL ̎擾
  std::vector<std::string> timeValues;
  if (operationExist) {
    getSyncTTL(ssmlInfo);
    cerr << "SSMLInfoFunc::getSSMLInfo:"
	 << "ssmlInfo.getSyncTTL() = " << ssmlInfo.getSyncTTL() << endl;
    
    getAsyncTTL(ssmlInfo);
    cerr << "SSMLInfoFunc::getSSMLInfo:"
	 << "ssmlInfo.getAsyncTTL() = " << ssmlInfo.getAsyncTTL() << endl;

    getHostname(ssmlInfo);
    cerr << "SSMLInfoFunc::getSSMLInfo:"
	 << "ssmlInfo.getHostname() = " << ssmlInfo.getHostname() << endl;
    
    getPort(ssmlInfo);
    cerr << "SSMLInfoFunc::getSSMLInfo:"
	 << "ssmlInfo.getPort() = " << ssmlInfo.getPort() << endl;
  }
  /* parse ssml end */
  
  return operationExist;

}

