/* -*- mode: c++; -*-
 *-----------------------------------------------------------------------------
 * $RCSfile: SrvConfAttrHandler.cpp,v $
 *
 * See Copyright for the status of this software.
 *
 * The OpenSOAP Project
 * http://opensoap.jp/
 *-----------------------------------------------------------------------------
 */

#if HAVE_CONFIG_H
# include "config.h"
#endif

#include <string>

#include <sys/socket.h>
#include <sys/un.h>
#include <cstdio>
#include <unistd.h>
#include <string>

//for shm
#include <sys/ipc.h>
#include <sys/shm.h>
#include <sys/sem.h>

#include "MapUtil.h"
#include "ServerCommon.h"
#include "SSMLAttrMgrProtocol.h"
#include "SrvConfAttrHandler.h"

using namespace OpenSOAP;

// for LOG
static const std::string CLASS_SIG = "SrvConfAttrHandler";

SrvConfAttrHandler::SrvConfAttrHandler()
  : socketAddr_(SRVCONFATTRMGR_SOCKET_ADDR)
{
  
}

SrvConfAttrHandler::~SrvConfAttrHandler()
{

}

int SrvConfAttrHandler::queryXml(const std::string& query, 
			      std::vector<std::string>& values)
{
#ifdef DEBUG
  cerr << CLASS_SIG << "::queryXml() arg=[" << query << "]" << endl;
#endif /* DEBUG */

  //check shm buffer
  if (getShmCache(query, values)) {
    return 0;
  }

  int sockfd = connectManager();
  if (0 > sockfd) {
    cerr << TAG_ERR << CLASS_SIG 
	 << "::queryXml: connect manager failed." << endl;
    return -1;
  }
  
  //MbZ[W쐬
  MapUtil writeDataMap(ITEM_DELMIT, VAL_DELMIT);
  writeDataMap.insert(make_pair(CMD,SEARCH));
  writeDataMap.insert(make_pair(QUERY,query));

#ifdef DEBUG
  writeDataMap.spy();
#endif /* DEBUG */

  //bZ[WM
  if (0 > write(sockfd, writeDataMap.toString())) {
    cerr << TAG_ERR << CLASS_SIG 
	 << "::queryXml: write to manager failed." << endl;
    close(sockfd);
    return -1;
  }
  
  std::string readBuf;
  if (0 > read(sockfd, readBuf)) {
    cerr << TAG_ERR << CLASS_SIG 
	 << "::queryXml: read from manager failed." << endl;
    close(sockfd);
    return -1;
  }

#ifdef DEBUG
  cerr << CLASS_SIG << "::queryXml: read=[" << readBuf << "]" << endl;
#endif /* DEBUG */

  MapUtil readDataMap(ITEM_DELMIT, VAL_DELMIT);
  readDataMap.insertItem(readBuf);

  //ʎo
  std::string result = readDataMap[RESULT];
  if (SUCCESS != result) {
#ifdef DEBUG
    cerr << CLASS_SIG << "::queryXml: result failure. result=[" 
	 << result << "]" << endl;
#endif /* DEBUG */
    close(sockfd);
    return -1;
  }

  //𓚎o
  std::string answer = readDataMap[ANSWER];
  
  std::string::size_type begIdx = 0;
  std::string::size_type endIdx = 0;
  for (;;) {
    begIdx = answer.find_first_not_of(SUB_ITEM_DELMIT, endIdx);
    if (begIdx == std::string::npos) {
#ifdef DEBUG
      cerr << "END" << endl;
#endif /* DEBUG */
      break;
    }
    endIdx = answer.find_first_of(SUB_ITEM_DELMIT, begIdx);
    if (endIdx == std::string::npos) {
      endIdx = answer.length();
    }
    std::string token = answer.substr(begIdx, (endIdx-begIdx));

#ifdef DEBUG
    cerr << "token=[" << token << "]" << endl;
#endif /* DEBUG */

    values.push_back(token);
  }

  close(sockfd);
  return 0;
}

int SrvConfAttrHandler::connectManager()
{
  typedef struct sockaddr SockAddr;
  typedef struct sockaddr_un SockAddrUn;

  int sockfd = int();
  SockAddrUn servAddr;

  memset((char*)&servAddr, 0, sizeof(servAddr));
  servAddr.sun_family = AF_UNIX;
  strcpy(servAddr.sun_path, socketAddr_.c_str());

  if (0 > (sockfd = socket(AF_UNIX, SOCK_STREAM, 0))) {
    cerr << TAG_ERR << CLASS_SIG 
	 << "::connectManager open socket failed" << endl;
    return -1;
  }

#ifdef DEBUG
  cerr << CLASS_SIG << "::connectManager: socket open OK!" << endl;
#endif
  
#if HAVE_TYPE_OF_ADDRLEN_AS_SOCKLEN_T
  socklen_t servAddrLen = socklen_t();
#endif
#if HAVE_TYPE_OF_ADDRLEN_AS_SIZE_T
  size_t servAddrLen = size_t();
#endif
#if HAVE_TYPE_OF_ADDRLEN_AS_INT
  int servAddrLen = int();
#endif
  
  servAddrLen = sizeof(servAddr.sun_family) + strlen(servAddr.sun_path);

  if (0 > connect(sockfd, (SockAddr*)&servAddr, servAddrLen)) {
    cerr << TAG_ERR << CLASS_SIG 
	 << "::connectManager connect failed" << endl;
    return -1;
  }

#ifdef DEBUG
  cerr << CLASS_SIG << "::connectManager: connect OK!" << endl;
#endif

    return sockfd;
}


int SrvConfAttrHandler::reloadXml()
{
  int sockfd = connectManager();
  if (0 > sockfd) {
    cerr << TAG_ERR << CLASS_SIG 
	 << "::reloadXml: connect manager failed." << endl;
    return -1;
  }
  
  //MbZ[W쐬
  MapUtil writeDataMap(ITEM_DELMIT, VAL_DELMIT);
  writeDataMap.insert(make_pair(CMD,RELOAD_XML));
  //writeDataMap.insert(make_pair(QUERY,query));

#ifdef DEBUG
  writeDataMap.spy();
#endif /* DEBUG */

  //bZ[WM
  if (0 > write(sockfd, writeDataMap.toString())) {
    cerr << TAG_ERR << CLASS_SIG 
	 << "::reloadXml: write to manager failed." << endl;
    close(sockfd);
    return -1;
  }
  
  std::string readBuf;
  if (0 > read(sockfd, readBuf)) {
    cerr << TAG_ERR << CLASS_SIG 
	 << "::reloadXml: read from manager failed." << endl;
    close(sockfd);
    return -1;
  }

#ifdef DEBUG
  cerr << CLASS_SIG << "::reloadXml: read=[" << readBuf << "]" << endl;
#endif /* DEBUG */

  MapUtil readDataMap(ITEM_DELMIT, VAL_DELMIT);
  readDataMap.insertItem(readBuf);

  //ʎo
  std::string result = readDataMap[RESULT];
  if (SUCCESS != result) {
#ifdef DEBUG
    cerr << CLASS_SIG << "::reloadXml: result failure. result=[" 
	 << result << "]" << endl;
#endif /* DEBUG */
    close(sockfd);
    return -1;
  }

  cerr << TAG_INF << CLASS_SIG << ": xml reload done." << endl;

  close(sockfd);
  return 0;

}

bool SrvConfAttrHandler::getShmCache(const std::string& query,
				     std::vector<std::string>& values)
{
  // SEM check
  key_t semKey = ftok(SRVCONFATTRMGR_SOCKET_ADDR.c_str(), 1);
  
  int semid = semget(semKey, 1, 0666);
  if (-1 == semid) {
    cerr << TAG_ERR << CLASS_SIG << "::getShmCache: semget failed." << endl;
    return false;
  }
  struct sembuf semBuf;
  semBuf.sem_op = 0;
  semBuf.sem_num = 0;
  semBuf.sem_flg = 0;
  if (-1 == semop(semid, &semBuf, 1)) { // if sem val not 0, process blocked.
    cerr << TAG_ERR << CLASS_SIG << "::getShmCache: semop failed." << endl;
    return false;
  }

  key_t shm_key = ftok(SRVCONFATTRMGR_SOCKET_ADDR.c_str(), 0);
  const int SHMSZ = 196608;
  char shmBuf[SHMSZ];
  int shmid = 0;
  char* shm = NULL;

  if ((shmid = shmget(shm_key, SHMSZ, IPC_CREAT|0666)) < 0) {
    cerr << TAG_ERR << CLASS_SIG << "::getShmCache: shmget failed." << endl;
    return false;
  }

  if ((shm = (char*)shmat(shmid, NULL, 0)) == (char*)-1) {
    cerr << TAG_ERR << CLASS_SIG << "::getShmCache: shmat failed." << endl;
    return false;
  }

  memset(shmBuf, 0x00, SHMSZ);
  strncpy(shmBuf, shm, SHMSZ);
  std::string data(shmBuf);

  std::string::size_type begIdx = 0;
  std::string::size_type endIdx = 0;
  const std::string delimit = "|";
  for (;;) {
    begIdx = data.find_first_not_of(delimit, endIdx);
    if (begIdx == std::string::npos) {
#ifdef DEBUG
      cerr << "END" << endl;
#endif /* DEBUG */
      break;
    }
    endIdx = data.find_first_of(delimit, begIdx);
    if (endIdx == std::string::npos) {
      endIdx = data.length();
    }
    std::string key = data.substr(begIdx, (endIdx-begIdx));

#ifdef DEBUG
    cerr << "key=[" << key << "]" << endl;
#endif /* DEBUG */

    begIdx = data.find_first_not_of(delimit, endIdx);
    if (begIdx == std::string::npos) {

#ifdef DEBUG
      cerr << "END" << endl;
#endif /* DEBUG */
      break;
    }
    endIdx = data.find_first_of(delimit, begIdx);
    if (endIdx == std::string::npos) {
      endIdx = data.length();
    }
    std::string val = data.substr(begIdx, (endIdx-begIdx));

#ifdef DEBUG
    cerr << "val=[" << val << "]" << endl;
#endif /* DEBUG */
    
    if (key == query) {
      std::string::size_type begIdx = 0;
      std::string::size_type endIdx = 0;
      for (;;) {
	begIdx = val.find_first_not_of(SUB_ITEM_DELMIT, endIdx);
	if (begIdx == std::string::npos) {
#ifdef DEBUG
	  cerr << "END" << endl;
#endif /* DEBUG */
	  break;
	}
	endIdx = val.find_first_of(SUB_ITEM_DELMIT, begIdx);
	if (endIdx == std::string::npos) {
	  endIdx = val.length();
	}
	std::string token = val.substr(begIdx, (endIdx-begIdx));
	
#ifdef DEBUG
	cerr << "token=[" << token << "]" << endl;
#endif /* DEBUG */
	
	values.push_back(token);
      }
      break;
    }
  }

  shmdt(shm);

  return values.size() > 0 ? true : false;
}
