/* -*- mode: c++; -*-
 *-----------------------------------------------------------------------------
 * $RCSfile: SrvConfAttrMgr.cpp,v $
 *
 * See Copyright for the status of this software.
 *
 * The OpenSOAP Project
 * http://opensoap.jp/
 *-----------------------------------------------------------------------------
 */

#if HAVE_CONFIG_H
# include "config.h"
#endif

#include <sys/types.h>
#include <dirent.h>
#include <sys/stat.h>

#include <sys/ipc.h>
#include <sys/shm.h>
#include <sys/sem.h>

#include <sys/socket.h>
#include <sys/un.h>
#include <unistd.h>

#include <cstdio>
#include <stdexcept>
#include <string>

#include <libxml/tree.h>
#include <libxml/parser.h>

#include "MapUtil.h"
#include "ServerCommon.h"
//#include "SrvConfAttrMgrProtocol.h"
#include "SSMLAttrMgrProtocol.h"
#include "XmlQuery.h"
#include "SrvConfAttrMgr.h"

using namespace OpenSOAP;

const std::string SrvConfAttrMgr::CONF_FILE_SUFFIX = "conf";

// for LOG
static const std::string CLASS_SIG = "SrvConfAttrMgr";

//UNIXhC\Pbg̃AhXXMLt@CfBNgw肷
SrvConfAttrMgr::SrvConfAttrMgr(const std::string& socketAddr, 
			 const std::string& xmlDir)
  : socketAddr_(socketAddr)
    , xmlDir_(xmlDir)
{
  //mutex̏
  int status = pthread_mutex_init(&loadXmlLock_, NULL);
  if (0 != status) {
    //throw OpenSOAPException();
    throw std::exception();
  }

  //fBNgpX/ŏI悤ɂĂ
  if (FILE_PATH_DELMIT != xmlDir_.at(xmlDir_.length() - 1)) {
    xmlDir_ += FILE_PATH_DELMIT;
  }
#ifdef DEBUG
  cerr << CLASS_SIG << "::" << CLASS_SIG << ": xmlDir_=["
       << xmlDir_ << "]" << endl;
#endif /* DEBUG */

  //SEM init
  key_t semKey = ftok(SRVCONFATTRMGR_SOCKET_ADDR.c_str(), 1);
  semid_ = semget(semKey, 1, IPC_CREAT|0666);
  if (-1 == semid_) {
    throw std::exception();
  }

#if defined(__GNU_LIBRARY__) && !defined(_SEM_SEMUN_UNDEFINED)
  union semun semUn;
#else
  union semun {
    int val;
    struct semid_ds *buf;
    unsigned short int *array;
    struct seminfo *__buf;
  } semUn;
#endif

  if (-1 == semctl(semid_, 0, GETVAL, semUn)) {
    throw std::exception();
  }
  semUn.val = 0;
  if (-1 == semctl(semid_, 0, SETVAL, semUn)) {
    throw std::exception();
  }

  //SSMLt@C̃[h
  if (0 != loadXml()) {
    cerr << TAG_ERR << CLASS_SIG << "::" << CLASS_SIG 
	 << ": load ssml file failed" << endl;
    throw std::exception();
  }
}

void 
SrvConfAttrMgr::freeXmlList()
{
  //xmlList_DOM\J
  std::list<xmlDocPtr>::iterator pos;
  for (pos = xmlList_.begin(); pos != xmlList_.end(); ++pos) {
    xmlFreeDoc(*pos);
  }
  xmlList_.clear();

}

SrvConfAttrMgr::~SrvConfAttrMgr()
{
  //xmlList_DOM\J
  freeXmlList();

  // mutex j
  int status = pthread_mutex_destroy(&loadXmlLock_);
  if (0 != status) {
    //throw OpenSOAPException();
    //throw std::exception();
    cerr << CLASS_SIG << "::" << CLASS_SIG
	 << ": mutext destroy failed" << endl;
  }
}

//ʐMJn
int SrvConfAttrMgr::run()
{
  //===================================================================
  // ʐM SrvConfAttrMgr <=> SSMLAttrHandler
  //===================================================================
  typedef struct sockaddr SockAddr;
  typedef struct sockaddr_un SockAddrUn;

  int sockfd = int();
  int newsockfd = int();
  SockAddrUn serverAddr;
  SockAddrUn clientAddr;
#if HAVE_TYPE_OF_ADDRLEN_AS_SOCKLEN_T
  socklen_t addrLen = socklen_t(); //socket_t();
  socklen_t clientLen = socklen_t();
#endif
#if HAVE_TYPE_OF_ADDRLEN_AS_SIZE_T
  size_t addrLen = size_t();
  size_t clientLen = size_t();
#endif
#if HAVE_TYPE_OF_ADDRLEN_AS_INT
  int addrLen = int();
  int clientLen = int();
#endif
  
  if (0 > (sockfd = socket(AF_UNIX, SOCK_STREAM, 0))) {
    cerr << TAG_ERR << CLASS_SIG << "::run: open socket failed" << endl;
    return -1;
  }

#ifdef DEBUG
  cerr << CLASS_SIG << "::run" 
       << ": socket open OK!" << endl;
#endif

  memset((char*)&serverAddr, 0, sizeof(serverAddr));
  serverAddr.sun_family = AF_UNIX;
  strcpy(serverAddr.sun_path, socketAddr_.c_str());
  addrLen = sizeof(serverAddr.sun_family) + strlen(serverAddr.sun_path);

  unlink(socketAddr_.c_str());
  
  if (0 > bind(sockfd, (SockAddr*)&serverAddr, addrLen)) {
    cerr << TAG_ERR << CLASS_SIG << "::run"
	 << ": bind failed" << endl;
    return -1;
  }

#ifdef DEBUG
  cerr << CLASS_SIG << "::run" 
       << ": bind OK!" << endl;
#endif

  if (0 > listen(sockfd, 5)) {
    cerr << TAG_ERR << CLASS_SIG << "::run"
	 << ": listen failed" << endl;
    return -1;
  }

#ifdef DEBUG
  cerr << CLASS_SIG << "::run" 
       << ": listen OK!" << endl;
#endif
    
  for (;;) {
    clientLen = sizeof(clientAddr);

#ifdef DEBUG
    cerr << CLASS_SIG << "::run" 
	 << ": accept ready!" << endl;
#endif
    
    newsockfd = accept(sockfd, (SockAddr*)&clientAddr, &clientLen);
    if (0 > newsockfd) {
      cerr << TAG_ERR << CLASS_SIG << "::run"
	   << ": accept failed" << endl;
      return -1;
    }

#ifdef DEBUG
    cerr << CLASS_SIG << "::run" 
	 << ": accept OK! fd=(" << newsockfd << ")" << endl;
#endif

    //XbhɓnKv̂f[^
    ThrData* thrData = new ThrData();
    thrData->that = this;
    thrData->sockfd = newsockfd;

#if 1
    //ʐMXbh쐬
    pthread_t thrId;
    int status = pthread_create(&thrId,
				NULL,
				connectionThread,
				(void*)thrData);
    if (0 != status) {
      cerr << TAG_ERR << CLASS_SIG << "::run"
	   << ": create thread failed" << endl;
      return -1;
    }
#else
    connectionThread(thrData);
#endif
  } // end for (;;)
}


//XMLt@C̓ǂݍ
int SrvConfAttrMgr::loadXml()
{ 
#ifdef DEBUG
  cerr << CLASS_SIG << "::loadXml()" << endl;
#endif /* DEBUG */

  // mutex lock
  // bNݒ
  int status = pthread_mutex_lock(&loadXmlLock_);
  if (0 != status) {
    //error
    cerr << TAG_ERR << CLASS_SIG << "::loadXml: mutex lock failed" << endl;
    return -1;
  }

  //xmlList_DOM\J
  freeXmlList();

  typedef struct direct Dir;

  DIR* dp = NULL;
  dirent* dir = NULL;
  //wfBNgJ
  if (NULL == (dp = opendir(xmlDir_.c_str()))) {
    cerr << TAG_ERR << CLASS_SIG << "::loadXml: cannot open directory.[" 
	 << xmlDir_ << "]" << endl;
    // "open dir failed " + fifoDirPath

    // bN
    status = pthread_mutex_unlock(&loadXmlLock_);
    if (0 != status) {
      cerr << TAG_ERR << CLASS_SIG << "::loadXml: mutex unlock failed" << endl;
      return -1;
    }
    return -1;
  }

  std::string ssmlFilePath;

  while (NULL != (dir = readdir(dp))) {
    if (0 == dir->d_ino) {
      cerr << "0 == dir->d_ino" << endl;
      continue;
    }

    //full path file name
    ssmlFilePath = xmlDir_ + dir->d_name;

    //t@Cgq̃`FbN
    if (CONF_FILE_SUFFIX != ssmlFilePath.substr(ssmlFilePath.rfind(".")+1)) {
      continue;
    }

#ifdef DEBUG
    cerr << CLASS_SIG << "::loadXml: xmlParseFile : file=["
	 << ssmlFilePath << "]" << endl;
#endif /* DEBUG */    

    /* parse ssml begin */
    /* COMPAT: Do not generate nodes for formatting spaces */
    LIBXML_TEST_VERSION
      xmlKeepBlanksDefault(0);
    xmlDocPtr doc = NULL;
    xmlNodePtr node1 = NULL;
  
    /* build an XML tree from a file */
    //std::string ssml_filename = "../ssml/Calc.ssml";
    
    doc = xmlParseFile((const char*)(ssmlFilePath.c_str()));
    if (doc == NULL) {
      cerr << TAG_WRN << CLASS_SIG << "::loadXml: "
	   << "parsing file " << ssmlFilePath << " failed." << endl;
      continue;
    }
    
    /* check document (root node) */
    node1 = xmlDocGetRootElement(doc);
    if (node1 == NULL) {
      cerr << TAG_WRN << CLASS_SIG << "::loadXml: "
	   << "empty document " << ssmlFilePath << endl;
      xmlFreeDoc(doc);
      continue;
    }
    //Xg֒ǉ
    xmlList_.push_back(doc);

#ifdef DEBUG
    cerr << CLASS_SIG << "::loadXml: add doc." << endl;
#endif /* DEBUG */
  }
  closedir(dp);

  // shm buffering clear
  clearShm();

  // bN
  status = pthread_mutex_unlock(&loadXmlLock_);
  if (0 != status) {
    cerr << TAG_ERR << CLASS_SIG << "::loadXml: mutex unlock failed" << endl;
    return -1;
  }

  return 0;
}

//ʐMXbh
void* SrvConfAttrMgr::connectionThread(void* arg)
{
#ifdef DEBUG
  cerr << CLASS_SIG << "::connectionThread()" << endl;
#endif /* DEBUG */

  //ñ`FbN
  if (!arg) {
    cerr << TAG_ERR << CLASS_SIG 
	 << "::connectionThread: invalid argument" << endl;
    pthread_exit(NULL);
  }

  //n̓WJ
  ThrData* thrData = (ThrData*)arg;
  SrvConfAttrMgr* that = thrData->that;
  int sockfd = thrData->sockfd;
  delete thrData;

#ifdef DEBUG
  cerr << CLASS_SIG << "::connectionThread: sockfd=(" << sockfd << ")" << endl;
#endif /* DEBUG */

  that->connectProc(sockfd);

  if (0 < sockfd) {
    close(sockfd);
  }

  pthread_exit(NULL);

  return NULL;
}

//ʐM
void SrvConfAttrMgr::connectProc(int sockfd)
{
#ifdef DEBUG
  cerr << CLASS_SIG << "::connectProc() arg=[" << sockfd << "]" << endl;
#endif /* DEBUG */

  //f[^M
  std::string readBuf;
  if (0 > read(sockfd, readBuf)) {
    cerr << TAG_ERR << CLASS_SIG 
	 << "::connectProc: read failed" << endl;
    return;
  }

#ifdef DEBUG
  cerr << CLASS_SIG << "::connectProc: read=["
       << readBuf << "]" << endl;
#endif

  //Mf[^̉

  MapUtil readDataMap(ITEM_DELMIT, VAL_DELMIT);
  readDataMap.insertItem(readBuf);

#ifdef DEBUG
  readDataMap.spy();
#endif /* DEBUG */

  //ԐMf[^
  std::string response;
  std::string retCode;
  MapUtil writeDataMap(ITEM_DELMIT, VAL_DELMIT);

  
  //R}h`FbN
  if (!readDataMap.exists(CMD)) {
    cerr << TAG_ERR << CLASS_SIG 
	 << "::connectProc: read failed" << endl;
    writeDataMap.insert(make_pair(RESULT, FAILURE));
    retCode = CMD + " not found";
    writeDataMap.insert(make_pair(RET_CODE, retCode));
    write(sockfd, writeDataMap.toString());

    return;
  }
  //R}heo
  std::string cmd = readDataMap[CMD];

  if (RELOAD_XML == cmd) {
    //SSMLt@C̃[h
    if (0 != loadXml()) {
      cerr << TAG_ERR << CLASS_SIG 
	   << "::connectProc: reload SSML failed" << endl;
      writeDataMap.insert(make_pair(RESULT, FAILURE));
      retCode = "reload failed";
      writeDataMap.insert(make_pair(RET_CODE, retCode));
      write(sockfd, writeDataMap.toString());

      return;
    }
    writeDataMap.insert(make_pair(RESULT, SUCCESS));
    write(sockfd, writeDataMap.toString());

    return;
  }
  else if (SEARCH == cmd) {
    //T[`f[^`FbN
    if (!readDataMap.exists(QUERY)) {
      cerr << TAG_ERR << CLASS_SIG 
	   << "::connectProc: search query not found." << endl;

      writeDataMap.insert(make_pair(RESULT, FAILURE));
      retCode = QUERY + " not found";
      writeDataMap.insert(make_pair(RET_CODE, retCode));
      write(sockfd, writeDataMap.toString());

      return;
    }
    //T[`eo
    std::string query = readDataMap[QUERY];

#ifdef DEBUG
    cerr << CLASS_SIG << "::connectProc: search query=[" << query << "]" << endl;
#endif /* DEBUG */

    response = queryXml(query);

#ifdef DEBUG
    cerr << CLASS_SIG << "::connectProc: queryXml response=["
	 << response << "]" << endl;
#endif /* DEBUG */
    
    write(sockfd, response);

    return;
    
  }
  else {
    cerr << TAG_ERR << CLASS_SIG 
	 << "::connectProc: invalid command. cmd=[" << cmd << "]" << endl;

    writeDataMap.insert(make_pair(RESULT, FAILURE));
    retCode = CMD + " invalid";
    writeDataMap.insert(make_pair(RET_CODE, retCode));
    write(sockfd, writeDataMap.toString());

    return;
  }
}

std::string SrvConfAttrMgr::queryXml(const std::string& query) 
{
#ifdef DEBUG
  cerr << CLASS_SIG << "::queryXml() arg=[" << query << "]" << endl;
#endif /* DEBUG */

  XmlQuery* xmlQuery = new XmlQuery(query);
  
  //mutex lock check.

#ifdef DEBUG
  cerr << "XmlQuery create done. doc size=(" 
       << xmlList_.size() << ")" << endl;
#endif /* DEBUG */

  std::vector<std::string> values;
  std::list<xmlDocPtr>::const_iterator pos;
  for (pos = xmlList_.begin(); pos != xmlList_.end(); ++pos) {
    xmlDocPtr doc = *pos;
    xmlNodePtr node = xmlDocGetRootElement(doc);
#ifdef DEBUG
    cerr << "### call getValue ###" << endl;
#endif /* DEBUG */

    int ret = xmlQuery->getValue(doc, node, values);

#ifdef DEBUG
    cerr << "xmlQuery ret=(" << ret << ")" << endl;
#endif /* DEBUG */

    if (0 == ret && !xmlQuery->isMulti()) {
      //if (0 == xmlQuery->getValue(doc, node, values)) {
      break;
    }
  }

  MapUtil writeDataMap(ITEM_DELMIT, VAL_DELMIT);

  if (0 < values.size()) {
    writeDataMap.insert(make_pair(RESULT, SUCCESS));
    std::string answer;
    std::vector<std::string>::const_iterator vpos;
    for (vpos = values.begin(); vpos != values.end(); ++vpos) {
      if (!answer.empty()) {
	answer += SUB_ITEM_DELMIT;
      }
      answer += *vpos;
    }
    writeDataMap.insert(make_pair(ANSWER, answer));

    //shm buffering
    addShm(query, answer);
  }
  else {
    writeDataMap.insert(make_pair(RESULT, FAILURE));
    std::string retCode = "querey failed";
    writeDataMap.insert(make_pair(RET_CODE, retCode));
  }

  delete xmlQuery;

  return writeDataMap.toString();
}

void SrvConfAttrMgr::clearShm()
{
  // SEM lock
  struct sembuf semBuf;
  // semVal -> 1
  semBuf.sem_num = 0;
  semBuf.sem_op = 1;
  semBuf.sem_flg = 0;
  if (-1 == semop(semid_, &semBuf, 1)) {
    cerr << TAG_ERR << CLASS_SIG << "::clearShm: semop up failed" << endl;
    return;
  }

  //̕ӂ̏͌ŐB(b)
  key_t shm_key = ftok(SRVCONFATTRMGR_SOCKET_ADDR.c_str(), 0);
  const int SHMSZ = 196608;
  int shmid = 0;
  char* shm = NULL;

  if ((shmid = shmget(shm_key, SHMSZ, IPC_CREAT|0666)) < 0) {
    cerr << TAG_ERR << CLASS_SIG << "::addShm: shmget failed." << endl;
    return;
  }

  if ((shm = (char*)shmat(shmid, NULL, 0)) == (char*)-1) {
    cerr << TAG_ERR << CLASS_SIG << "::addShm: shmat failed." << endl;
    return ;
  }

  *shm = 0x00;

  shmdt(shm);

  // SEM unlock
  // semVal -> 0
  semBuf.sem_num = 0;
  semBuf.sem_op = -1;
  semBuf.sem_flg = 0;
  if (-1 == semop(semid_, &semBuf, 1)) {
    cerr << TAG_ERR << CLASS_SIG << "::clearShm: semop down failed" << endl;
    return;
  }
}

void SrvConfAttrMgr::addShm(const std::string& query, const std::string& value)
{
  // SEM lock
  struct sembuf semBuf;
  // semVal -> 1
  semBuf.sem_num = 0;
  semBuf.sem_op = 1;
  semBuf.sem_flg = 0;
  if (-1 == semop(semid_, &semBuf, 1)) {
    cerr << TAG_ERR << CLASS_SIG << "::addShm: semop up failed" << endl;
    return;
  }

  //̕ӂ̏͌ŐB(b)
  key_t shm_key = ftok(SRVCONFATTRMGR_SOCKET_ADDR.c_str(), 0);
  const int SHMSZ = 196608;
  char shmBuf[SHMSZ];
  int shmid = 0;
  char* shm = NULL;

  if ((shmid = shmget(shm_key, SHMSZ, IPC_CREAT|0666)) < 0) {
    cerr << TAG_ERR << CLASS_SIG << "::addShm: shmget failed." << endl;
    return;
  }

  if ((shm = (char*)shmat(shmid, NULL, 0)) == (char*)-1) {
    cerr << TAG_ERR << CLASS_SIG << "::addShm: shmat failed." << endl;
    return ;
  }

  memset(shmBuf, 0x00, SHMSZ);
  strncpy(shmBuf, shm, SHMSZ);
  std::string newBuf(shmBuf);
  newBuf += "|";
  newBuf += query;
  newBuf += "|";
  newBuf += value;

  //newBuf.length >= SHMSZ -> delete cache
  strcpy(shm, newBuf.c_str());

  shmdt(shm);

  // SEM unlock
  // semVal -> 0
  semBuf.sem_num = 0;
  semBuf.sem_op = -1;
  semBuf.sem_flg = 0;
  if (-1 == semop(semid_, &semBuf, 1)) {
    cerr << TAG_ERR << CLASS_SIG << "::addShm: semop down failed" << endl;
    return;
  }
}
