/* -*- mode: c++; -*-
 *-----------------------------------------------------------------------------
 * $RCSfile: SrvDrvCreator_main.cpp,v $
 *
 * See Copyright for the status of this software.
 *
 * The OpenSOAP Project
 * http://opensoap.jp/
 *-----------------------------------------------------------------------------
 */

#include <fstream>
#include <cstdio>
#include <string>
#include <unistd.h>

#include "ServerCommon.h"
#include "SrvDrvCreator.h"

using namespace OpenSOAP;

int 
main(int argc, char** argv) {
  bool isAsync = false;
  
  // It works as Asynchronous Service Driver, 
  // except for argument is 0 or no argument exists
  if (argc > 1) {
    if (atoi(argv[1]) != 0) {
      isAsync = true;
    }
  }
  
  // make it to daemon & write pid file
#if HAVE_DAEMON
  daemon(0, 1);
#endif
  
  pid_t pid = getpid();
  
  ofstream pidFile;
  if (isAsync) {
    // Asynchronous Communication
    cout << "Async SrvDrvCreator" << endl;
    pidFile.open(ASYNC_SRVDRV_PID_FILE.c_str());
    if (!pidFile) {
      cerr << "Error: pid file [" 
	   << ASYNC_SRVDRV_PID_FILE << "] cannot open!" << endl;
      exit(1);
    }
    pidFile << pid;
    pidFile.close();
  }
  else {
    // Synchronous Communication
    cout << "Sync SrvDrvCreator" << endl;
    pidFile.open(SYNC_SRVDRV_PID_FILE.c_str());
    if (!pidFile) {
      cerr << "Error: pid file [" 
	   << SYNC_SRVDRV_PID_FILE << "] cannot open!" << endl;
      exit(1);
    }
  }
  
  pidFile << pid;
  pidFile.close();
  
  // create instance of QueueManager and run
  ChannelManager* chnlMngr = new SrvDrvCreator(isAsync);
  chnlMngr->run();

  delete chnlMngr;

  return EXIT_SUCCESS;
}
