/* -*- mode: c++; -*-
 *-----------------------------------------------------------------------------
 * $RCSfile: TTLManager.cpp,v $
 *
 * See Copyright for the status of this software.
 *
 * The OpenSOAP Project
 * http://opensoap.jp/
 *-----------------------------------------------------------------------------
 */

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <cstdio>
#include <unistd.h>
#include <string>

#include <stdexcept>

#include "StringUtil.h"
#include "MapUtil.h"
#include "ServerCommon.h"
#include "ChannelCommon.h"
#include "ChannelDescriptor.h"
#include "TTLManager.h"
#include "TTL.h"


using namespace OpenSOAP;

//for LOG
static const std::string CLASS_SIG = "TTLManager";

TTLManager::TTLManager()
{
#ifdef DEBUG
  cerr << CLASS_SIG << "::" << CLASS_SIG << endl;
#endif /* DEBUG */  
  
  std::string sockAddr = TTLMGR_SOCKET_ADDR;
  setSocketAddr(sockAddr);
  
  //CX^X
  ttlPtr = new TTL("./ttl_map");
  
  ttlPtr->watchInterval = 1;
  ttlPtr->ttlMaxValue = 3600;
  
}

TTLManager::~TTLManager()
{
#ifdef DEBUG
  cerr << CLASS_SIG << "::~" << CLASS_SIG << endl;
#endif /* DEBUG */  

}

void TTLManager::doProc(int sockfd)
{
#ifdef DEBUG
  cerr << CLASS_SIG << "::doProc() arg=[" << sockfd << "]" << endl;
#endif /* DEBUG */

  //f[^M
  std::string readBuf;
  if (0 > read(sockfd, readBuf)) {
    cerr << TAG_ERR << CLASS_SIG 
	 << "::doProc: read failed" << endl;
    return;
  }

#ifdef DEBUG
  cerr << CLASS_SIG << "::doProc: read=["
       << readBuf << "]" << endl;
#endif

  //Mf[^Kvɉĉ͂AԐMʂ𓾂

  MapUtil readDataMap(OpenSOAP::ChnlCommon::ITEM_DELMIT,
		      OpenSOAP::ChnlCommon::KEY_DELMIT);
  readDataMap.insertItem(readBuf);
  
#ifdef DEBUG
  readDataMap.spy();
#endif /* DEBUG */
  
  //bZ[W̉
  std::string response;  
  MapUtil writeDataMap(OpenSOAP::ChnlCommon::ITEM_DELMIT,
		       OpenSOAP::ChnlCommon::KEY_DELMIT);

  if (!readDataMap.exists(OpenSOAP::ChnlCommon::CMD)) {
    cerr << TAG_ERR << CLASS_SIG 
	 << "::doProc: read failed" << endl;
    return;
  }
  std::string cmd = readDataMap[OpenSOAP::ChnlCommon::CMD];
  
  if (OpenSOAP::ChnlCommon::OPEN == cmd) {
    //P[X
    
    //
    std::string s1(OpenSOAP::ChnlCommon::RESULT);
    std::string s2(OpenSOAP::ChnlCommon::SUCCESS);
    writeDataMap.insert(make_pair(s1,s2));
    //ʐM^Cv
    s1 = OpenSOAP::ChnlCommon::CHNL_TYPE;
    s2 = OpenSOAP::ChnlCommon::UNIX_SOCKET;
    writeDataMap.insert(make_pair(s1,s2));
    
    //ԐMf[^
    response = writeDataMap.toString();
  
    ChannelDescriptor chnlDesc;
    chnlDesc.setReadFd(sockfd);
    chnlDesc.setWriteFd(sockfd);
    
    if (0 > write(sockfd, response)) {
      cerr << TAG_ERR << CLASS_SIG 
	   << "::doProc: write failed" << endl;
      return;
    }
    
    //s
    ttlPtr->setChnlDesc(chnlDesc);
    
    if (readDataMap.exists(OpenSOAP::ChnlCommon::OPERATION)) {
      
      std::string opr = readDataMap[OpenSOAP::ChnlCommon::OPERATION];
      if (opr == OpenSOAP::ChnlCommon::PUSH) {
	ttlPtr->execPushOpr();
      }
      else if (opr == OpenSOAP::ChnlCommon::REF) {
	ttlPtr->execRefOpr();
      }
      else {
	cerr << TAG_ERR << CLASS_SIG 
	     << "::doProc: invalid operation[" << opr << "]" << endl;
      }
    }// end of if (readDataMap.exists(OpenSOAP::ChnlCommon::OPERATION))
    
  }
  else if (OpenSOAP::ChnlCommon::CLOSE == cmd) {
    //\[XJ
#ifdef DEBUG
    cerr << CLASS_SIG << "::doProc: CLOSE proc." << endl;
#endif /* DEBUG */

    std::string s1(OpenSOAP::ChnlCommon::RESULT);
    std::string s2(OpenSOAP::ChnlCommon::SUCCESS);
    writeDataMap.insert(make_pair(s1,s2));
    response = writeDataMap.toString();
  
    if (0 > write(sockfd, response)) {
      cerr << TAG_ERR << CLASS_SIG 
	   << "::doProc: write failed" << endl;
      return;
    }
    
  }
  close(sockfd);
  
#ifdef DEBUG
  cerr << CLASS_SIG << "::doProc: close(" << sockfd << ")" << endl;
#endif /* DEBUG */
  
}
