/* -*- mode: c++; -*-
 *-----------------------------------------------------------------------------
 * $RCSfile: XmlModifier.cpp,v $
 *
 * See Copyright for the status of this software.
 *
 * The OpenSOAP Project
 * http://opensoap.jp/
 *-----------------------------------------------------------------------------
 */

#include <string>

#include <stdexcept>

#include <libxml/tree.h>
#include <libxml/parser.h>

#include "ServerCommon.h"
#include "XmlModifier.h"

#include "XmlQuery2.h"

using namespace OpenSOAP;

// for LOG
static const std::string CLASS_SIG = "XmlModifier";

XmlModifier::XmlModifier(const std::string& xmlMsg)
  : xmlDoc(NULL)
{
  /* COMPAT: Do not generate nodes for formatting spaces */
  LIBXML_TEST_VERSION
    xmlKeepBlanksDefault(0);

  xmlDoc = xmlParseMemory((char*)xmlMsg.c_str(), xmlMsg.length());
  if (NULL == xmlDoc) {
    cerr << TAG_ERR << CLASS_SIG << "::" << CLASS_SIG
	 << ": xmlParseMemory returned NULL" << endl;
    throw std::exception();
  }
  /* check document (root node) */
  xmlNodePtr node1 = xmlDocGetRootElement(xmlDoc);
  if (node1 == NULL) {
    cerr << TAG_ERR << CLASS_SIG << "::" << CLASS_SIG
	 << ": empty document" << endl;
    xmlFreeDoc(xmlDoc);
    throw std::exception();
  }
}

XmlModifier::~XmlModifier()
{
  if (xmlDoc) {
    xmlFreeDoc(xmlDoc);
  }
}

std::string 
XmlModifier::toString()
{
  /* output doc tree */
  xmlChar* memPtr;
  int size;
  xmlDocDumpMemory(xmlDoc, &memPtr, &size);
  
  return std::string((char*)memPtr, size);
}

int
XmlModifier::attach(const std::string& fmt, const std::string& value)
{
  XmlQuery2 xmlQuery(fmt);
  return xmlQuery.attach(xmlDoc, xmlDocGetRootElement(xmlDoc), NULL,
			 value,
			 false, false, NULL);
}

int 
XmlModifier::attach(const std::string& fmt, 
		    const std::string& value,
		    const NameSpaceDef* nsDef)
{
  XmlQuery2 xmlQuery(fmt);
  XmlQuery2::NameSpaceDef* def = NULL;
  if (nsDef) {
    def = new XmlQuery2::NameSpaceDef;
    def->href = nsDef->href;
    def->prefix = nsDef->prefix;
  }
  return xmlQuery.attach(xmlDoc, xmlDocGetRootElement(xmlDoc), NULL,
			 value,
			 false, true, def);

}

int 
XmlModifier::update(const std::string& fmt, const std::string& value)
{
  XmlQuery2 xmlQuery(fmt);
  return xmlQuery.attach(xmlDoc, xmlDocGetRootElement(xmlDoc), NULL,
			 value,
			 true, false, NULL);
  
}


int
XmlModifier::append(const std::string& fmt, const std::string& value)
{
  XmlQuery2 xmlQuery(fmt);
  return xmlQuery.attach(xmlDoc, xmlDocGetRootElement(xmlDoc), NULL,
			 value,
			 false, false, NULL, 1);
}

int 
XmlModifier::append(const std::string& fmt, 
		    const std::string& value,
		    const NameSpaceDef* nsDef)
{
  XmlQuery2 xmlQuery(fmt);
  XmlQuery2::NameSpaceDef* def = NULL;
  if (nsDef) {
    def = new XmlQuery2::NameSpaceDef;
    def->href = nsDef->href;
    def->prefix = nsDef->prefix;
  }
  return xmlQuery.attach(xmlDoc, xmlDocGetRootElement(xmlDoc), NULL,
			 value,
			 false, true, def, 1);

}


    
